﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using System.Data.SqlClient;
using Ice.Orm.Core.Entity; 
using Ice.Orm.Core.Criteria;

namespace Ice.Net.Ecommerce.BusinessObjects.Database.VISTACART
{
	public partial class ProductCollection : List<Product>
	{
		#region Database methods
		// Database methods

		/// <summary>
		/// Load all products by categoryID
		/// </summary>
		/// <param name="categoryID"></param>
		/// <returns></returns>
		public static List<Product> LoadByCategoryID(int categoryID)
		{
			return new Query()
			.Select(new SqlName(db.Product + ".*"))
			.From(db.Product)
			.InnerJoin(db.ProductInCategory, db.Product.ProductID, db.ProductInCategory.ProductID)
			.Where(db.ProductInCategory.CategoryID, categoryID)
			.OrderBy(db.Product.Ordinal, true)
			.OrderBy(db.Product.ProductID, false)
			.ToList<Product>();
		}

		/// <summary>
		/// Load all products by campaignID
		/// </summary>
		/// <param name="campaignID"></param>
		/// <returns></returns>
		public static List<Product> LoadByCampaignID(int campaignID)
		{
			return new Query()
			.Select(new SqlName(db.Product + ".*"))
			.From(db.Product)
			.InnerJoin(db.ProductInCampaign, db.Product.ProductID, db.ProductInCampaign.ProductID)
			.InnerJoin(db.Campaign, db.Campaign.CampaignID, db.ProductInCampaign.CampaignID)
			.Where(db.ProductInCampaign.CampaignID, campaignID)
			.ToList<Product>();
		}

		public static int CountTotalProducts()
		{
			return new Query().Count(db.Product).ExecuteScalar<Product, int>();
		}

		public static int CountByCategoryID(int categoryID, bool activeOnly)
		{
			var q = new Query()
			.Count(db.Product)
			.InnerJoin(db.ProductInCategory, db.ProductInCategory.ProductID, db.Product.ProductID)
			.Where(db.ProductInCategory.CategoryID, categoryID)
			.Case(activeOnly, o => o.Where(db.Product.IsActive, true));

			return q.ExecuteScalar<Product, int>();
		}

		public static int CountByCampaignID(int campaignID, bool activeOnly)
		{
			var q = new Query()
			.Count(db.Product)
			.InnerJoin(db.ProductInCampaign, db.ProductInCampaign.ProductID, db.Product.ProductID)
			.Where(db.ProductInCampaign.CampaignID, campaignID)
			.Case(activeOnly, o => o.Where(db.Product.IsActive, true));

			return q.ExecuteScalar<Product, int>();
		}

		public static int CountBySearchExpression(string searchExpression, bool activeOnly)
		{
			var q = new Query()
			.Count(db.Product)
			.Case(activeOnly, o => o.Where(db.Product.IsActive, true))
			.Where(db.Product.Name, searchExpression, DBWhereKeyOperator.Like, DBWhereKeyLogic.And, DBWhereBrackets.Open)
			.Where(db.Product.ShortDescription, searchExpression, DBWhereKeyOperator.Like, DBWhereKeyLogic.Or)
			.Where(db.Product.LongDescription, searchExpression, DBWhereKeyOperator.Like, DBWhereKeyLogic.Or, DBWhereBrackets.Close);

			return q.ExecuteScalar<Product, int>();
		}

		public static List<Product> LoadPaged(int recordsPerPage, int currentPageIndex, object orderByColumn, bool orderAsc, bool activeOnly)
		{
			return new Query()
			.Case(activeOnly, o => o.Where(db.Product.IsActive, true))
			.Case(orderByColumn != null, o => o.OrderBy(new SqlName(orderByColumn + ""), !orderAsc))
			.Skip(currentPageIndex * recordsPerPage)
			.Take(recordsPerPage)
			.ToList<Product>();
		}

		public static List<Product> LoadPagedByCategoryID(int categoryID, int recordsPerPage, int currentPageIndex, object orderByColumn, bool orderAsc, bool activeOnly)
		{
			return new Query()
			.Select(new SqlName(db.Product + ".*"))
			.From(db.Product)
			.InnerJoin(db.ProductInCategory, db.Product.ProductID, db.ProductInCategory.ProductID)
			.Where(db.ProductInCategory.CategoryID, categoryID)
			.Case(activeOnly, o => o.Where(db.Product.IsActive, true))
			.Case(orderByColumn != null, o => o.OrderBy(new SqlName(orderByColumn + ""), !orderAsc))
			.Skip(currentPageIndex * recordsPerPage)
			.Take(recordsPerPage)
			.ToList<Product>();
		}

		public static List<Product> LoadPagedByCampaignID(int campaignID, int recordsPerPage, int currentPageIndex, object orderByColumn, bool orderAsc, bool activeOnly)
		{
			return new Query()
			.Select(new SqlName(db.Product + ".*"))
			.From(db.Product)
			.InnerJoin(db.ProductInCampaign, db.Product.ProductID, db.ProductInCampaign.ProductID)
			.Where(db.ProductInCampaign.CampaignID, campaignID)
			.Case(activeOnly, o => o.Where(db.Product.IsActive, true))
			.Case(orderByColumn != null, o => o.OrderBy(new SqlName(orderByColumn + ""), !orderAsc))
			.Skip(recordsPerPage * currentPageIndex)
			.Take(recordsPerPage)
			.ToList<Product>();
		}

		public static List<Product> LoadPagedBySearchExpression(string searchExpression, int recordsPerPage, int currentPageIndex, object orderByColumn, bool orderAsc, bool activeOnly)
		{
			return new Query()
			.Case(activeOnly, o => o.Where(db.Product.IsActive, true))
			.Where(db.Product.Name, searchExpression, DBWhereKeyOperator.Like, DBWhereKeyLogic.And, DBWhereBrackets.Open)
			.Where(db.Product.ShortDescription, searchExpression, DBWhereKeyOperator.Like, DBWhereKeyLogic.Or)
			.Where(db.Product.LongDescription, searchExpression, DBWhereKeyOperator.Like, DBWhereKeyLogic.Or, DBWhereBrackets.Close)
			.Case(orderByColumn != null, o => o.OrderBy(new SqlName(orderByColumn + ""), !orderAsc))
			.Skip(recordsPerPage * currentPageIndex)
			.Take(recordsPerPage)
			.ToList<Product>();
		}

		/// <summary>
		/// Load all downloadable products
		/// <para>set categoryID = 0 to load products from all categories</para>
		/// </summary>
		public static List<Product> LoadDownloadableProducts(bool activeOnly)
		{
			object orderByColumn = db.Product.Ordinal;
			bool orderAsc = false;

			return new Query()
			.Where(db.Product.IsDownloadable, true)
			.Case(activeOnly, o => o.Where(db.Product.IsActive, true))
			.Case(orderByColumn != null, o => o.OrderBy(new SqlName(orderByColumn + ""), !orderAsc))
			.Take(100)
			.ToList<Product>();
		}

		public static List<Product> LoadNewProducts(int maxObjectsToLoad, bool sortAsc, bool activeOnly)
		{
			return new Query()
			.Case(activeOnly, o => o.Where(db.Product.IsActive, true))
			.OrderBy(db.Product.ProductID, !sortAsc)
			.Take(maxObjectsToLoad)
			.ToList<Product>();
		}

		#endregion

		#region Custom methods

		/// <summary>
		/// Returns Product or null if not found
		/// </summary>
		/// <param name="productId"></param>
		/// <returns></returns>
		public Product FindProduct(int productId)
		{
			foreach (Product product in this)
			{
				if (product.ProductID == productId)
				{
					return product;
				}
			}
			return null;
		}

		#endregion
	}
}
