﻿using System;
using System.Collections.Generic;
using System.Text;
using Ice.Net.Ecommerce.BusinessObjects.Database.VISTACART;
using GCART = Ice.Net.Ecommerce.BusinessObjects.Database.VISTACART;
using Ice.Net.Ecommerce.BusinessObjects.Common;
using Ice.Net.Ecommerce.BusinessObjects.Gallery;
using Ice.Net.Ecommerce.EmailUtility;

namespace Ice.Net.Ecommerce.NewsletterUtility
{
	#region .NET / JSON objects

	class NewsObject
	{
		public String ToEmail;
		public String Subject;
		public String Body;

		public NewsObject(string toEmail, string subject, string body)
		{
			this.ToEmail = toEmail;
			this.Subject = subject;
			this.Body = body;
		}
	}

	public class NewsStatus
	{
		public int QueueCount;
		public int SentCount;
		public int FailedCount;
		public int SendingCount;

	}

	public class NewsLog
	{
		public List<string> SentEmails;
		public List<string> FailedEmails;
	}

	#endregion

	internal class NewsletterSender
	{
		private static WorkerThreadsController mNewsletterSenderController = null;

		private static List<string> mListSentEmails = new List<string>();
		private static List<string> mListFailedEmails = new List<string>();
		private static List<NewsObject> mNewsObjectCollection = new List<NewsObject>();

		private static Object mLockSentEmails = new object();
		private static Object mLockFailedEmails = new object();
		private static Object mLockNewsObjectCollection = new object();
		
		private static string mFromEmail = null;
		private static string mFromName = null;
		private static string mServer = null;
		private static int mPort = 25;
		private static string mUsername = null;
		private static string mPassword = null;

		/// <summary>
		/// Add emails and start sending.
		/// </summary>
		/// <param name="emails"></param>
		/// <param name="subject"></param>
		/// <param name="bodyHtml"></param>
		/// <returns></returns>
		public static bool SendNewsletter(List<string> emails, string subject, string bodyHtml)
		{
			mFromEmail = VISTACARTHelper.AppDefaults.AppGlobalSettingsItem.FromEmail;
			mFromName = VISTACARTHelper.AppDefaults.AppGlobalSettingsItem.FromName;
			mServer = VISTACARTHelper.AppDefaults.AppGlobalSettingsItem.EmailServer;
			mPort = VISTACARTHelper.AppDefaults.AppGlobalSettingsItem.EmailPort;
			mUsername = VISTACARTHelper.AppDefaults.AppGlobalSettingsItem.EmailUsername;
			mPassword = VISTACARTHelper.AppDefaults.AppGlobalSettingsItem.EmailPassword;

			lock (mLockNewsObjectCollection)
			{
				foreach (string email in emails)
				{
					mNewsObjectCollection.Add(new NewsObject(email, subject, bodyHtml));
				}
			}

			StartSending();

			return true;
		}

		private static void StartSending()
		{
			if (mNewsletterSenderController == null)
			{
				mNewsletterSenderController = new WorkerThreadsController(10, SendNext);
			}

			mNewsletterSenderController.StartThreads();
		}

		/// <summary>
		/// Abort sending threads, but don't reset status.
		/// </summary>
		public static void Abort()
		{
			if (mNewsletterSenderController != null)
			{
				mNewsletterSenderController.AbortThreads();
			}
			mNewsletterSenderController = null;
		}

		/// <summary>
		/// Start sending threads.
		/// </summary>
		public static void Start()
		{
			StartSending();
		}

		/// <summary>
		/// Abort sending threads and reset status.
		/// </summary>
		public static void Reset()
		{
			Abort();
			mListSentEmails = new List<string>();
			mListFailedEmails = new List<string>();
			mNewsObjectCollection = new List<NewsObject>();
		}

		/// <summary>
		/// Get status.
		/// </summary>
		/// <returns></returns>
		public static NewsStatus GetStatus()
		{
			NewsStatus status = new NewsStatus();

			lock (mLockNewsObjectCollection)
			{
				status.QueueCount = mNewsObjectCollection.Count;
			}
			lock (mLockFailedEmails)
			{
				status.FailedCount = mListFailedEmails.Count;
			}
			lock (mLockSentEmails)
			{
				status.SentCount = mListSentEmails.Count;
			}
			if (mNewsletterSenderController == null)
			{
				status.SendingCount = 0;
			}
			else
			{
				status.SendingCount = mNewsletterSenderController.AliveThreadsCount;
			}
			return status;
		}

		/// <summary>
		/// Get object with list of sent emails and list of failed emails.
		/// </summary>
		/// <returns></returns>
		public static NewsLog GetLog()
		{
			NewsLog nlog = new NewsLog();
			nlog.SentEmails = mListSentEmails;
			nlog.FailedEmails = mListFailedEmails;
			return nlog;
		}


		#region sending engine

		private static NewsObject GetNextEmail()
		{
			lock (mLockNewsObjectCollection)
			{
				if (mNewsObjectCollection.Count > 0)
				{
					NewsObject newsObject = mNewsObjectCollection[0];
					mNewsObjectCollection.RemoveAt(0);
					return newsObject;
				}
				else
				{
					return null;
				}
			}
		}

		private static WorkerThreadMethodResult SendNext()
		{
			NewsObject newsObject = GetNextEmail();
			if (newsObject == null)
			{
				return WorkerThreadMethodResult.StopThread;
			}
			else
			{
				List<string> failedEmails = EmailSenderImpl.SendEmails(
					mServer, 
					mPort, 
					mUsername, 
					mPassword, 
					newsObject.Subject, 
					newsObject.Body, 
					true, 
					null, 
					mFromEmail, 
					mFromName, 
					new String[] { newsObject.ToEmail });

				if (failedEmails.Count > 0)
				{
					lock (mLockFailedEmails)
					{
						foreach (string errorEmail in failedEmails)
						{
							mListFailedEmails.Add(errorEmail);
						}
					}
				}
				else
				{
					lock (mLockSentEmails)
					{
						mListSentEmails.Add(newsObject.ToEmail);
					}
				}

				return WorkerThreadMethodResult.Successful;
			}
		}

		#endregion


	}
}
