﻿using System;
using System.Collections.Generic;
using System.Text;

namespace Ice.Net.Ecommerce.Security.Cryptography
{

	internal class RSASecretKey
	{
		// private key: d, n
		// pubilc key: e, n

		/// <summary>
		/// Modulus - used for private and public key.
		/// </summary>
		public BigInteger n;	    // public modulus
		/// <summary>
		/// Exponent - public key.
		/// </summary>
		public BigInteger e;	    // public exponent
		/// <summary>
		/// D - Private key.
		/// </summary>
		public BigInteger d;	    // private exponent
		/// <summary>
		/// Prime P
		/// </summary>
		public BigInteger p;	    // prime  p. // not required
		/// <summary>
		/// Prime Q
		/// </summary>
		public BigInteger q;	    // prime  q. // not required
		/// <summary>
		/// Inverse of P mod Q
		/// </summary>
		public BigInteger u;	    // inverse of p mod q. // not requeireds

		#region public properties

		public bool IsPublicKey
		{
			get { return ( Object.ReferenceEquals( this.e , null) == false && Object.ReferenceEquals(this.n, null) == false); }
		}

		public bool IsPrivateKey
		{
			get { return (Object.ReferenceEquals(this.d, null) == false && Object.ReferenceEquals(this.n, null) == false); }
		}

		#endregion

		#region private methods

		private static BigInteger GetBi(byte[] data)
		{
			if (data == null)
			{
				return null;
			}
			else
			{
				return new BigInteger(data);
			}
		}

		#endregion

		#region constructors

		public RSASecretKey()
		{
		}

		public RSASecretKey(RSAParameters2 rsaPar2)
		{
			this.d = GetBi(rsaPar2.D);
			this.e = GetBi(rsaPar2.Exponent);
			this.n = GetBi(rsaPar2.Modulus);
			this.p = GetBi(rsaPar2.P);
			this.q = GetBi(rsaPar2.Q);
		}

		#endregion

		#region Key generator

		/// <summary>
		/// Default key size: 1024, use internal BigInteger to generate keys.
		/// </summary>
		/// <returns></returns>
		public static RSASecretKey GenerateKeys()
		{
			return GenerateKeys(1024);
		}

		public static RSASecretKey GenerateKeys(int nbits, bool useMSRSACryptoProvider)
		{
			if (useMSRSACryptoProvider)
			{
				System.Security.Cryptography.RSACryptoServiceProvider rsa = new System.Security.Cryptography.RSACryptoServiceProvider(nbits);
				System.Security.Cryptography.RSAParameters par = rsa.ExportParameters(true);
				RSASecretKey skey = new RSASecretKey();
				skey.e = new BigInteger(par.Exponent);
				skey.d = new BigInteger(par.D);
				skey.n = new BigInteger(par.Modulus);

				skey.p = new BigInteger(par.P); // not required
				skey.q = new BigInteger(par.Q); // not required
				return skey;
			}
			else
			{
				return GenerateKeys(nbits);
			}
		}

		private static RSASecretKey GenerateKeys(int nbits)
		{
			BigInteger p, q; /* the two primes */
			BigInteger d;    /* the private key */
			BigInteger u;
			BigInteger t1, t2;
			BigInteger n = new BigInteger();    /* the public key */
			BigInteger e;    /* the exponent */
			BigInteger phi;  /* helper: (p-1)(q-1) */
			BigInteger g;
			BigInteger f;
			Random rand = new Random();

			if ((nbits < 768) || (nbits > 4096))
				throw new ArgumentException("Only keysizes betwen 768 and 4096 bit are allowed!");

			/* make sure that nbits is even so that we generate p, q of equal size */
			if ((nbits & 1) == 1)
				nbits++;

			do
			{
				/* select two (very secret) primes */
				p = new BigInteger();
				q = new BigInteger();

				p = BigInteger.genPseudoPrime(nbits / 2, 10, rand);
				q = BigInteger.genPseudoPrime(nbits / 2, 10, rand);

				/* p shall be smaller than q (for calc of u)*/
				if (q > p)
				{
					BigInteger tmp = p;
					p = q;
					q = tmp;
				}

				/* calculate the modulus */
				n = p * q;
			} while (n.bitCount() != nbits);

			/* calculate Euler totient: phi = (p-1)(q-1) */
			t1 = p - new BigInteger(1);
			t2 = q - new BigInteger(1);
			phi = t1 * t2;

			g = t1.gcd(t2);
			f = phi / g;

			/* find an public exponent.
			We use 41 as this is quite fast and more secure than the
			commonly used 17.
			*/

			e = new BigInteger(41);
			t1 = e.gcd(phi);
			if (t1 != new BigInteger(1))
			{
				e = new BigInteger(257);
				t1 = e.gcd(phi);
				if (t1 != new BigInteger(1))
				{
					e = new BigInteger(65537);
					t1 = e.gcd(phi);

					/* (while gcd is not 1) */
					while (t1 != new BigInteger(1))
					{
						e += 2;
						t1 = e.gcd(phi);
					}
				}
			}

			/* calculate the secret key d = e^1 mod phi */
			d = e.modInverse(f);

			/* calculate the inverse of p and q (used for chinese remainder theorem)*/
			u = p.modInverse(q);

			RSASecretKey sk = new RSASecretKey();

			sk.n = n;
			sk.e = e;
			sk.p = p;
			sk.q = q;
			sk.d = d;
			sk.u = u;

			//this.biGeneratedKey = ParseSecretKey(sk);

			return sk;// this.biGeneratedKey;

			/* now we can test our keys (this should never fail!) */
			// test_keys( sk, nbits - 64 );
		}

		#endregion

	}
}
