﻿/***********************************************
 * Modified by Ice to return Sequence
 * datbase file is created by Elmah code
 * removed parameter Application from WHERE
 ***********************************************/

[assembly: Ice.Elmah.Scc("$Id: AccessErrorLog.cs 829 2011-05-03 21:59:40Z azizatif $")]

namespace Ice.Elmah
{
	#region Imports

	using System;
	using System.Data;
	using System.Data.OleDb;
	using System.Diagnostics;
	using System.Globalization;
	using System.IO;
	using System.Text;
	using IDictionary = System.Collections.IDictionary;
	using IList = System.Collections.IList;

	#endregion

	/// <summary>
	/// An <see cref="ErrorLog"/> implementation that uses Microsoft Access 
	/// as its backing store.
	/// </summary>
	/// <remarks>
	/// The MDB file is automatically created at the path specified in the 
	/// connection string if it does not already exist.
	/// </remarks>

	public class AccessErrorLog : ErrorLog
	{
		private readonly string _connectionString;

		private const int _maxAppNameLength = 60;
		private const string _scriptResourceName = "mkmdb.vbs";

		private static readonly object _mdbInitializationLock = new object();

		/// <summary>
		/// Initializes a new instance of the <see cref="AccessErrorLog"/> class
		/// using a dictionary of configured settings.
		/// </summary>

		public AccessErrorLog(IDictionary config)
		{
			if (config == null)
				throw new ArgumentNullException("config");

			string connectionString = ConnectionStringHelper.GetConnectionString(config);

			//
			// If there is no connection string to use then throw an 
			// exception to abort construction.
			//

			if (connectionString.Length == 0)
				throw new ApplicationException("Connection string is missing for the Access error log.");

			_connectionString = connectionString;

			InitializeDatabase();

			//
			// Set the application name as this implementation provides
			// per-application isolation over a single store.
			//

			string appName = Mask.NullString((string)config["applicationName"]);

			if (appName.Length > _maxAppNameLength)
			{
				throw new ApplicationException(string.Format(
					"Application name is too long. Maximum length allowed is {0} characters.",
					_maxAppNameLength.ToString("N0")));
			}

			ApplicationName = appName;
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="AccessErrorLog"/> class
		/// to use a specific connection string for connecting to the database.
		/// </summary>

		public AccessErrorLog(string connectionString)
		{
			if (connectionString == null)
				throw new ArgumentNullException("connectionString");

			if (connectionString.Length == 0)
				throw new ArgumentException(null, "connectionString");

			_connectionString = connectionString;
		}

		/// <summary>
		/// Gets the name of this error log implementation.
		/// </summary>

		public override string Name
		{
			get { return "Microsoft Access Error Log"; }
		}

		/// <summary>
		/// Gets the connection string used by the log to connect to the database.
		/// </summary>

		public virtual string ConnectionString
		{
			get { return _connectionString; }
		}

		/// <summary>
		/// Logs an error to the database.
		/// </summary>
		/// <remarks>
		/// Use the stored procedure called by this implementation to set a
		/// policy on how long errors are kept in the log. The default
		/// implementation stores all errors for an indefinite time.
		/// </remarks>

		public override string Log(Error error)
		{
			if (error == null)
				throw new ArgumentNullException("error");

			string errorXml = ErrorXml.EncodeString(error);

			Guid id = Guid.NewGuid();

			using (OleDbConnection connection = new OleDbConnection(this.ConnectionString))
			using (OleDbCommand command = connection.CreateCommand())
			{
				command.CommandText = @"INSERT INTO ELMAH_Error 
([ErrorId], [Application], [Host], [Type], [Source], [Message], [User], [StatusCode], [TimeUtc], [AllXml] )
	VALUES 
(@ErrorId, @Application, @Host, @Type, @Source, @Message, @User, @StatusCode, @TimeUtc, @AllXml);";

				OleDbParameterCollection parameters = command.Parameters;

				parameters.Add("@ErrorId", OleDbType.Guid).Value = id;
				parameters.Add("@Application", OleDbType.VarChar, _maxAppNameLength).Value = ApplicationName;
				parameters.Add("@Host", OleDbType.VarChar, 30).Value = error.HostName;
				parameters.Add("@Type", OleDbType.VarChar, 100).Value = error.Type;
				parameters.Add("@Source", OleDbType.VarChar, 60).Value = error.Source;
				parameters.Add("@Message", OleDbType.LongVarChar, error.Message.Length).Value = error.Message;
				parameters.Add("@User", OleDbType.VarChar, 50).Value = error.User;
				parameters.Add("@StatusCode", OleDbType.Integer).Value = error.StatusCode;
				parameters.Add("@TimeUtc", OleDbType.Date).Value = error.Time.ToUniversalTime();
				parameters.Add("@AllXml", OleDbType.LongVarChar, errorXml.Length).Value = errorXml;

				connection.Open();
				command.ExecuteNonQuery();
				return id.ToString();
			}
		}

		/// <summary>
		/// Returns a page of errors from the databse in descending order 
		/// of logged time.
		/// </summary>

		public override int GetErrors(int pageIndex, int pageSize, IList errorEntryList)
		{
			if (pageIndex < 0)
				throw new ArgumentOutOfRangeException("pageIndex", pageIndex, null);

			if (pageSize < 0)
				throw new ArgumentOutOfRangeException("pageSize", pageSize, null);

			using (OleDbConnection connection = new OleDbConnection(this.ConnectionString))
			using (OleDbCommand command = connection.CreateCommand())
			{
				command.CommandType = CommandType.Text;
				command.CommandText = "SELECT COUNT(*) FROM ELMAH_Error";

				connection.Open();
				int totalCount = (int)command.ExecuteScalar();

				if (errorEntryList != null && pageIndex * pageSize < totalCount)
				{
					int maxRecords = pageSize * (pageIndex + 1);
					if (maxRecords > totalCount)
					{
						maxRecords = totalCount;
						pageSize = totalCount - pageSize * (totalCount / pageSize);
					}

					StringBuilder sql = new StringBuilder(1000);
					sql.Append("SELECT e.* FROM (");
					sql.Append("SELECT TOP ");
					sql.Append(pageSize.ToString(CultureInfo.InvariantCulture));
					sql.Append(" TimeUtc, Sequence FROM (");
					sql.Append("SELECT TOP ");
					sql.Append(maxRecords.ToString(CultureInfo.InvariantCulture));
					sql.Append(" TimeUtc, Sequence FROM ELMAH_Error ");
					sql.Append("ORDER BY TimeUtc DESC, Sequence DESC) ");
					sql.Append("ORDER BY TimeUtc ASC, Sequence ASC) AS i ");
					sql.Append("INNER JOIN Elmah_Error AS e ON i.Sequence = e.Sequence ");
					sql.Append("ORDER BY e.TimeUtc DESC, e.Sequence DESC");

					command.CommandText = sql.ToString();

					using (OleDbDataReader reader = command.ExecuteReader())
					{
						Debug.Assert(reader != null);

						while (reader.Read())
						{
							string id = reader["ErrorId"].ToString();

							Error error = new Error();
							error.Sequence = Convert.ToInt32(reader["Sequence"]); // Ice added
							error.ApplicationName = reader["Application"].ToString();
							error.HostName = reader["Host"].ToString();
							error.Type = reader["Type"].ToString();
							error.Source = reader["Source"].ToString();
							error.Message = reader["Message"].ToString();
							error.User = reader["User"].ToString();
							error.StatusCode = Convert.ToInt32(reader["StatusCode"]);
							error.Time = Convert.ToDateTime(reader["TimeUtc"]).ToLocalTime();
							errorEntryList.Add(new ErrorLogEntry(this, id, error));
						}

						reader.Close();
					}
				}

				return totalCount;
			}
		}

		/// <summary>
		/// Returns the specified error from the database, or null 
		/// if it does not exist.
		/// </summary>
		public override ErrorLogEntry GetError(string id)
		{
			if (id == null)
				throw new ArgumentNullException("id");

			if (id.Length == 0)
				throw new ArgumentException(null, "id");

			Guid errorGuid;

			try
			{
				errorGuid = new Guid(id);
			}
			catch (Exception e)
			{
				throw new ArgumentException(e.Message, "id", e);
			}

			using (OleDbConnection connection = new OleDbConnection(this.ConnectionString))
			{
				using (OleDbCommand command = connection.CreateCommand())
				{
					command.CommandText = @"SELECT [AllXml] FROM [ELMAH_Error] WHERE [ErrorId] = @ErrorId";

					command.Parameters.Add("@ErrorId", OleDbType.Guid).Value = errorGuid;

					connection.Open();

					string errorXml = (string)command.ExecuteScalar();

					if (errorXml == null)
						return null;

					Error error = ErrorXml.DecodeString(errorXml);
					return new ErrorLogEntry(this, id, error);
				}
			}
		}

		private void InitializeDatabase()
		{
			string connectionString = ConnectionString;
			Debug.AssertStringNotEmpty(connectionString);

			string dbFilePath = ConnectionStringHelper.GetDataSourceFilePath(connectionString);
			if (File.Exists(dbFilePath))
				return;

			//
			// Make sure that we don't have multiple instances trying to create the database.
			//

			lock (_mdbInitializationLock)
			{
				//
				// Just double-check that no other thread has created the database while
				// we were waiting for the lock.
				//

				if (File.Exists(dbFilePath))
					return;

				//
				// Create a temporary copy of the mkmdb.vbs script.
				// We do this in the same directory as the resulting database for security permission purposes.
				//

				string scriptPath = Path.Combine(Path.GetDirectoryName(dbFilePath), _scriptResourceName);

				using (FileStream scriptStream = new FileStream(scriptPath, FileMode.Create, FileAccess.Write, FileShare.None))
				{
					ManifestResourceHelper.WriteResourceToStream(scriptStream, _scriptResourceName);
				}

				//
				// Run the script file to create the database using batch 
				// mode (//B), which suppresses script errors and prompts 
				// from displaying.
				//

				ProcessStartInfo psi = new ProcessStartInfo(
					"cscript", "\"" + scriptPath + "\" \"" + dbFilePath + "\" //B //NoLogo");

				psi.UseShellExecute = false;    // i.e. CreateProcess
				psi.CreateNoWindow = true;      // Stay lean, stay mean

				try
				{
					using (Process process = Process.Start(psi))
					{
						//
						// A few seconds should be plenty of time to create the database.
						//

						TimeSpan tolerance = TimeSpan.FromSeconds(2);
						if (!process.WaitForExit((int)tolerance.TotalMilliseconds))
						{
							//
							// but it wasn't, so clean up and throw an exception!
							// Realistically, I don't expect to ever get here!
							//

							process.Kill();

							throw new Exception(string.Format(
								"The Microsoft Access database creation script took longer than the allocated time of {0} seconds to execute. "
								+ "The script was terminated prematurely.",
								tolerance.TotalSeconds));
						}

						if (process.ExitCode != 0)
						{
							throw new Exception(string.Format(
								"The Microsoft Access database creation script failed with exit code {0}.",
								process.ExitCode));
						}
					}
				}
				finally
				{
					//
					// Clean up after ourselves!!
					//

					File.Delete(scriptPath);
				}
			}
		}
	}
}
