﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Ice.Orm.Core.Criteria;
using System.Collections;
using System.Collections.ObjectModel;
using Ice.Orm.Core.Criteria.SqlBlocks;
using System.Data;
using Ice.Orm.Core.DataStore.MSSQL;

namespace Ice.Orm.Core.Criteria
{
	internal class CompiledQuery
	{
		public int SkipWithReader { get; private set;}
		internal QueryBase BaseQuery { get; private set; }

		public string QueryText{get;private set;}
		public string QueryWithValues{get; private set;}
		private ReadOnlyCollection<QParameter> Parameters{get; set;}
		public ReadOnlyCollection<IDbDataParameter> DbParameters 
		{ 
			get
			{
				// Buduci da SchemaTableHelper dohvaca shemu iz baze on demand, nek se i parametri kreiraju on-demand da ne baca exception ako se zeli samo dobiti query text 
				return this.BaseQuery.DataStore.ConvertParameters(this.Parameters, this.BaseQuery.SchemaTableHelper);
			}
		}

		internal CompiledQuery(QueryBase query)
		{
			this.BaseQuery = query;
			this.Compile();
		}

		private static string GetSqlQuotedValue(object value)
		{
			if (value == null || value == DBNull.Value)
			{
				return "NULL";
			}
			else if (value is Boolean)
			{
				return ((bool)value) ? "1" : "0";
			}
			else if (value is DateTime)
			{
				return "'" + ((DateTime)value).ToString("yyyy-MM-dd HH:mm:ss.fff") + "'";
			}
			else if (value is string || value is char)
			{
				return "'" + value + "'";
			}
			else if (value is IEnumerable)
			{
				return "IN (" + ((IEnumerable)value).Cast<object>().ToStringJoined(", ") + ")";
			}
			else
			{
				return value + "";
			}
		}

		private static bool IsSpaceRequired(string word, string previous)
		{
			char? firstChar = !string.IsNullOrEmpty(word) ? word[0] : (char?)null;
			char? lastChar = !string.IsNullOrEmpty(previous) ? previous[previous.Length - 1] : (char?)null;

			if (firstChar != null && lastChar != null)
			{
				switch (firstChar)
				{
					case '\r':
					case '\n':
					case ',':
					case ')':
					case ';':
						return false; // don't add space if text starts with start char
				}

				switch (lastChar)
				{
					case ' ':
					case '\r':
					case '\n':
					case '\t':
					case '(':
						return false; // don't add space if last char in QueryText is c
				}

				return true;
			}
			return false;
		}

		private void Compile()
		{
			if (this.BaseQuery.DataStore == null)
			{
				// use MS SQL data store as default, required for self-test in new Query().CreateQuery2("SELECT....
				this.BaseQuery.DataStore = new MSSQLDataStore(null);
			}

			SqlSelectBlock select = this.BaseQuery.GetSqlBlock<SqlSelectBlock>();
			select.ParentherizeTop = this.BaseQuery.DataStore.ParentherizeTop;
			select.UseOffsetAndLimit = this.BaseQuery.DataStore.UseOffsetAndLimit;

			if (this.BaseQuery.DataStore.UseOffsetAndLimit)
			{
				SqlOffsetLimitBlock offsetLimit = this.BaseQuery.GetSqlBlock<SqlOffsetLimitBlock>();
				offsetLimit.UseOffsetAndLimit = this.BaseQuery.DataStore.UseOffsetAndLimit;
				offsetLimit.Offset = select.Skip;
				offsetLimit.Limit = select.Take;
			}

			StringBuilder sql = new StringBuilder();
			StringBuilder sqlWithValues = new StringBuilder();
			QParameterCollection parameters = new QParameterCollection() { ParameterPrefix = this.BaseQuery.DataStore.ParameterPreffix };

			string previous = null;

			foreach(SqlSegment o in this.BaseQuery.SelectSqlSegments())
			{
				string val = null;
				string word = null;

				if (o.Parameter != null)
				{
					val = (GetSqlQuotedValue(o.Parameter.Value));
					// create new parameters every time to have it with Name = null so the new name will be created according to paramenter name
					word = (parameters.Add(new QParameter(o.Parameter.Value) { Column = o.Parameter.Column, Table = o.Parameter.Table }).Name);
					if (!this.BaseQuery.DataStore.UseNamedParameters)
					{
						word = "?";
					}
				}
				else if (o.Text != null)
				{
					// trim only space, don't trim \r or \n because it is used in \r\nWHERE, \r\nJOIN etc. inside SqlWhereBlock and others
					val = word = o.Text.Trim(' ');
				}
				else
				{
					// both Text and Parameter values are null
					continue;
				}

				if (!string.IsNullOrEmpty(word))
				{
					if (previous != null && IsSpaceRequired(word, previous))
					{
						sql.Append(" ");
						sqlWithValues.Append(" ");
					}

					sql.Append(word);
					previous = word;

					if (!string.IsNullOrEmpty(val))
					{
						sqlWithValues.Append(val);
					}
				}
			}

			this.SkipWithReader = this.BaseQuery.DataStore.UseOffsetAndLimit ? 0 : select.Skip;
			this.Parameters = parameters.AsReadOnly();
			this.QueryText = this.BaseQuery.DataStore.CleanUpSqlText(sql.ToString().Trim());
			this.QueryWithValues = this.BaseQuery.DataStore.CleanUpSqlText(sqlWithValues.ToString().Trim());
		}
	}
}