﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.Linq.Expressions;
using System.Linq;
using System.Reflection;
using Ice.Orm.Core.DataStore;
using System.IO;
using Ice.Orm.Core.LINQ;
using System.Collections;
using Ice.Orm.Core.Utility;
using Ice.Utility;
using Ice.Orm.Core.Schema;
using Ice.Orm.Core.Entity;
using Ice.Orm.Core.Criteria.SqlBlocks;
using System.Collections.ObjectModel;
using System.Data;

namespace Ice.Orm.Core.Criteria
{
	[System.Diagnostics.DebuggerDisplay("{QueryWithValues}")]
	internal class QueryBase : ICompilableQuery
	{
		#region memeber variables

		public event Action<string> WriteLog;
		public bool SimulationMode { get; set; }
		public bool WriteTime { get; set; }
		private List<ISqlBlock> SqlBlockItems = new List<ISqlBlock>();
		public DataStore.DataStoreBase DataStore { get; set; }
		public EntityCache CachePolicy { get; set; }
		public SchemaTableHelper SchemaTableHelper { get; set; }
		private List<QueryBase> SubQueries = new List<QueryBase>();

		#endregion

		#region query helpers

		private string QueryWithValues { get { return this.CompileQuery().QueryWithValues; } }
		private CompiledQuery CompiledQuery { get { return this.CompileQuery(); } }

		/// <summary>
		/// Get with create if don't exist
		/// </summary>
		/// <typeparam name="TBlock"></typeparam>
		/// <returns></returns>
		internal TBlock GetSqlBlock<TBlock>() where TBlock : ISqlBlock
		{
			ISqlBlock curr = this.SqlBlockItems.OfType<TBlock>().FirstOrDefault();
			if (curr == null)
			{
				curr = (ISqlBlock)Activator.CreateInstance<TBlock>();
				this.SqlBlockItems.Add(curr);
			}
			return (TBlock)curr;
		}

		public void AppendQuery(QueryBase query)
		{
			this.SubQueries.Add(query);
		}

		/// <summary>
		/// test if contains non-empty block, including test for SqlBlock which is empty if there is only Skip or Take set, but no sql names
		/// </summary>
		/// <param name="kind"></param>
		/// <returns></returns>
		internal bool ContainsValidBlock(BlockKindEnum kind)
		{
			return this.SelectQueries().SelectMany(o=>o.SqlBlockItems).Where(o => o.Kind == kind && !o.IsEmpty).Any(); // IsEmpty test is required for SqlSelect block where Skip or Take can be set, but there is no sql in it, so it is Empty block
		}

		#endregion

		#region query compiling

		/// <summary>
		/// get query and it's subqueries
		/// </summary>
		/// <returns></returns>
		private IEnumerable<QueryBase> SelectQueries()
		{
			yield return this;
			foreach (QueryBase qq in this.SubQueries)
			{
				yield return qq;
			}
		}

		/// <summary>
		/// select SqlSegments from current query and it's subqueries
		/// </summary>
		/// <returns></returns>
		public IEnumerable<SqlSegment> SelectSqlSegments()
		{
			bool addSeparator = false;

			foreach (QueryBase query in SelectQueries())
			{
				if (addSeparator)
				{
					yield return new SqlSegment("\r\n");
				}

				addSeparator = true;

				foreach (SqlSegment seg in query.SqlBlockItems.Where(o => !o.IsEmpty).OrderBy(o => o.Kind).SelectMany(o => o.SelectSqlSegments()))
				{
					yield return seg;
				}
			}
		}

		public CompiledQuery CompileQuery()
		{
			return new CompiledQuery(this);
		}

		#endregion

		#region WriteToSqlBlocks methods and properties

		/// <summary>
		/// Number of rows to skip from database
		/// </summary>
		internal int SkipCount
		{
			get { return this.GetSqlBlock<SqlSelectBlock>().Skip; }
			set { this.GetSqlBlock<SqlSelectBlock>().Skip = value; }
		}

		/// <summary>
		/// Max. number of rows to select from database
		/// </summary>
		internal int? TakeCount
		{
			get { return this.GetSqlBlock<SqlSelectBlock>().Take; }
			set { this.GetSqlBlock<SqlSelectBlock>().Take = value; }
		}
	
		/// <summary>
		/// if true, will create SELECT DISTINCT
		/// </summary>
		internal bool IsDistinct
		{
			get { return this.GetSqlBlock<SqlSelectBlock>().IsDistinct; }
			set { this.GetSqlBlock<SqlSelectBlock>().IsDistinct = value; }
		}

		public void WriteToCountBlock(SqlName fromTable)
		{
			this.GetSqlBlock<SqlSelectBlock>().AddSegment(new SqlSegment("COUNT(1)"));
			this.GetSqlBlock<SqlFromBlock>().AddTable(fromTable);
		}

		public void WriteToDeleteBlock(SqlName fromTable)
		{
			this.GetSqlBlock<SqlDeleteBlock>().AddTable(fromTable);
		}

		internal void WriteToFromBlock(SqlName tableName)
		{
			SqlFromBlock from = this.GetSqlBlock<SqlFromBlock>();
			from.AddTable(tableName);
		}

		internal void WriteToJoinBlock(SqlName tableName, SqlName onA, SqlName equalsB, JoinKindEnum joinKind)
		{
			SqlFromBlock from = this.GetSqlBlock<SqlFromBlock>();
			from.AddJoinedTable(joinKind, tableName, onA.ToFullSqlName() + " = " + equalsB.ToFullSqlName());
		}

		#endregion

		#region helper methods for query execution

		private void LogQuery(QueryBase query)
		{
			if (this.WriteLog != null)
			{
				CompiledQuery cc = this.CompileQuery();

				this.WriteLog(@"================================");
				this.WriteLog("\r\n");

				if (this.WriteTime)
				{
					this.WriteLog(DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss"));
					this.WriteLog("\r\n");
				}

				this.WriteLog(@"QueryText:

" + cc.QueryText + @"

QueryTextWithValues:

" + cc.QueryWithValues);

				this.WriteLog("\r\n");
				this.WriteLog("\r\n");

			}
		}

		#endregion

		#region query execution methods

		private void EnsureDataStore<TSource>()
		{
			Entity.EntitySetting esource = Entity.EntitySetting.Get(typeof(TSource));

			if (this.SchemaTableHelper == null)
			{
				this.SchemaTableHelper = esource.SchemaTableHelper;
			}
			if (this.DataStore == null)
			{
				this.DataStore = esource.DataStore;
			}
		}

		private void EnsureSelectAndDataStore<TSource>()
		{
			Entity.EntitySetting esource = Entity.EntitySetting.Get(typeof(TSource));

			if (this.SchemaTableHelper == null)
			{
				this.SchemaTableHelper = esource.SchemaTableHelper;
			}
			if (this.DataStore == null)
			{
				this.DataStore = esource.DataStore;
			}

			if (!this.ContainsValidBlock(BlockKindEnum.Select) && !this.ContainsValidBlock(BlockKindEnum.Command))
			{
				SqlName tableName = new SqlName(esource.SchemaTableHelper.TableName);
				this.GetSqlBlock<SqlSelectBlock>().AddSegment(new SqlSegment(tableName.ToFullSqlName() + ".*"));

				this.GetSqlBlock<SqlFromBlock>().AddTable(tableName);
			}
		}

		public int ExecuteNonQuery<TSource>()
		{
			this.EnsureDataStore<TSource>();

			LogQuery(this);

			if (this.SimulationMode)
			{
				return 0;
			}
			else
			{
				CompiledQuery cc = new CompiledQuery(this);
				return (this.DataStore).ExecuteNonQuery(cc.QueryText, cc.DbParameters);
			}
		}

		public TResult ExecuteScalar<TSource, TResult>()
		{
			this.EnsureDataStore<TSource>();

			LogQuery(this);

			if (this.SimulationMode)
			{
				return default(TResult);
			}
			else
			{
				CompiledQuery cc = new CompiledQuery(this);
				// sqlite returns int64 when counting records, can't just cast to int, must convert
				// if executing Sum(colName) where rows count = 0: ExecuteScalar returns NULL. Must be able to convert NULL to int without exception. The same is with Min(colName) and Max(colName).
				TResult res;
				IceTypeConverter.TryChangeType<TResult>((this.DataStore).ExecuteScalar(cc.QueryText, cc.DbParameters), out res);
				return res;
			}
		}

		static Ice.Caching.Cache<string, object> _cachedManager = new Caching.Cache<string, object>();

		public List<SimpleDataTable> ToSimpleDataSet<TSource>()
		{
			this.EnsureSelectAndDataStore<TSource>();

			return ToSimpleDataSet(typeof(TSource));
		}

		internal List<SimpleDataTable> ToSimpleDataSet(Type sourceType)
		{
			this.CachePolicy = this.CachePolicy ?? AttributedTableHelper.GetEntityCachePolicy(sourceType);

			CompiledQuery cc = this.CompileQuery();

			if (this.CachePolicy != null)
			{
				lock (_cachedManager)
				{
					string key = cc.QueryWithValues;
					object obj;
					if (!_cachedManager.TryGetValue(key, out obj))
					{
						obj = this.DataStore.GetDataSet(cc.QueryText, cc.DbParameters, cc.SkipWithReader);
						_cachedManager.SetValue(key, obj, TimeSpan.FromSeconds(this.CachePolicy.DurationSek), this.CachePolicy.IsSlidingExpiration);
					}
					return (List<SimpleDataTable>)obj;
				}
			}

			return this.DataStore.GetDataSet(cc.QueryText, cc.DbParameters, cc.SkipWithReader);
		}

		internal List<TResult> ToList<TSource, TResult>()
		{
			var ds = this.ToSimpleDataSet<TSource>();
			return ObjectMaterializer.Translate<TResult>(ds.First(), this.DataStore);
		}

		#endregion

		#region ICompilableQuery Members

		IEnumerable<SqlSegment> ICompilableQuery.SelectSqlSegments()
		{
			return this.SelectSqlSegments();
		}

		#endregion

		#region ICompilableQuery Members

		CompiledQuery ICompilableQuery.CompileQuery()
		{
			return this.CompileQuery();
		}

		#endregion
	}

}
