﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Data.Common;
using System.Data.Odbc;
using System.Linq;
using System.Text.RegularExpressions;

namespace Ice.Orm.Core.DataStore
{
	using Criteria;
	using Entity;
	using LINQ;
	using Schema;
	using Utility;
	using System.Collections.ObjectModel;

	public abstract class DataStoreBase
	{
		#region cached schema manager

		class CachedSchemaManager
		{
			class CachedShemaItem
			{
				public string Key { get; set; }
				public SimpleDataTable Schema { get; set; }
				public DateTime TimeUsed { get; set; }
				public uint UsedCount { get; set; }
			}

			int maxCachedSchemas = 10000;

			private object _rwLock = new object();

			public static bool CacheReaderSchema = true;

			Dictionary<string, CachedShemaItem> Items = new Dictionary<string, CachedShemaItem>();

			////static Regex mReSchemaCache = new Regex("[\\s]+TOP[\\s]+[\\d\\(\\)]+", RegexOptions.IgnoreCase);

			////public static string CreateSchemaKey(string sqlStatement)
			////{
			////    // replace SELECT TOP (10) with SELECT TOP X to have single schemaKey if the rest of sql is the same
			////    return mReSchemaCache.Replace(sqlStatement, ""); // remove top (n)
			////}

			public SimpleDataTable this[string key]
			{
				get
				{
					lock (_rwLock)
					{
						CachedShemaItem item;
						if (this.Items.TryGetValue(key, out item))
						{
							if (item.UsedCount < uint.MaxValue) // keep max value to avoid removal if items.Count > max items
							{
								item.UsedCount++;
							}
							item.TimeUsed = DateTime.Now;
							return item.Schema;
						}
					}
					return null;
				}
				set
				{
					lock (_rwLock)
					{
						// remove if requred first, than will add new item since new item will have UsedCount == 1
						if (this.Items.Count > maxCachedSchemas * 2)
						{
							// if items count is double, create new dictionary from the first maxCachedSchemas items
							this.Items = this.Items.Values
								.OrderByDescending(o => o.UsedCount) // take the most used
								.ThenByDescending(o => o.TimeUsed) // than take the most recent
								.Take(maxCachedSchemas)
								.ToDictionary(o => o.Key, o => o);
						}

						this.Items[key] = new CachedShemaItem() { TimeUsed = DateTime.Now, Schema = value, Key = key, UsedCount = 1 };
					}
				}
			}
		}

		#endregion
		
		#region schema table class

		[Serializable]
		public class Schemas
		{
			public DataTable SchemaTable { get; set; }
			public DataTable ReaderSchema { get; set; }
		}

		#endregion

		#region Private members

		// 1 DataStore = 1 database
		// 1 DataStore have connection pool with x connection contexts (1 connection context = 1 db connection)
		private ConnectionContextPool mConnectionsPool = new ConnectionContextPool();

		CachedSchemaManager mCachedSchemaManager = new CachedSchemaManager();

		private DataStoreSetting mDbSetting = null;

		/// <summary>
		/// for ODBC or non auto number fields
		/// </summary>
		private static object mRetrieveIdentityLock = new object();

		#endregion

		#region Constructors

		private DataStoreBase()
		{
		}

		protected DataStoreBase(DataStoreSetting dbSetting)
		{
			mDbSetting = dbSetting;
			if (dbSetting != null)
			{
				lock (mConnectionsPool)
				{
					mConnectionsPool.Initialize(dbSetting.ConnPoolSize);
				}
			}
		}

		#endregion

		#region Properties

		protected ConnectionContextPool ConnectionsPool
		{
			get { return mConnectionsPool; }
		}

		public DataStoreSetting DbSetting
		{
			get { return mDbSetting; }
		}

		/// <summary>
		/// <para>returns true if data store requres inner criteria name:</para>
		/// <para>MSSQL: select * from (select * from [Article]) as criteriaName</para>
		/// <para>ORACLE: select * from (select * from [Article])</para>
		/// </summary>
		/// <returns></returns>
		protected virtual bool CriteriaNameRequired
		{
			get { return true; }
		}

		/// <summary>
		/// Defines the parameter preffix to use in commands for this data store.
		/// </summary>
		public virtual string ParameterPreffix
		{
			get { return "@"; }
		}

		/// <summary>
		/// Defines the object preffix to use in queries for this data store.
		/// </summary>
		protected virtual string ColumnPreffix
		{
			get { return @"["; }
		}

		/// <summary>
		/// Defines the object suffix to use in queries for this data store.
		/// </summary>
		protected virtual string ColumnSuffix
		{
			get { return @"]"; }
		}

		/// <summary>
		/// Gets the separator for objects as tables and columns
		/// </summary>
		protected virtual char NameSeparator
		{
			get { return '.'; }
		}

		/// <summary>
		/// If true, use named parameters, else use ? instead
		/// </summary>
		/// <returns></returns>
		public virtual bool UseNamedParameters
		{
			get { return false; }
		}

		/// <summary>
		/// Query to execute after insert, used to retrieve identity value.
		/// </summary>
		/// <returns></returns>
		protected virtual string PostInsertSQL
		{
			get { return "SELECT @@IDENTITY"; }
		}

		/// <summary>
		/// Returns true if data store support 2 or more query statement at once. ODBC or OleDB support only 1 query! This is used for retriving identity after insert!
		/// </summary>
		/// <returns></returns>
		protected virtual bool SupportCompoundQuery
		{
			get { return false; }
		}

		/// <summary>
		/// Return default terminator ;
		/// </summary>
		/// <returns></returns>
		protected virtual string SqlStatementTerminator
		{
			get { return ";"; }
		}

		/// <summary>
		/// if false, create TOP n, if true create TOP (n)
		/// </summary>
		public virtual bool ParentherizeTop
		{
			get { return false; }
		}

		public virtual bool UseOffsetAndLimit
		{
			get { return false; }
		}

		#endregion

		#region Helper methods

		protected IDbConnection OpenConnection(ConnectionContext context)
		{
			if (context.DbConnection == null)
			{
				context.DbConnection = NewDbConnection();
				context.DbConnection.ConnectionString = mDbSetting.ConnectionString;
			}

			IDbConnection conn = context.DbConnection;
			if (conn.State != ConnectionState.Open)
			{
				conn.Open();
			}
			return conn;
		}

		protected void CloseConnection(ConnectionContext context)
		{
			if (context.DbConnection != null && !mDbSetting.KeepOpen && context.Transaction == null)
			{
				context.DbConnection.Close();
			}

			// release context
			mConnectionsPool.ReleaseConnContext(context);
		}

		protected virtual IDbConnection NewDbConnection()
		{
			return new OdbcConnection();
		}

		protected virtual IDbCommand NewDbCommand()
		{
			return new OdbcCommand();
		}

		protected virtual IDbDataAdapter NewDbDataAdapter()
		{
			return new OdbcDataAdapter();
		}

		/// <summary>
		/// Common settings. Returns true if schemaColumn != null, else false
		/// </summary>
		/// <param name="parameter"></param>
		/// <param name="schemaColumn"></param>
		/// <returns></returns>
		protected virtual bool FixDbParameterValueAndType(IDbDataParameter parameter, EntityFieldSetting schemaColumn)
		{
			// if null, must be System.DBNull.Value
			if (parameter.Value == null)
			{
				parameter.Value = System.DBNull.Value;
			}

			if (schemaColumn == null)
			{
				if (parameter.Value != null && parameter.Value != System.DBNull.Value)
				{
					// set parameter type based on Value type
					parameter.DbType = TypeConverters.GetDbType(parameter.Value.GetType());
				}
			}
			else // schemaColumn != null
			{
				// DateTime.MinValue is Unspecified, so comparison with Local kind is not correct, if year is < 0002, consider as MinValue
				if (schemaColumn.AllowDBNull && parameter.Value != null && parameter.Value is DateTime && ((DateTime)parameter.Value).Year < 2) // 0002 year
				{
					parameter.Value = System.DBNull.Value;
					return true;
				}

				// Ext compatibility
				if (parameter.Value as string == "" && schemaColumn.DataType != typeof(string)) // Ext compatibility!!! if data column is not string type, "" means null
				{
					if (schemaColumn.AllowDBNull)
					{
						parameter.Value = System.DBNull.Value;
					}
					else
					{
						// get default value
						parameter.Value = TypeConverters.GetDefaultNotNullNetValue(schemaColumn.DataType);
					}
				}

				// check MagicValue
				if (schemaColumn != null && schemaColumn.MagicValue != null && (parameter.Value + "") == (schemaColumn.MagicValue + ""))
				{
					parameter.Value = System.DBNull.Value;
				}

				// set parameter type based on schemaColumn type
				parameter.DbType = TypeConverters.GetDbType(schemaColumn.DataType);


				if (parameter.Value != null && parameter.Value != System.DBNull.Value)
				{
					Type t = parameter.Value.GetType();

					if (t != schemaColumn.DataType)
					{
						// conversion is required
						if (t == typeof(Boolean))
						{
							bool value = ((Boolean)parameter.Value);
							if (schemaColumn.DataType == typeof(string))
							{
								parameter.Value = value ? "1" : "0";
							}
							else
							{
								parameter.Value = value ? 1 : 0;
							}
						}
						else if (schemaColumn.DataType == typeof(DateTime))
						{
							parameter.Value = DateTime.ParseExact(parameter.Value + "", this.DbSetting.DateTimePattern, null);
						}
						else
						{
							parameter.Value = Convert.ChangeType(parameter.Value, schemaColumn.DataType);
						}
					}
				}

				// set column size for String and Byte[]
				if (schemaColumn.ColumnSize > 0 && parameter.Value != null && parameter.Value != System.DBNull.Value)
				{
					parameter.Size = schemaColumn.ColumnSize;

					if (parameter.Value is string)
					{
						int len = ((String)parameter.Value).Length;
						if (len > schemaColumn.ColumnSize)
						{
							// ODBC hack: If string is longer than column size, odbc exception is thrown. We have to trim string to the exact column size.
							// this is requred for ODBC only, but let's be safe

							if (this.DbSetting.TruncateToFit)
							{
								parameter.Value = ((String)parameter.Value).Substring(0, parameter.Size);
							}
							else
							{
								

								// we may throw exception here that string is to long
								throw new Exception("String value is too long for column: " + schemaColumn.TableAndColumnName + ". Database column size: " + schemaColumn.ColumnSize + ", string length: " + len + ", text: " + parameter.Value);
							}
						}
						// trimm string if it's value > ColumnSize
					}
					else if (parameter.Value is byte[])
					{
						int len = ((byte[])parameter.Value).Length;
						if (len > schemaColumn.ColumnSize)
						{
							throw new Exception("byte[] data is too long for column: " + schemaColumn.TableAndColumnName + ". Database column size: " + schemaColumn.ColumnSize + ", data length: " + len);
						}
					}
				}
			}
			// return true if schemaColumn is not null
			return schemaColumn != null;
		}

		internal ReadOnlyCollection<IDbDataParameter> ConvertParameters(ReadOnlyCollection<QParameter> parameters, SchemaTableHelper schema)
		{
			List<IDbDataParameter> list = new List<IDbDataParameter>();
			if (parameters == null || parameters.Count == 0)
			{
				return list.AsReadOnly();
			}
			foreach (QParameter par in parameters)
			{
				EntityFieldSetting es = schema == null || string.IsNullOrEmpty(par.Column) ? null : schema.SchemaColumns[par.Column];
				IDbDataParameter dbpar = this.CreateIDbDataParameter(par.Name, par.Value, es);
				list.Add(dbpar);
			}
			return list.AsReadOnly();
		}

		protected virtual IDbDataParameter CreateIDbDataParameter(string parameterName, object value, EntityFieldSetting schemaColumn)
		{
			throw new NotImplementedException("method has to be implemented in child class");
		}

		protected void AddParametersToDbCommand(IDbCommand cmd, ReadOnlyCollection<IDbDataParameter> parameters)
		{
			if (parameters != null && parameters.Count > 0)
			{
				foreach (IDbDataParameter par in parameters)
				{
					cmd.Parameters.Add(par);
				}
			}
		}

		internal virtual string CleanUpSqlText(string sqlStatement)
		{
			return sqlStatement;
		}

		private string DumpParameters(ReadOnlyCollection<IDbDataParameter> col)
		{
			try
			{
				return col.Cast<IDbDataParameter>().Select(o => "declare " + o.ParameterName + " " + SqlTypesVsNetTypes.FindSQLType(o.DbType) + " = " + o.Value).ToList().ToStringJoined("\r\n");
			}
			catch
			{
			}
			return null;
		}

		/// <summary>
		/// If query is not made by CompiledQuery, it has to be parsed and replace parameters names @pX with ?
		/// </summary>
		/// <param name="sql"></param>
		/// <param name="parameters"></param>
		/// <returns></returns>
		private string CreateUnnamedParameters(string sql, ReadOnlyCollection<IDbDataParameter> parameters)
		{
			if (this.UseNamedParameters || parameters == null || parameters.Count == 0)
			{
				return sql;
			}

			Dictionary<string, IDbDataParameter> dic = parameters.ToDictionary(o => o.ParameterName, o => o);
			Dictionary<string, IDbDataParameter> finalDic = new Dictionary<string, IDbDataParameter>();

			// arrange parameters in the correct order
			string parsed = new Regex("@p[\\d]+").Replace(sql, (m) =>
			{
				finalDic.Add(m.Value, dic[m.Value]); // will not allow to add the same parameter twice
				return "?";
			});
			parameters = dic.Values.ToList().AsReadOnly();
			return parsed;
		}

		#endregion

		#region BuildGDataTableForDbReader

		private SimpleDataTable BuildSchemaGDataTable(DataTable readerSchema)
		{
			return SimpleDataTable.Create(BuildDataTableForDbReaderInternal(readerSchema, true));
		}

		private DataTable BuildDataTableForDbReaderInternal(DataTable readerSchema, bool columnDetails)
		{
			////readerSchema.WriteXml("d:\\reader_" + DateTime.Now.Ticks + ".xml");

			int count = 0;
			DataTable dt = new DataTable();
			foreach (DataRow drschema in readerSchema.Rows)
			{
				string columnName = SqlName.CreateShortName((String)drschema["ColumnName"]); // OleDb uses Product.ProductID naming!!! We want ProductID only!
				DataColumn column = new DataColumn(columnName, (Type)drschema["DataType"]);
				if (dt.Columns.Cast<DataColumn>().Where(o=>o.ColumnName.Equals(column.ColumnName)).Any()) // join may select duplicated columns
				{
					throw new NotSupportedException("duplicated columns in sql select: " + columnName);
					////column.ColumnName += count;
				}

				if (columnDetails)
				{
					column.AllowDBNull = (Boolean)drschema["AllowDBNull"];
					column.AutoIncrement = (Boolean)drschema["IsAutoIncrement"];
					column.ReadOnly = (Boolean)drschema["IsReadOnly"]; // this is not correct value
					column.Unique = (Boolean)drschema["IsUnique"];  // this is not correct value
					////if (column.DataType == typeof(string))
					////{
					////    // sometimes non string types are read as string. setting MaxLength for non string type, throws exception
					////    column.MaxLength = (int)drschema["ColumnSize"]; // for string types only, this throws exception for non string types (like byte[])
					////}
				}
				count++;
				dt.Columns.Add(column);
			}
			return dt;
		}

		////private GDataTable BuildGDataTableForDbReader(DataTable readerSchema, bool columnDetails)
		////{
		////    int count = 0;
		////    List<string> columns = new List<string>();
		////    foreach (DataRow drschema in readerSchema.Rows)
		////    {
		////        string columnName = SqlName.CreateShortName((String)drschema["ColumnName"]); // OleDb uses Product.ProductID naming!!! We want ProductID only!

		////        if (columns.Where(o=>o.Equals(columnName, StringComparison.OrdinalIgnoreCase)).Any()) // join may select duplicated columns
		////        {
		////            throw new NotSupportedException("duplicated columns selected: " + columnName);
		////            //columnName += count;
		////        }

		////        columns.Add(columnName);

		////        count++;
		////    }

		////    return new GDataTable(columns.AsReadOnly());
		////}

		#endregion

		#region GetSchemaTable & GetTablesNames from database

		protected virtual string GetSelectSQLForSchema(string dbTableName)
		{
			return "SELECT TOP 1 * FROM " + SqlName.CreateFullSqlName(dbTableName) + " WHERE 1<>1";
		}

		/// <summary>
		/// Try to load from db with RetryIfFail retrys
		/// </summary>
		/// <param name="tableName"></param>
		/// <returns></returns>
		public virtual Schemas GetSchemaTable(string tableName)
		{
			for (int n = 0; ; n++)
			{ 
				DbDataAdapter adapter = (DbDataAdapter)NewDbDataAdapter();
				IDbCommand cmd = NewDbCommand();
				cmd.CommandText = GetSelectSQLForSchema(tableName); // this select nothing! DataReader loads too many data on close if select * from... is valid!
				IDataReader reader = null;

				// critical section start
				// get next available context, or wait for one to become available
				ConnectionContext connContext = mConnectionsPool.GetConnContext(); // thread locks here if connectionContext pool is bussy

				try
				{
					DataTable dt = new DataTable(tableName);
					cmd.Connection = OpenConnection(connContext);
					cmd.Transaction = connContext.Transaction;
					adapter.SelectCommand = (DbCommand)cmd;
					adapter.FillSchema(dt, SchemaType.Mapped);

					Schemas sss = new Schemas() { SchemaTable = dt };

					try
					{
						reader = cmd.ExecuteReader();
						sss.ReaderSchema = reader.GetSchemaTable();
						reader.Close();
					}
					catch
					{
						if (reader != null && !reader.IsClosed)
						{
							reader.Close();
						}
					}

					cmd.Parameters.Clear();
					// release context object
					CloseConnection(connContext);
					return sss;
				}
				catch (Exception ex)
				{
					cmd.Parameters.Clear();
					// release context object
					CloseConnection(connContext);

					if (n >= mDbSetting.RetrysIfFails)
					{
						throw ex;
					}
				}
			}
		}

		/// <summary>
		/// Try to load from db with RetryIfFail retrys
		/// </summary>
		/// <param name="tableName"></param>
		/// <returns></returns>
		protected virtual DataTable LoadTablesNamesFromDB()
		{
			for (int n = 0; ; n++)
			{
				// critical section start
				// get next free db connection or wait for one to become available
				ConnectionContext connContext = mConnectionsPool.GetConnContext(); // thread locks here if connectionContext pool is bussy

				try
				{
					DbConnection conn = (DbConnection)OpenConnection(connContext);
					//// try select * from INFORMATION_SCHEMA.COLUMNS to examine this 
					//// conn.GetSchema("Columns", new string[] { null, null, "State" }); // null = not important, table name State is 3rd column in INFORMATION_SCHEMA.COLUMNS
					DataTable dt = conn.GetSchema("Tables"); // works with all datastore providers
					CloseConnection(connContext);
					return dt;
				}
				catch (Exception ex)
				{
					CloseConnection(connContext);
					if (n >= mDbSetting.RetrysIfFails)
					{
						throw ex;
					}
				}
			}
		}

		public virtual List<string> GetTablesNames()
		{
			#region schema xml examples

			/* sql 2008:
			<Tables>
			<TABLE_CATALOG>GSMS</TABLE_CATALOG>
			<TABLE_SCHEMA>dbo</TABLE_SCHEMA>
			<TABLE_NAME>WapBrowserSetting</TABLE_NAME>
			<TABLE_TYPE>BASE TABLE</TABLE_TYPE>
		  </Tables>
		  <Tables>
			<TABLE_CATALOG>GSMS</TABLE_CATALOG>
			<TABLE_SCHEMA>dbo</TABLE_SCHEMA>
			<TABLE_NAME>sysdiagrams</TABLE_NAME>
			<TABLE_TYPE>BASE TABLE</TABLE_TYPE>
		  </Tables>
			 *
			 * 
			 * sqlce:
			 <Tables>
				<TABLE_NAME>BillingStatus</TABLE_NAME>
				<TABLE_TYPE>TABLE</TABLE_TYPE>
			  </Tables>
			 * 
			 * 
			 * sqlite:
		 <Tables>
			<TABLE_CATALOG>main</TABLE_CATALOG>
			<TABLE_NAME>bluetoothdelivery</TABLE_NAME>
			<TABLE_TYPE>table</TABLE_TYPE>
			<TABLE_ID>3</TABLE_ID>
			<TABLE_ROOTPAGE>5</TABLE_ROOTPAGE>
			<TABLE_DEFINITION>CREATE TABLE "bluetoothdelivery" ...</TABLE_DEFINITION>
		  </Tables>
			 * 
			 * 
			 * oledb:
			  <Tables>
				<TABLE_NAME>Address</TABLE_NAME>
				<TABLE_TYPE>TABLE</TABLE_TYPE>
				<DATE_CREATED>2011-09-23T21:45:35.986+02:00</DATE_CREATED>
				<DATE_MODIFIED>2012-03-19T15:13:38.515+01:00</DATE_MODIFIED>
			  </Tables>
			   <Tables>
				<TABLE_NAME>MSysAccessStorage</TABLE_NAME>
				<TABLE_TYPE>ACCESS TABLE</TABLE_TYPE>
				<DATE_CREATED>2012-03-19T15:13:38.531+01:00</DATE_CREATED>
				<DATE_MODIFIED>2012-03-19T15:13:38.531+01:00</DATE_MODIFIED>
			  </Tables>
			  <Tables>
				<TABLE_NAME>MSysACEs</TABLE_NAME>
				<TABLE_TYPE>SYSTEM TABLE</TABLE_TYPE>
				<DATE_CREATED>2011-09-23T21:45:14.046+02:00</DATE_CREATED>
				<DATE_MODIFIED>2011-09-23T21:45:14.046+02:00</DATE_MODIFIED>
			  </Tables>
			 * 
			 * 
			 * odbc:
			 <Tables>
				<TABLE_CAT>D:\=VSPROJECTS=\SQLDB\VISTACART</TABLE_CAT>
				<TABLE_NAME>Address</TABLE_NAME>
				<TABLE_TYPE>TABLE</TABLE_TYPE>
			  </Tables>
			  <Tables>
				<TABLE_CAT>D:\=VSPROJECTS=\SQLDB\VISTACART</TABLE_CAT>
				<TABLE_NAME>MSysAccessStorage</TABLE_NAME>
				<TABLE_TYPE>SYSTEM TABLE</TABLE_TYPE>
			  </Tables>
		  */

			#endregion

			DataTable dt = LoadTablesNamesFromDB();

			// Oracle has column name TYPE, all other has TABLE_TYPE
			DataColumn typecolumn = dt.Columns.Cast<DataColumn>().Where(o=>o.ColumnName.IndexOf("TYPE", StringComparison.OrdinalIgnoreCase) >= 0).First();

			// MSSQL type: BASE TABLE
			// ODBC, OLEDB type: TABLE
			// ORACLE type: User
			// SQLITE type: table
			var items = dt.Rows.Cast<DataRow>()
				.Select(o => new { Name = o["TABLE_NAME"] + "", Type = o[typecolumn.ColumnName] + "" })
				.ToList();

			var filtered = items.Where(o => "TABLE, BASE TABLE, USER".IndexOf(o.Type, StringComparison.OrdinalIgnoreCase) >= 0)
				.Where(o=>o.Name != "sysdiagrams") // exclude ms sql system table
				.Select(o => o.Name)
				.OrderBy(o=>o)
				.ToList();

			return filtered;
		}

		#endregion

		#region GetDataTable

		protected virtual SimpleDataTable GetDataTable(string sqlStatement, ReadOnlyCollection<IDbDataParameter> dbParameters = null, int skipWithReader = 0)
		{
			return GetDataSet(sqlStatement, dbParameters, skipWithReader).FirstOrDefault();
		}

		public virtual List<SimpleDataTable> GetDataSet(string sqlStatement, ReadOnlyCollection<IDbDataParameter> dbParameters = null, int skipWithReader = 0)
		{
			for (int n = 0; ; n++)
			{
				sqlStatement = CreateUnnamedParameters(sqlStatement, dbParameters);
				IDbCommand cmd = NewDbCommand();
				cmd.CommandText = sqlStatement;
				AddParametersToDbCommand(cmd, dbParameters);
				IDataReader reader = null;
				List<SimpleDataTable> resultsList = new List<SimpleDataTable>();

				// critical section start
				// get next free db connection or wait for one to become available
				ConnectionContext connContext = mConnectionsPool.GetConnContext(); // thread locks here if connectionContext pool is bussy

				try
				{
					cmd.Connection = OpenConnection(connContext);
					cmd.Transaction = connContext.Transaction;
					reader = cmd.ExecuteReader();
					int resultIndex = -1; // if there are more than 1 table returned

					do
					{
						resultIndex++;

						// create GDataTable or get it from cache
						// BE CAREFUL: if using cached version, and db columns names or columns order is changed, invalid data will be read (acctuall data is associated with wrong column)
						// if using cached version, whole object loading is 3x faster!!!! (non-cached: loading 10.000 objects takes 7.1 s, cached: loading 10.000 objects takes 2.2 s)
						SimpleDataTable gdt = null;

						if (CachedSchemaManager.CacheReaderSchema)
						{
							////string readerSchemaKey = "[" + resultIndex + "]" + CachedSchemaManager.CreateSchemaKey(sqlStatement);
							string readerSchemaKey = "[" + resultIndex + "]" + sqlStatement;

							gdt = mCachedSchemaManager[readerSchemaKey];

							if (gdt == null || gdt.GetColumnsCount() != reader.FieldCount) // if columns count in database has changed, but we can't check if column order or name is changed!!!!
							{
								DataTable readerSchemaTable = reader.GetSchemaTable(); // very time consuming, don't keep mGDataTables locked for other threads!, 95% of time consumed is here
								gdt = BuildSchemaGDataTable(readerSchemaTable);

								mCachedSchemaManager[readerSchemaKey] = gdt;
							}
							gdt = gdt.Clone();
						}
						else // don't use cached reader schema!
						{
							DataTable readerSchemaTable = reader.GetSchemaTable(); // very time consuming, don't keep mGDataTables locked for other threads!, 95% of time consumed is here
							gdt = BuildSchemaGDataTable(readerSchemaTable);
						}

						resultsList.Add(gdt);

						int currentRecordIndex = -1;
						////int finalRecordIndex = -1; // if it is -1, there is no limitation, take all available records
						////if (take != null && take >= 0)// new: maxRecords <= 0, load all available
						////{
						////    finalRecordIndex = skipWithReader + take.GetValueOrDefault() - 1;
						////}

						int fieldsCount = reader.FieldCount;

						while (reader.Read())
						{
							currentRecordIndex++;

							if (currentRecordIndex < skipWithReader)
							{
								continue;
							}

							object[] data = new Object[fieldsCount];
							reader.GetValues(data);
							gdt.Rows.Add(data);

							// if maxRecords < 0, will load all available records by SQL criteria
							// new: maxRecords <= 0, load all available
							////if (finalRecordIndex >= 0 && currentRecordIndex >= finalRecordIndex)
							////{
							////    break;
							////}
						}
					}
					while (reader != null && !reader.IsClosed && reader.NextResult());

					reader.Close();
					cmd.Parameters.Clear();
					CloseConnection(connContext);
					return resultsList;

				}
				catch (Exception ex)
				{
					if (reader != null && reader.IsClosed != true)
					{
						reader.Close();
					}

					cmd.Parameters.Clear();
					CloseConnection(connContext);
					if (n >= mDbSetting.RetrysIfFails)
					{
						throw new Exception("\r\n" + DumpParameters(dbParameters) + "\r\n" + sqlStatement + "\r\n", ex);
					}
					else
					{
						// release parameter to be reused
						cmd.Parameters.Clear();
					}
				}
			}
		}

		#endregion

		#region ExecuteNonQuery

		public virtual int ExecuteNonQuery(string sqlStatement, ReadOnlyCollection<IDbDataParameter> dbParameters = null)
		{
			for (int n = 0; ; n++)
			{
				sqlStatement = CreateUnnamedParameters(sqlStatement, dbParameters);
				IDbCommand cmd = NewDbCommand();
				cmd.CommandText = sqlStatement;
				AddParametersToDbCommand(cmd, dbParameters);

				// critical section start
				ConnectionContext connContext = mConnectionsPool.GetConnContext(); // thread locks here if connectionContext pool is bussy

				try
				{
					cmd.Connection = OpenConnection(connContext);
					cmd.Transaction = connContext.Transaction;
					int rows = cmd.ExecuteNonQuery();
					cmd.Parameters.Clear();
					CloseConnection(connContext);

					return rows;
				}
				catch (Exception ex)
				{
					cmd.Parameters.Clear();
					CloseConnection(connContext);
					if (n >= mDbSetting.RetrysIfFails)
					{
						throw new Exception("\r\n" + DumpParameters(dbParameters) + "\r\n" + sqlStatement + "\r\n", ex);
					}
					else
					{
						// release parameter to be reused
						cmd.Parameters.Clear();
					}
				}
			}
		}

		#endregion

		#region  ExecuteScalar

		public virtual Object ExecuteScalar(string sqlStatement, ReadOnlyCollection<IDbDataParameter> dbParameters = null)
		{
			for (int n = 0; ; n++)
			{
				sqlStatement = CreateUnnamedParameters(sqlStatement, dbParameters);
				IDbCommand cmd = NewDbCommand();
				cmd.CommandText = sqlStatement;
				AddParametersToDbCommand(cmd, dbParameters);

				// critical section start
				ConnectionContext connContext = mConnectionsPool.GetConnContext(); // thread locks here if connectionContext pool is bussy

				try
				{
					cmd.Connection = OpenConnection(connContext);
					cmd.Transaction = connContext.Transaction;
					Object obj = cmd.ExecuteScalar();
					cmd.Parameters.Clear();
					CloseConnection(connContext);
					return obj;
				}
				catch (Exception ex)
				{
					cmd.Parameters.Clear();
					CloseConnection(connContext);
					if (n >= mDbSetting.RetrysIfFails)
					{
						throw new Exception("\r\n" + DumpParameters(dbParameters) + "\r\n" + sqlStatement + "\r\n", ex);
					}
					else
					{
						// release parameter to be reused
						cmd.Parameters.Clear();
					}
				}
			}
		}

		#endregion

		#region Save methods

		/// <summary>
		/// <para>Save newValues to database.</para>
		/// </summary>
		/// <param name="schemaTableHelper"></param>
		/// <param name="dbValues">if null, Insert wil be invoked, else Update will be invoked</param>
		/// <param name="newValues">values to write to database</param>
		/// <returns>Gdictionary with new values (including new autonumber value)</returns>
		public virtual int Save(SchemaTableHelper schemaTableHelper, SimpleDataRow dbValues, CaselessDictionary newValues)
		{
			if (newValues != null)
			{
				// update
				if (dbValues != null)
				{
					// update object previously loaded from database
					return Update(schemaTableHelper, dbValues, newValues);
				}
				else if (dbValues == null)
				{
					// insert new object with all default values
					return Insert(schemaTableHelper, newValues);
				}
			}
			return 0; // nothing to update
		}

		#endregion

		#region Insert methods

		private int RetrieveIdentityBeforeInsert(SchemaTableHelper schemaTableHelper, EntityFieldSetting uniqueIdentityColumn)
		{
			if (uniqueIdentityColumn == null)
			{
				throw new Exception("Insert exception. uniqueIdentityColumn not defined, can't retrieve it's value from database.");
			}
			string sqlsel = "SELECT MAX(" + SqlName.CreateFullSqlName(uniqueIdentityColumn.ColumnName) + ") FROM " + SqlName.CreateFullSqlName(schemaTableHelper.TableName);
			object oValue = this.ExecuteScalar(sqlsel); // returns int, Oracle returns decimal
			int value;
			int.TryParse(oValue + "", out value);

			if (!(value > 0))
			{
				value = 1;
			}
			else
			{
				value++;
			}
			return value;
		}

		private int InsertExecutor(SchemaTableHelper schemaTableHelper, CaselessDictionary newValues, string sql, ReadOnlyCollection<IDbDataParameter> dbParameters, string postInsertSql)
		{
			if (postInsertSql == null)
			{
				// identity is not autonumber, nothing to retrun after insert
				return ExecuteNonQuery(sql, dbParameters);
			}
			else // postInsertSql != null,  we have to retrieve identity after insert
			{
				if (SupportCompoundQuery)
				{
					// MSSQL, MYSQL, SQLITE...
					// insert than retrieve identity at once
					sql = sql + SqlStatementTerminator + postInsertSql + SqlStatementTerminator;
					object identity = ExecuteScalar(sql, dbParameters);
					int id = Convert.ToInt32(identity);
					if (id > 0)
					{
						newValues[schemaTableHelper.IdentityColumn.ColumnName] = id;
						return 1;
					}
				}
				else
				{
					// ODBC, OLEDB
					// insert, than retrieve identity in new query
					int rows = this.ExecuteNonQuery(sql, dbParameters);
					if (rows > 0)
					{
						Object obj = this.ExecuteScalar(postInsertSql, null); // get new autonumber value
						newValues[schemaTableHelper.IdentityColumn.ColumnName] = Convert.ToInt32(obj);
						return rows;
					}
				}
			}

			return 0; // nothing inserted
		}

		private int InsertCriticalSection(SchemaTableHelper schemaTableHelper, CaselessDictionary newValues, string postInsertSql, EntityFieldSetting uniqueIdentityColumn, bool retrieveIdentity)
		{
			QParameterCollection parameters = new QParameterCollection();

			if (retrieveIdentity)
			{
				object value = RetrieveIdentityBeforeInsert(schemaTableHelper, uniqueIdentityColumn);
				// insert or set into newValues
				newValues[uniqueIdentityColumn.ColumnName] = value;
			}

			EntityFieldSetting dummyColumn = null;
			bool identityInsert = false;

			// get all values from mNewValues and insert it into database
			foreach (EntityFieldSetting schemaColumn in schemaTableHelper.SchemaColumns)
			{
				Object value;
				bool isValueSet = newValues.TryGetValue(schemaColumn.ColumnName, out value);

				// for VistaCart, form values always have "" value, adjust it's type:
				if(value is string && (string)value == "" && schemaColumn.DataType != typeof(string))
				{
					value = null;
					isValueSet = false;
				}

				if (schemaColumn.IsAutoIncrement)
				{
					if (isValueSet && ((int)value) > 0)
					{
						// identity insert, only if value > 0
						identityInsert = true;
						parameters.Add(new QParameter(value, schemaColumn.ColumnName));
					}
				}
				else if (schemaColumn.IsReadOnly)
				{
					// can't write it to database
				}
				else if (schemaColumn.AllowDBNull)
				{
					if (!isValueSet) // is value set in newValues, value in newValues may be NULL and will be written NULL to database
					{
						// get default value
						value = schemaColumn.DefaultValue;
					}

					if (value != null && value != System.DBNull.Value)
					{
						// write to db
						newValues[schemaColumn.ColumnName] = value;
						parameters.Add(new QParameter(value, schemaColumn.ColumnName));
					}
					else if (dummyColumn == null)
					{
						// remember column, we have to write at least 1 value to database for compatibility, 
						// eg. if all fields are null, it would create INSERT INTO table () VALUES (). This throws exception.
						// we'll insert null
						dummyColumn = schemaColumn;
					}
				}
				else // don't allow db null
				{
					if (!isValueSet || value == null || value == System.DBNull.Value) // test if value is null too
					{
						// get default value
						value = schemaColumn.DefaultValue;
					}

					if (value == null || value == System.DBNull.Value)
					{
						value = TypeConverters.GetDefaultNotNullNetValue(schemaColumn.DataType);
						//throw new Exception("Can't write null to " + schemaColumn.ColumnName);
						// set new value
					}

					newValues[schemaColumn.ColumnName] = value;
					parameters.Add(new QParameter(value, schemaColumn.ColumnName));
				}
			}

			if (parameters.Count == 0)
			{
				if (dummyColumn != null)
				{
					// nothing to insert, we have to insert at least 1 value to avoid  INSERT INTO table () VALUES (). This throws exception.
					parameters.Add(new QParameter(dummyColumn.ColumnName, null));
				}
				else
				{
					throw new Exception("Impossible situation in Insert, seems data table has no columns!");
				}
			}

			// write to db
			string sql = parameters.GetInsertStatement(schemaTableHelper.TableName);
			if (identityInsert)
			{
				sql = "SET IDENTITY_INSERT [" + schemaTableHelper.TableName + "] ON;\r\n" + sql + ";\r\nSET IDENTITY_INSERT [" + schemaTableHelper.TableName + "] OFF;";
			}

			ReadOnlyCollection<IDbDataParameter> dbParameters = ConvertParameters(parameters.AsReadOnly(), schemaTableHelper);

			return InsertExecutor(schemaTableHelper, newValues, sql, dbParameters, postInsertSql);
		}

		/// <summary>
		/// Returns rows affected count, fills newValues with data saved to database
		/// </summary>
		/// <param name="schemaTableHelper"></param>
		/// <param name="dbValues"></param>
		/// <param name="newValues"></param>
		/// <returns></returns>
		public virtual int Insert(SchemaTableHelper schemaTableHelper, CaselessDictionary newValues)
		{
			if (mDbSetting.DisableWrite)
			{
				return 1;
			}

			// insert newValues to database
			if (newValues == null)
			{
				throw new Exception("newValues can't be null!");
			}

			string postInsertSql = null;
			EntityFieldSetting identityColumn = schemaTableHelper.IdentityColumn;
			bool retrieveIdentity = false;

			if (identityColumn == null)
			{
				// no identity column
			}
			else if (identityColumn.IsAutoIncrement)
			{
				// autoincrement column exist, we have to return it's value after insert
				postInsertSql = PostInsertSQL;
			}
			else // identityColumn is not autoincrement
			{
				if (newValues == null || !newValues.ContainsKey(identityColumn.ColumnName)) // if pkValue is not set in newValues hashtable
				{
					// try to set identity value, but it's better if value is already set by the user!

					if (identityColumn.DataType == typeof(int) || identityColumn.DataType == typeof(long) || identityColumn.DataType == typeof(decimal)) // oracle uses decimal
					{
						retrieveIdentity = true; // retrieve max value from db and increment it by 1
					}
					else if (identityColumn.DataType == typeof(Object))
					{
						retrieveIdentity = true; // sqLite is object for autonumber or number PK
					}
					else if (identityColumn.DataType == typeof(string))
					{
						// insert into newValues
						newValues[identityColumn.ColumnName] = Guid.NewGuid().ToString();
					}
					else
					{
						throw new Exception("Can't calculate primary key value, it has to be set manually.");
					}
				}
			}

			// critical section
			if (retrieveIdentity)
			{
				lock (mRetrieveIdentityLock) // DON'T ALLOW MORE THAN 1 THREAD TO ENTER HERE IF HAVE TO RETIEVE IDENTITY FIRST
				{
					return InsertCriticalSection(schemaTableHelper, newValues, postInsertSql, identityColumn, retrieveIdentity);
				}
			}
			else
			{
				return InsertCriticalSection(schemaTableHelper, newValues, postInsertSql, identityColumn, retrieveIdentity);
			}
		}

		#endregion

		#region Where sql for Update & Delete

		/// <summary>
		/// Create Where statement for Update or Delete. Used values are removed from newValues and placed into usedValues
		/// </summary>
		/// <param name="schemaTableHelper"></param>
		/// <param name="dbValues"></param>
		/// <param name="notForWriting"></param>
		/// <param name="tmpValues"></param>
		/// <returns></returns>
		private DBWhereKeys GetWhereStatementForWrite(SchemaTableHelper schemaTableHelper, SimpleDataRow dbValues, CaselessDictionary valuesToWrite)
		{
			// create where statement
			DBWhereKeys whereKeys = new DBWhereKeys();

			// use this AutoIncrementColumn value as reference
			SchemaColumnCollection referenceColumns = schemaTableHelper.WriteReference;

			//////if (referenceColumns == null || referenceColumns.Count == 0)
			//////{
			//////    referenceColumns = schemaTableHelper.SchemaColumns; // use all old row values for where statement, works with Delete only, for Update all columns will be removed from newValues
			//////}

			// build where statement
			if (referenceColumns == null || referenceColumns.Count == 0) // referenceColumns is null or referenceColumns.Count = 0
			{
				// OracleDataStore uses Oracle's ROWID
				//	ROWID is selected from DB every time an object is selected from db
				whereKeys = GetDataStoreSpecificWhere(valuesToWrite, dbValues);
			}
			else
			{
				foreach (EntityFieldSetting schemaColumn in referenceColumns)
				{
					// get from newValues first and remove it from newValues, value won't be updated in database, it's reference only
					Object value = null;

					if (dbValues != null)
					{
						value = dbValues[schemaColumn.ColumnName];
					}

					// write or delete reference should not be null

					// valuesToWrite can be null if Entity is loaded from db and invoked .DeleteFromDB()
					if (valuesToWrite != null)
					{
						if (value == null && valuesToWrite.TryGetValue(schemaColumn.ColumnName, out value))
						{
							valuesToWrite.Remove(schemaColumn.ColumnName); // REMOVE FROM valuesToWrite
						}
						else if (schemaColumn.IsAutoIncrement)
						{
							valuesToWrite.Remove(schemaColumn.ColumnName); // REMOVE FROM valuesToWrite
						}
					}

					whereKeys.Add(new DBWhereKey(schemaColumn.ColumnName, value));
				}
			}

			if (whereKeys == null || whereKeys.Count == 0)
			{
				throw new Exception("Can't create where condition! No WriteReference columns found in table " + schemaTableHelper.TableName);
			}

			return whereKeys; // required for update & delete statement
		}

		protected virtual DBWhereKeys GetDataStoreSpecificWhere(CaselessDictionary newValues, SimpleDataRow dbValues)
		{
			return null;
		}

		#endregion

		#region Update methods

		/// <summary>
		/// Returns Gdictionary with new values, if there is nothing to update returns null
		/// </summary>
		public virtual int Update(SchemaTableHelper schemaTableHelper, SimpleDataRow dbValues, CaselessDictionary newValues)
		{
			if (mDbSetting.DisableWrite)
			{
				return 1;
			}

			if (newValues == null || newValues.Count == 0)
			{
				// nothing to update
				return 0;
			}

			CaselessDictionary valuesToWrite = newValues.Clone();

			// create where statement
			// GetWhereStatementForWrite(...) removes values from valuesToWrite which can't be written, eg. PK, Autoincrement, ReadOnly
			DBWhereKeys whereKeys = GetWhereStatementForWrite(schemaTableHelper, dbValues, valuesToWrite);
			// IMPORTANT: valuesToWrite now contains only values to write to db, reference columns are removed from valuesToWrite

			// specific, for update only
			if (valuesToWrite == null || valuesToWrite.Count == 0)
			{
				// nothing to update, newValues had AutoIncrement field only or if no Autoincrement field, than PK only
				return 0; // nothing updated
			}

			// create parameters to update
			QParameterCollection parameters = new QParameterCollection();

			foreach (KeyValuePair<string, object> entry in valuesToWrite)
			{
				if (schemaTableHelper.SchemaColumns.Contains(entry.Key)) // newValues may contain keys which doesn't belong to database table
				{
					parameters.Add(new QParameter(entry.Value, entry.Key));
				}
			}

			// be careful, this is the correct order:

			// 1) get update sql first
			string updateSql = parameters.GetUpdateStatement(schemaTableHelper.TableName);

			// 2) merge where parameters with parameters to be updated, to make sure where parameters names are unique
			foreach (DBWhereKey whereKey in whereKeys.AsReadOnly())
			{
				whereKey.Name = null; // force to create new unique parameter name
				parameters.Add(whereKey); // makes ParameterName unique in sql
			}

			// 3) get where sql with unique parameters names
			string whereSql = whereKeys.GetWhereStatement();

			// 4) get all parameters (values to update and where values)
			ReadOnlyCollection<IDbDataParameter> dbParameters = ConvertParameters(parameters.AsReadOnly(), schemaTableHelper);

			return this.ExecuteNonQuery(updateSql + whereSql, dbParameters);
		}

		#endregion

		#region Delete methods

		/// <summary>
		/// Delete an object from database. Returns true if deleted, else false.
		/// </summary>
		/// <param name="schemaTableHelper"></param>
		/// <param name="dbValues"></param>
		/// <param name="newValues"></param>
		/// <returns></returns>
		public virtual int Delete(SchemaTableHelper schemaTableHelper, SimpleDataRow dbValues, CaselessDictionary newValues)
		{
			if (mDbSetting.DisableWrite)
			{
				return 1;
			}

			if (dbValues == null && newValues == null)
			{
				return 0; // no reference for deletion
			}

			//GDictionary notForWriting = null; // dummy, not used

			DBWhereKeys whereKeys = GetWhereStatementForWrite(schemaTableHelper, dbValues, newValues);

			string deleteSql = "DELETE FROM " + SqlName.CreateFullSqlName(schemaTableHelper.TableName) + whereKeys.GetWhereStatement();

			ReadOnlyCollection<IDbDataParameter> dbParameters = ConvertParameters(whereKeys.AsReadOnly(), schemaTableHelper);

			return this.ExecuteNonQuery(deleteSql, dbParameters);
		}

		#endregion

		#region StartTransaction method

		/// <summary>
		/// Start transaction. All actions will be executed inside of the current transaction and using same db connection. Commit when done and return T.
		/// </summary>
		/// <param name="action">action to execute in transaction</param>
		/// <param name="rollbackTransaction"></param>
		/// <returns></returns>
		public void UseTransaction(Action action, bool rollbackTransaction = false)
		{
			/* example
			new BlueToothCall().GetEntitySetting().DataStore.UseTransaction(() =>
			{
				BlueToothInbox bbb = new BlueToothInbox();
				bool success = bbb.SaveToDB();
			});
			*/

			if(this.DbSetting.DisableWrite)
			{
				return;
			}

			// critical section start
			ConnectionContext connContext = mConnectionsPool.GetConnContext(); // thread locks here if connectionContext pool is bussy

			try
			{
				connContext.Transaction = OpenConnection(connContext).BeginTransaction();

				action.Invoke();

				if (rollbackTransaction)
				{
					connContext.Transaction.Rollback();
				}
				else
				{
					connContext.Transaction.Commit();
				}

				connContext.Transaction = null;

				CloseConnection(connContext);// release context lock
			}
			catch (Exception ex)
			{
				// commit failed
				if (connContext.Transaction != null)
				{
					IDbTransaction trans = connContext.Transaction;
					connContext.Transaction = null; // release transaction before Rollback because Rollback may fail too!
					try
					{
						trans.Rollback();
					}
					catch
					{
					}
				}

				CloseConnection(connContext); // release context lock

				throw ex;
			}
		}

		#endregion


	}


}


