﻿using System;
using System.Reflection;
using System.Text.RegularExpressions;
using System.Xml;
using Ice.Orm.Core.Utility;
using System.Linq;
using System.IO;
using Ice.IO;
using Ice.Utility;

namespace Ice.Orm.Core.DataStore
{

	public class DataStoreSetting
	{

		const string mXmlFilename = "Ice.Orm.Settings.xml";

		/// <summary>
		/// <para>Defalut value is 3</para>
		/// <para> 0 means don't retry, 1..n means retry n times</para>
		/// </summary>
		private int mRetrysIfFails = 0;
		public int RetrysIfFails { get { return mRetrysIfFails; } private set { mRetrysIfFails = value; } }

		/// <summary>
		/// Has to be provided
		/// </summary>
		public string DatabaseName { get; private set; }

		/// <summary>
		/// Has to be provided
		/// </summary>
		public string ConnectionString { get; private set; }

		/// <summary>
		/// Default value is SQL
		/// </summary>
		private DataStoreProviderEnum mDBProvider = DataStoreProviderEnum.SQL;
		public DataStoreProviderEnum DBProvider { get { return mDBProvider; } private set { mDBProvider = value; } }

		/// <summary>
		/// <para>Default value is true</para>
		/// <para>once opened, leave connection opened - ConnPoolSize DBFactories per database will be created</para>
		/// </summary>
		private bool mKeepOpen = true;
		public bool KeepOpen { get { return mKeepOpen; } private set { mKeepOpen = value; } }

		/// <summary>
		/// <para>Default value is false</para>
		/// <para>disable update, insert or delete database data</para>
		/// </summary>
		private bool mDisableWrite = false;
		public bool DisableWrite { get { return mDisableWrite; } private set { mDisableWrite = value; } }

		/// <summary>
		/// <para>Default value is 5</para>
		/// <para>db connections pool size</para>
		/// </summary>
		private int mConnPoolSize = 5;
		public int ConnPoolSize { get { return mConnPoolSize; } private set { mConnPoolSize = value; } }

		private string mDateTimePattern = "yyyy-MM-dd HH:mm:ss"; // ISO STANDARD
		public string DateTimePattern { get { return mDateTimePattern; } private set { mDateTimePattern = value; } }

		/// <summary>
		/// if true, text values will be truncated to database field size to fit into
		/// </summary>
		public bool TruncateToFit { get; private set; }

		/// <summary>
		/// Default constructor
		/// </summary>
		public DataStoreSetting()
		{
		}

		/// <summary>
		/// Loads settings from Ice.Orm.Settings.xml, web.config or app.config
		/// </summary>
		/// <param name="dbname"></param>
		public DataStoreSetting(string dbname)
		{
			/* XML example:
				<DbConnections>
					<GSMS>
						<!--<provider>RemoteDB</provider><connectionString>http://www.hitoglasnik.test/RemoteDb.axd</connectionString>-->
						<provider>SQL</provider><connectionString>Data Source=localhost; Initial Catalog=GSMS;Integrated Security=True;Connect Timeout=60;</connectionString>
	 					<!--<provider>ORACLE</provider><connectionString>Data Source=XE;User Id=hr;Password=a;Integrated Security=no;</connectionString>-->
						<!--<provider>OLEDB</provider><connectionString>Provider=Microsoft.SQLSERVER.CE.OLEDB.3.5;Data Source=d:\mojmobi.sdf;</connectionString>-->
						<!--<provider>SQLCE</provider><connectionString>Data Source=d:\mojmobi.sdf;Persist Security Info=False;</connectionString>-->
						<!--<provider>ODBC</provider><connectionString>Driver={SQL Native Client};Server=OEMUSER-lap;Database=MOJMOBI; Trusted_Connection=yes;</connectionString>-->
						<!--<provider>OLEDB</provider><connectionString>Provider=SQLNCLI;Server=OEMUSER-lap;Database=MOJMOBI;Trusted_Connection=yes;</connectionString>-->
						<!--<provider>SQLITE2</provider><connectionString>Data Source=d:\mojmobi.db3;Version=3;</connectionString>-->
						<!--<provider>MYSQL</provider><connectionString>server=localhost;user id=root; password=; database=mojmobi; pooling=false</connectionString>-->
						<!--<provider>ODBC</provider><connectionString>Driver={Microsoft Access Driver (*.mdb)};Dbq=|DataDirectory|goffice.mdb;Uid=Admin;Pwd=;</connectionString>-->
						<!--<provider>ODBC</provider><connectionString>Driver={Microsoft Access Driver (*.mdb)};Dbq=d:\mojmobi.mdb;Uid=Admin;Pwd=;</connectionString>-->
						<provider>SQL</provider><connectionString>Data Source=localhost; Initial Catalog=BIDMASTER;Integrated Security=True;Pooling=false;Connect Timeout=30;</connectionString>
						<!--<provider>SQL</provider><connectionString>Data Source=OEMUSER-lap;Initial Catalog=GOFFICE;Integrated Security=False;User ID=aa;Password=malimo;</connectionString>-->
						<!--<provider>SQL</provider><connectionString>Data Source=.\SQLEXPRESS;AttachDbFilename=|DataDirectory|GOFFICE.mdf;Integrated Security=True;User Instance=True;</connectionString>-->
						<!--<provider>OLEDB</provider><connectionString>provider=Microsoft.Jet.OLEDB.4.0;Data Source=D:\mojmobi.mdb;User Id=admin;Password=;</connectionString>-->
						<!--<provider>OLEDB</provider><connectionString>provider=Microsoft.Jet.OLEDB.4.0;Data Source=|DataDirectory|goffice.mdb;User Id=admin;Password=;</connectionString>-->
						<retryIfFails>3</retryIfFails><!--  0 means don't retry, 1..n means retry n times-->
						<keepOpen>true</keepOpen><!--once opened, leave connection opened - one DBFactory per database will be created-->
						<disableWrite>false</disableWrite> <!--disable update, insert or delete database data-->
						<connPoolSize>5</connPoolSize> <!--db connections pool size-->
						<truncateToFit>true</truncateToFit><!-- trim text values if required to fit into database field -->
					</GSMS>
				</DbConnections>
			*/

			this.DatabaseName = dbname;

			if (dbname.StartsWith("SelfTestDB")) // jer se dodaje _LOCAL
			{
				this.DBProvider = DataStoreProviderEnum.SQL;
				return;
			}

			LoadXMLSettings(GetXmlRootNode(dbname));

			if (this.DBProvider == DataStoreProviderEnum.MYSQL)
			{
				this.KeepOpen = false; // MySQL connection has to be closed after each request
			}
		}

		public DataStoreSetting(DataStoreProviderEnum provider, string server, string dbname, string user, string pass)
		{
			this.DBProvider = provider;
			this.DatabaseName = dbname;

			switch (provider)
			{
				case DataStoreProviderEnum.SQL:
					if (string.IsNullOrEmpty(user + pass))
					{
						this.ConnectionString = "Data Source=" + server + ";Initial Catalog=" + dbname + ";Integrated Security=True;Connect Timeout=60;";
					}
					else
					{
						this.ConnectionString = "Data Source=" + server + ";Initial Catalog=" + dbname + ";Integrated Security=False;User ID=" + user + ";Password=" + pass + ";Connect Timeout=60;";
					}


					break;

				default:
					throw new Exception("Not implemented, provider: " + provider);
			}
		}

		private XmlNode GetXmlRootNode(string dbname)
		{
			DatabaseName = dbname;
			XmlDocument doc = GetXmlDocument(mXmlFilename);
			string xpath = "//DbConnections/" + dbname;
			XmlNode rootNode = doc.DocumentElement.SelectSingleNode(xpath);
			if (rootNode == null)
			{
				throw new Exception("Error! Can't find " + xpath + " xml node.");
			}
			return rootNode;
		}

		private void LoadXMLSettings(XmlNode rootNode)
		{
			bool providerFound = false;

			foreach (XmlNode node in rootNode.ChildNodes)
			{
				if (node.NodeType != XmlNodeType.Element)
				{
					continue;
				}

				switch (node.Name.ToLower())
				{
					case "connectionstring":
						ConnectionString = node.InnerText;
						Regex reDataDirectory = new Regex("\\|DataDirectory\\|", RegexOptions.IgnoreCase);
						Regex reAppRoot = new Regex("\\|AppRoot\\|", RegexOptions.IgnoreCase);
						if (reDataDirectory.IsMatch(ConnectionString))
						{
							// replace it with absolute path
							ConnectionString = reDataDirectory.Replace(ConnectionString, GetAppDataPath(true));
						}
						else if (reAppRoot.IsMatch(ConnectionString))
						{
							ConnectionString = reAppRoot.Replace(ConnectionString, GetAppDataPath(false));
						}
						break;

					case "retryiffails":
						int.TryParse(node.InnerText, out mRetrysIfFails);
						break;

					case "provider":
						if (providerFound)
						{
							throw new NotSupportedException("invalid config settings, database provider defined more than once");
						}
						else
						{
							providerFound = true;
							DBProvider = (DataStoreProviderEnum)Enum.Parse(typeof(DataStoreProviderEnum), node.InnerText.ToUpper(), true);
						}
						break;

					case "keepopen":
						KeepOpen = node.InnerText.Trim().ToLower() == "true";
						break;

					case "disablewrite":
						DisableWrite = node.InnerText.Trim().ToLower() == "true";
						break;

					case "connpoolsize":
						int.TryParse(node.InnerText, out mConnPoolSize);
						break;

					case "datetimeformat":
						this.DateTimePattern = node.InnerText.Trim();
						break;

					case "truncatetofit":
						this.TruncateToFit = node.InnerText.Trim() == "true";
						break;
				}
			}
		}

		private static string GetAppDataPath(bool isAppDataDir)
		{
			try
			{
				return System.Web.HttpRuntime.AppDomainAppPath + (isAppDataDir ? "App_Data\\" : ""); // if starting from different thread (there is no HttpContext.Current)
			}
			catch
			{
				try
				{
					return Path.GetFullPath(Assembly.GetExecutingAssembly().Location) + (isAppDataDir ? "App_Data\\" : "");
				}
				catch
				{

				}
			}
			return "";
		}

		private static string GetUpperDirectory(string path)
		{
			int intIndex = path.LastIndexOf("\\");
			if (intIndex > 0)
			{
				path = path.Substring(0, intIndex);
			}
			path = Path.GetFullPath(path);
			return path;
		}

		private static XmlDocument GetXmlDocument(string xmlFilename)
		{
			XmlDocument doc = new XmlDocument();

			try
			{
				// web application root\\bin\\ directory
				string webxmlConfigPath = System.Web.HttpRuntime.AppDomainAppPath + "bin\\" + xmlFilename;
				doc.Load(webxmlConfigPath);
			}
			catch
			{
				try
				{
					// directory where dlls are
					string xmlConfigPath = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location) + "\\" + xmlFilename;
					doc.Load(xmlConfigPath);
				}
				catch
				{
					try
					{
						// app.config
						string filepath = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location) + "\\" +  "app.config";
						doc.Load(filepath);
					}
					catch
					{
						try
						{
							// web.config
							string filepath = System.Web.HttpRuntime.AppDomainAppPath + "web.config";
							doc.Load(filepath);
						}
						catch
						{
							try
							{
								// 2x upper directory: ..\\..\\
								string upperPath = Path.GetFullPath(Assembly.GetExecutingAssembly().Location) + "\\";
								upperPath = GetUpperDirectory(upperPath);
								upperPath = GetUpperDirectory(upperPath);
								upperPath += xmlFilename;
								doc.Load(upperPath);

							}
							catch
							{
								try
								{
									// try embedded resource
									byte[] data = EmbeddedHelper.GetFileBytes(xmlFilename, Assembly.GetEntryAssembly()); // let's throw exception if there is no items
									MemoryStream ms = new MemoryStream(data);
									doc.Load(ms);
								}
								catch
								{
									throw new Exception("Error! Can't load configuration file.");
								}
							}
						}
					}
				}
			}
			return doc;


		}

	}
}

