﻿using System.Data;
using System.Data.OracleClient;
using System.Collections.Generic;
using System.Collections;

namespace Ice.Orm.Core.DataStore.ORACLE
{
	using Criteria;
	using Schema;
	using Ice.Orm.Core.Entity;

	public class OracleDataStore : DataStoreBase
	{
		public OracleDataStore(DataStoreSetting dbSetting)
			: base(dbSetting)
		{
		}

		protected override bool CriteriaNameRequired
		{
			get { return false; }
		}

		/// <summary>
		/// Defines the parameter preffix to use in commands for this data store.
		/// </summary>
		public override string ParameterPreffix
		{
			get { return ":"; }
		}

		/// <summary>
		/// Defines the object preffix to use in queries for this data store.
		/// </summary>
		protected override string ColumnPreffix
		{
			get { return "\""; }
		}

		/// <summary>
		/// Defines the object suffix to use in queries for this data store.
		/// </summary>
		protected override string ColumnSuffix
		{
			get { return "\""; }
		}

		/// <summary>
		/// If true, use named parameters, else use ? instead
		/// </summary>
		/// <returns></returns>
		public override bool UseNamedParameters
		{
			get { return true; }
		}

		protected override bool SupportCompoundQuery
		{
			get { return true; }
		}

		protected override string GetSelectSQLForSchema(string dbTableName)
		{
			return "SELECT * FROM " + dbTableName + " WHERE 1<>1"; // oracle doesn't understand TOP X, eg: SELECT TOP 10 * FROM...
		}

		protected override IDbConnection NewDbConnection()
		{
			return new OracleConnection();
		}

		protected override IDbCommand NewDbCommand()
		{
			return new OracleCommand();
		}

		protected override IDbDataAdapter NewDbDataAdapter()
		{
			return new OracleDataAdapter();
		}

		protected override IDbDataParameter CreateIDbDataParameter(string parameterName, object value, EntityFieldSetting schemaColumn)
		{
			OracleParameter parameter = new OracleParameter(parameterName, value);
			FixDbParameterValueAndType(parameter, schemaColumn);

			if(schemaColumn != null)
			{
				if (schemaColumn.ColumnSize > 4000)
				{
					parameter.OracleType = OracleType.NClob;
				}
				else if(schemaColumn.ColumnSize > 0)
				{
					parameter.OracleType = OracleType.NVarChar;
				}

				// Oracle hack, empty string oracle understand as null, so can't insert "" into non-nullable string column!
				if (!schemaColumn.AllowDBNull && parameter.Value is string && (string)parameter.Value == "")
				{
					parameter.Value = " ";
				}

				if (schemaColumn.ProviderType != null)
				{
					parameter.OracleType = (OracleType)schemaColumn.ProviderType;
				}
			}

			return parameter;
		}

		internal override string CleanUpSqlText(string sqlStatement)
		{
			return sqlStatement
				.Replace(" AS ", " ")
				.Replace(" as ", " ")
				.Replace("[", ColumnPreffix)
				.Replace("]", ColumnPreffix);
		}

		protected override DBWhereKeys GetDataStoreSpecificWhere(CaselessDictionary newValues, SimpleDataRow dbValues)
		{
			DBWhereKeys whereKeys = new DBWhereKeys();
			string uniqueKey = "ORA_ROWID";

			if (dbValues != null && dbValues.Table.GetColumnIndex(uniqueKey, false) >= 0)
			{
				DBWhereKey whereKey = new DBWhereKey("ROWID", dbValues[uniqueKey]);
				whereKeys.Add(whereKey);
				
				return whereKeys;
			}
			else if (newValues != null && newValues.ContainsKey(uniqueKey))
			{
				DBWhereKey whereKey = new DBWhereKey("ROWID", newValues[uniqueKey]);
				whereKeys.Add(whereKey);

				return whereKeys;
			}
			else
			{
				return null;
			}

		}

	}
}
