﻿using System;
using System.Collections;
using System.Data;
using System.Collections.Generic;
using System.Linq;

namespace Ice.Orm.Core.Entity
{
	using Criteria;
	using DataStore;
	using Utility;
	using Schema;
	using System.Data.SqlClient;
	using Ice.Serialization;
	using Ice.Utility;

	public enum WriteModeEnum
	{
		/// <summary>
		/// Default. Standard write.
		/// </summary>
		Default = 0,

		/// <summary>
		/// Update all fields.
		/// </summary>
		UpdateAll,
		
		/// <summary>
		/// Insert all fields as new object to db. Identity value will be removed and will be returned new value from database engine.
		/// </summary>
		InsertAll,

		/// <summary>
		/// Insert all fields, use identity insert if identity &gt; 0.
		/// </summary>
		IdentityInsertAll,
		
		/// <summary>
		/// if identity &gt; 0, update modified values only, else insert all values
		/// </summary>
		SaveWithIDCheck,
		
		/// <summary>
		/// if identity &gt; 0, update all values, else insert all values
		/// </summary>
		SaveAllWithIDCheck
	}

	public abstract class EntityBase: IICeSerialization
	{
		#region Private members
	
		/// <summary>
		/// values loaded than deleted from database, will be used if save is invoked after delete
		/// </summary>
		private SimpleDataRow mDeletedDatabaseValues = null;
		
		/// <summary>
		/// new, saved to db or after deletion from db, create mDbValues on first use
		/// </summary>
		private CaselessDictionary mSavedDatabaseValues = null;
		
		/// <summary>
		/// values loaded from database
		/// </summary>
		private SimpleDataRow mDbValuesItem = null;
		
		/// <summary>
		/// values set using business object public properties, on delete remains untouched, on save values are moved to mSavedDatabaseValues
		/// </summary>
		private CaselessDictionary mNewValuesItem = null;
		
		/// <summary>
		/// data store may be different than one defined by entity setting
		/// </summary>
		private DataStoreBase mNonDefaultDataStore = null;

		#endregion

		private EntitySetting mEntitySetting = null;

		private EntitySetting GetEntitySetting()
		{
			if (mEntitySetting == null)
			{
				mEntitySetting = EntitySetting.Get(this.GetType());
			}
			return mEntitySetting;
		}

		private Query mLastQuery = null;

		private SimpleDataRow DbValues
		{
			get
			{
				if (mSavedDatabaseValues != null)
				{
					if (this.mDbValuesItem == null)
					{
						this.mDbValuesItem = GetSchemaTableHelper().NewSchemaTableDataRow();
					}

					foreach (KeyValuePair<string, object> entry in mSavedDatabaseValues)
					{
						if (!(this.mDbValuesItem.Table.GetColumnIndex(entry.Key, false) >= 0))
						{
							continue;
						}

						if (entry.Value == null)
						{
							this.mDbValuesItem[entry.Key] = System.DBNull.Value;
						}
						else
						{
							this.mDbValuesItem[entry.Key] = entry.Value;
						}
					}

					mSavedDatabaseValues = null;
				}

				return mDbValuesItem;
			}
			set
			{
				this.mDbValuesItem = value;
				mSavedDatabaseValues = null;
			}
		}

		public CaselessDictionary GetModifiedValues()
		{
			return this.NewValues ?? new CaselessDictionary();
		}

		private CaselessDictionary NewValues
		{
			get
			{
				if (mDeletedDatabaseValues != null)
				{
					// if object was deleted, than save is invoked, insert all values to db
					mDbValuesItem = mDeletedDatabaseValues;
					mDeletedDatabaseValues = null;
					SetWriteMode(WriteModeEnum.IdentityInsertAll); // mDatabaseValues will be set null
				}
				return mNewValuesItem;
			}
			set
			{
				this.mNewValuesItem = value;
				mDeletedDatabaseValues = null;
			}
		}

		public DataStoreBase GetDataStore()
		{
			if (this.mNonDefaultDataStore != null)
			{
				return mNonDefaultDataStore;
			}
			else
			{
				return GetEntitySetting().DataStore;
			}
		}

		protected SchemaTableHelper GetSchemaTableHelper()
		{
			return GetEntitySetting().SchemaTableHelper;
		}

		/// <summary>
		/// Change data store. If dataStoreName is null, set default database.
		/// </summary>
		/// <param name="dataStoreName"></param>
		public void SetDataStore(string dataStoreName)
		{
			if (dataStoreName == null)
			{
				mNonDefaultDataStore = null;
			}
			else
			{
				SetDataStore(DataStoreCollection.GetByDatabaseName(dataStoreName));
			}
		}

		/// <summary>
		/// Change data store.
		/// </summary>
		/// <param name="dataStore"></param>
		public void SetDataStore(DataStoreBase dataStore)
		{
			mNonDefaultDataStore = dataStore;
		}

		/// <summary>
		/// <para>Copy all values from dbValues to mNewValues only if the value already doesn't exist in mNewValues</para>
		/// <para>if forceUpdate is true, on Persist Update will be executed, else Insert will be executed</para>
		/// </summary>
		private void SetWriteMode(WriteModeEnum mode)
		{
			switch (mode)
			{
				case WriteModeEnum.Default:
					return;

				case WriteModeEnum.IdentityInsertAll:
				case WriteModeEnum.InsertAll:
				case WriteModeEnum.UpdateAll:
					{
						if (this.NewValues == null)
						{
							this.NewValues = new CaselessDictionary();
						}

						if (this.DbValues != null)
						{
							foreach (EntityFieldSetting column in GetSchemaTableHelper().SchemaColumns)
							{
								if (!this.NewValues.ContainsKey(column.ColumnName)) // if value doesn't exist in NewValues yet
								{
									this.NewValues[column.ColumnName] = this.DbValues[column.ColumnName];
								}
							}
						}

						if (mode == WriteModeEnum.IdentityInsertAll)
						{
							this.DbValues = null;
						}
						else if (mode == WriteModeEnum.InsertAll)
						{
							// force insert, DbValues must be null
							this.DbValues = null;
							this.NewValues.Remove(GetSchemaTableHelper().IdentityColumn.ColumnName); // remove identity field
						}
						else if (mode == WriteModeEnum.UpdateAll && this.DbValues == null)
						{
							// force Update, DbValues must not be null
							this.DbValues = GetSchemaTableHelper().NewSchemaTableDataRow();
						}
					}
					break;

				case WriteModeEnum.SaveAllWithIDCheck:
				case WriteModeEnum.SaveWithIDCheck:
					{
						// if identity > 0, invoke Update, else Insert
						// requered for VistaCart and ExtJS comaptibility where an object is edited in ExtJS grid and than saved, all values are send from client including identity field value which is reference for update
						SchemaTableHelper sch = GetSchemaTableHelper();
						if (sch.IdentityColumn != null)
						{
							int? idValue = this.GetValue<int?>(sch.IdentityColumn.ColumnName);
							if (idValue > 0) // invoke update
							{
								// invoke Update, mDbValues != null && mNewValues != null
								if (mode == WriteModeEnum.SaveAllWithIDCheck)
								{
									SetWriteMode(WriteModeEnum.UpdateAll);
								}
								else if (this.DbValues == null) // if WriteKind.SaveWithIDCheck, update modified values only to speed up writing
								{
									this.DbValues = GetSchemaTableHelper().NewSchemaTableDataRow();
								}
								else
								{
									// do nothing here, DbValues != null
									return;
								}
							}
							else // invoke insert, insert all values available (from DbValues and NewValues), do not insert only modified values from NewValues!!!
							{
								SetWriteMode(WriteModeEnum.InsertAll);
							}
						}
						else
						{
							throw new Exception("SetWriteMode invoked, but there is no IdentityColumn!");
						}
					}
					break;

				default:
					throw new Exception("WriteMode " + mode + " not implemented!");

			}
		}

		/// <summary>
		/// Returns true if entity was loaded from database.
		/// </summary>
		/// <returns></returns>
		public bool IsLoadedFromDB()
		{
			return this.DbValues != null;
		}

		/// <summary>
		/// Retruns true if any value was set and need to be updated to database.
		/// </summary>
		/// <returns></returns>
		public bool IsDirty()
		{
			return this.NewValues != null && this.NewValues.Count > 0;
		}

		internal SimpleDataRow GetDataRow()
		{
			return this.DbValues;
		}

		internal bool SetDataRow(SimpleDataRow dr)
		{
			if (dr != null)
			{
				this.NewValues = null;
				this.DbValues = dr;
				return true;
			}
			else
			{
				return false;
			}
		}

		private bool SetDataTable(SimpleDataTable dt)
		{
			if (dt != null && dt.Rows.Count > 0)
			{
				return this.SetDataRow(dt.GetDataRow(0));
			}
			return false;
		}

		/// <summary>
		/// Get or set field's value
		/// </summary>
		/// <param name="fieldName"></param>
		/// <returns></returns>
		public object this[string fieldName]
		{
			get { return this.GetValue<object>(fieldName); }
			set { this.SetValue<object>(fieldName, value); }
		}

		/// <summary>
		/// Get value. First look in mDbValues, than into mNewValues, than return default(M) if all previous values are null.
		/// </summary>
		/// <typeparam name="M"></typeparam>
		/// <param name="fieldName"></param>
		/// <returns></returns>
		protected M GetValue<M>(string fieldName)
		{
			Object value = null;

			if (this.NewValues != null && this.NewValues.TryGetValue(fieldName, out value)) // value was set using set property
			{
				// OK, "value" is loaded with value
			}
			else if (this.DbValues != null && this.DbValues.Table.GetColumnIndex(fieldName, false) >= 0) // value was loaded from database
			{
				value = this.DbValues[fieldName];
			}

			if (value == null || value == System.DBNull.Value)
			{
				// get db default value
				EntityFieldSetting efs = this.GetSchemaTableHelper().SchemaColumns[fieldName];
				value = efs != null ? efs.DefaultValue : null;
				
				if(value == null && !efs.AllowDBNull)
				{
					value = IceTypeConverter.DefaultNetValue(efs.DataType);
				}

				if (value != null)
				{
					// eg. decimal can not be casted, must use Convert.ChangeType()
					return IceTypeConverter.ChangeType<M>(value);
				}

				return default(M);
			}

			if (value is M)
			{
				return (M)value;
			}
			else // conversion required, egg. int->bool, double->decimal, float->decimal...
			{
				if (typeof(M) == typeof(Boolean) && !(value is Boolean)) // MySQL don't have boolean type!!!
				{
					return (M)(Object)(Object.Equals(value, 1) || (value+"") == "1"); // can't cast bool->M, use bool->object->M
				}

				// can't convert to Nullable type, must convert to it's underlying type (int, decimal, ...)
				Type convertTo = Nullable.GetUnderlyingType(typeof(M));
				if (convertTo == null) // if it's not Nullable type
				{
					convertTo = typeof(M);
				}
				return (M)Convert.ChangeType(value, convertTo); // can't cast double to decimal, must use Convert!!!
			}
		}

		/// <summary>
		/// Set value. New value is set into mNewValues hashtable.
		/// </summary>
		/// <param name="fieldName"></param>
		/// <param name="value"></param>
		protected void SetValue<M>(string fieldName, Object value)
		{
			if (this.NewValues == null)
			{
				this.NewValues = new CaselessDictionary();
			}

			this.NewValues[fieldName] = value;
		}

		/// <summary>
		/// <para>Save entity to database.</para> 
		/// <para>Insert or Update is invoked depending on if enitity was loaded from database.</para>
		/// <para>Returns true if successful, else false</para>
		/// </summary>
		/// <returns>true if successful, else false</returns>
		public virtual bool SaveToDB()
		{
			return SaveToDB(WriteModeEnum.Default);
		}

		public virtual bool SaveToDB(WriteModeEnum mode)
		{
			SetWriteMode(mode);

			CaselessDictionary nValues = this.NewValues ?? new CaselessDictionary();
			SimpleDataRow dbValues = this.DbValues;

			// on insert, nValues is updated with autonumber field value
			int affected = GetDataStore().Save(GetSchemaTableHelper(), dbValues, nValues);

			if (affected > 0)
			{
				// nValues contains values written to database & autonumber returned from database
				// this.DbValues will be updated with savedValues on first read using this.DbValues property
				this.NewValues = null;
				this.mSavedDatabaseValues = nValues;
				return true;
			}
			else
			{
				return false;
			}
		}

		/// <summary>
		/// <para>Delete entity from database</para>
		/// <para>Returns true if successful, else false</para>
		/// </summary>
		/// <returns>return true if successful, else false</returns>
		public virtual bool DeleteFromDB()
		{
			if (GetDataStore().Delete(GetSchemaTableHelper(), this.DbValues, this.NewValues) > 0)
			{
				mDeletedDatabaseValues = this.DbValues;
				this.DbValues = null;
				// this.NewValues remains untouched, on first use invoke force update
				return true;
			}
			return false;
		}

		/// <summary>
		/// <para>Load entity from database by AutoIncrementColumn value.</para>
		/// <para>Returns true if successful, else false.</para>
		/// </summary>
		/// <param name="id">AutoIncrementColumn value</param>
		/// <returns>true if successful, else false</returns>
		public bool LoadFromDB(int? identityId)
		{
			if (identityId > 0)
			{
				Query query = new Query()
				.Select()
				.Take(1)
				.From(new SqlName(this.GetSchemaTableHelper().TableName))
				.Where(new SqlName(GetSchemaTableHelper().IdentityColumn.ColumnName), identityId.Value);

				mLastQuery = query;
				query._query.DataStore = GetDataStore();
				query._query.SchemaTableHelper = GetSchemaTableHelper();
				return SetDataTable(query._query.ToSimpleDataSet(this.GetType()).First());
			}
			else
			{
				return false;
				// do not throw exception because a lot code invokes this method with null or 0 parameter
				//	throw new NotSupportedException("error in LoadFromDB(), identityId has to be greater than 0");
			}
		}

		protected virtual T CloneValues<T>() where T:EntityBase
		{
			return CloneValues((T) Activator.CreateInstance(typeof(T)));
		}

		private T CloneValues<T>(T destination) where T:EntityBase
		{
			foreach (string key in GetSchemaTableHelper().SchemaColumns.Keys)
			{
				destination[key] = this[key];
			}
			return destination;
		}

		/// <summary>
		/// Reload entity from database, identityId is key. Returns true if entity is loaded, else false.
		/// </summary>
		/// <returns></returns>
		public virtual bool ReloadFromDB()
		{
			int identityId = GetValue<int>(GetSchemaTableHelper().IdentityColumn.ColumnName);
			if (identityId > 0)
			{
				return LoadFromDB(identityId);
			}
			else
			{
				return false;
				// do not throw exception because a lot code invokes this method with null or 0 parameter
				//throw new NotSupportedException("error in ReloadFromDB(), there is no identity field");
			}
		}

		protected T Case<T>(bool condition, Action<T> onTrue, Action<T> onFalse = null) where T:EntityBase
		{
			if (condition && onTrue != null)
			{
				onTrue.Invoke((T)this);
			}
			else if (!condition && onFalse != null)
			{
				onFalse.Invoke((T)this);
			}
			return (T)this;
		}

		public Dictionary<string, object> ToDictionary()
		{
			Dictionary<string, object> dic = new Dictionary<string, object>();
			foreach (EntityFieldSetting schemaColumn in this.GetSchemaTableHelper().SchemaColumns)
			{
				dic[schemaColumn.ColumnName] = this[schemaColumn.ColumnName];
			}
			return dic;
		}


	}

}
