﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Ice.Threading;
using System.Threading;

namespace Ice.Caching
{
	public class Cache<TKey, TValue>
	{
		public class CacheItem
		{
			internal void AdjustExpiration(double expiresAfterMs, bool isSlidingExpiration)
			{
				this.ValidityMiliseconds = expiresAfterMs;
				this.IsSlidingExpiration = isSlidingExpiration;
				this.ExpirationDate = DateTime.Now.AddMilliseconds(expiresAfterMs);
			}

			internal void AdjustSlidingExpiration()
			{
				if(this.IsSlidingExpiration && this.ValidityMiliseconds > 0)
				{
					this.ExpirationDate = DateTime.Now.AddMilliseconds(this.ValidityMiliseconds.Value);
				}
			}

			private bool IsSlidingExpiration { get; set; }
			private double? ValidityMiliseconds { get; set; }
			private DateTime? ExpirationDate { get; set; }

			internal bool IsExpired()
			{
				return this.ExpirationDate != null && this.ExpirationDate < DateTime.Now;
			}

			internal Cache<TKey, TValue> Owner { get; set; }

			public TKey Key { get; set; }
			public TValue Value { get; set; }
		}

		public event Action<CacheItem> ItemRemoved;

		private object _rwLock = new object();

		private Dictionary<TKey, CacheItem> mDicPermanent = new Dictionary<TKey, CacheItem>();
		private Dictionary<TKey, CacheItem> mDicExpiration = new Dictionary<TKey, CacheItem>();

		private int _cleaningCycleMs = 60 * 1000; // 1 minute
		public int CleaningCycleMs { get { return _cleaningCycleMs; } set { _cleaningCycleMs = value; } }

		private Thread _cleanerPool;

		private void EnsureCleanerThreadStarted()
		{
			if (_cleanerPool == null)
			{
				lock(_rwLock)
				{
					if(_cleanerPool == null)
					{
						_cleanerPool = new Thread(() =>
						{
							while (true)
							{
								Thread.Sleep(_cleaningCycleMs);

								lock (_rwLock)
								{
									this.mDicExpiration.Values.Where(o => o.IsExpired()).ToList().ForEach(o => this.mDicExpiration.Remove(o.Key));
								}
							}
						}) { IsBackground = true };
						_cleanerPool.Start();
					}
				}
			}
		}

		public Cache()
		{

		}

		public int Count
		{
			get
			{
				lock (_rwLock)
				{
					return this.mDicPermanent.Count + this.mDicExpiration.Count;
				}
			}
		}

		public Cache<TKey, TValue> Clear()
		{
			lock (_rwLock)
			{
				this.mDicPermanent.Clear();
				this.mDicExpiration.Clear();
			}
			GC.Collect();
			return this;
		}

		public Cache<TKey, TValue> Remove(TKey key)
		{
			CacheItem cc;

			lock (_rwLock)
			{
				if (this.mDicPermanent.TryGetValue(key, out cc))
				{
					this.mDicPermanent.Remove(key);
				}
				else if(this.mDicExpiration.TryGetValue(key, out cc))
				{
					this.mDicExpiration.Remove(key);
				}
			}

			if (cc != null && this.ItemRemoved != null)
			{
				try { this.ItemRemoved.Invoke(cc); }
				catch { }
			}

			return this;
		}

		public Cache<TKey, TValue> SetValue(TKey key, TValue value, TimeSpan? expiration = null, bool isSlidingExpiration = false)
		{
			lock (_rwLock)
			{
				CacheItem ci = new CacheItem() { Owner = this, Key = key, Value = value };
				
				if(expiration == null)
				{
					this.mDicPermanent[key] = ci;
				}
				else
				{
					this.mDicExpiration[key] = ci;
					EnsureCleanerThreadStarted();
					ci.AdjustExpiration(expiration.Value.TotalMilliseconds, isSlidingExpiration);
				}
			}
			return this;
		}

		public bool TryGetValue(TKey key, out TValue value)
		{
			lock (_rwLock)
			{
				CacheItem cc;

				if(this.mDicPermanent.TryGetValue(key, out cc))
				{
					// item found
				}
				else if(this.mDicExpiration.TryGetValue(key, out cc))
				{
					// item found
					if(cc.IsExpired())
					{
						this.mDicExpiration.Remove(cc.Key);
						cc = null;
					}
					else
					{
						cc.AdjustSlidingExpiration();
					}
				}

				if(cc == null)
				{
					value = default(TValue);
					return false;
				}
				else
				{
					value = cc.Value;
					return true;
				}
			}
		}

		public bool ContainsKey(TKey key)
		{
			lock (_rwLock) 
			{
				return this.mDicPermanent.ContainsKey(key) || this.mDicExpiration.ContainsKey(key);
			}
		}

		public Cache<TKey, TValue> OnItemRemoved(Action<CacheItem> action)
		{
			this.ItemRemoved += action;
			return this;

		}



	}
}
