﻿#if !SILVERLIGHT
using Ice.Web;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Web;
using System.Web.SessionState;

namespace Ice.Caching
{
	public class SessionStateProvider
	{
		const int _sessionExpirationMinutes = 20;

		public class SessionStateItem
		{
			private Dictionary<string, object> _values = new Dictionary<string, object>();

			public string SessionID { get; private set; }

			internal DateTime Expiration { get; private set; }

			internal SessionStateItem(string sessionId)
			{
				this.SessionID = sessionId;
			}

			private void AdjustExpiration()
			{
				this.Expiration = DateTime.Now.AddMinutes(_sessionExpirationMinutes);
				CookieHelper.SaveCookie(_sessionCookieName, this.SessionID, this.Expiration.AddMinutes(1)); // let's cookie expire 1 minutes later
			}

			/// <summary>
			/// Remove value by key.
			/// </summary>
			/// <param name="key"></param>
			public void Remove(string key)
			{
				lock (_values)
				{
					_values.Remove(key);
				}
			}

			/// <summary>
			/// Get/set value from current session state by key.
			/// </summary>
			/// <param name="key"></param>
			/// <returns></returns>
			public object this[string key]
			{
				get
				{
					this.AdjustExpiration();

					object obj = null;
					lock (_values)
					{
						_values.TryGetValue(key, out obj);
					}
					return obj;
				}
				set
				{
					this.AdjustExpiration();

					lock (_values)
					{
						_values[key] = value;
					}

					lock (_sessions)
					{
						if (!_sessions.ContainsKey(this.SessionID))
						{
							// if not in _sessions dictionary, add it now
							_sessions.Add(this.SessionID, this);
						}
					}
				}
			}
		}



		const string _sessionCookieName = "__icessid";

		static Dictionary<string, SessionStateItem> _sessions = new Dictionary<string, SessionStateItem>();

		static Thread _cleanerThread = new Func<Thread>(() =>
		{
			Thread thr = new Thread(() =>
			{
				while (true)
				{
					Thread.Sleep(60000); // cleaner thread runs every 1 minute
					lock (_sessions)
					{
						// remove expired sessions
						_sessions.Where(o => o.Value.Expiration < DateTime.Now).ToList().ForEach(o => _sessions.Remove(o.Key));
					}
				}
			}) { IsBackground = true };
			
			thr.Start();
			
			return thr;

		}).Invoke();

		/// <summary>
		/// Get existing or create new session state
		/// </summary>
		/// <param name="sessionId"></param>
		/// <returns></returns>
		private static SessionStateItem GetSession()
		{
			if (HttpContext.Current == null)
			{
				// there is no cookie, so return null
				return null;
			}

			// session is temporary saved into context items
			SessionStateItem ss = HttpContext.Current.Items[_sessionCookieName] as SessionStateItem;

			if (ss == null)
			{
				string sessionId = CookieHelper.GetCookieValue(_sessionCookieName);

				if(!string.IsNullOrEmpty(sessionId))
				{
					lock(_sessions)
					{
						if(_sessions.TryGetValue(sessionId, out ss) && ss.Expiration < DateTime.Now)
						{
							ss = null;
							_sessions.Remove(sessionId);
						}
					}
				}

				if(ss == null)
				{
					// create new SessionState instance and insert into Items collection
					// create new cookie for this session too, overwrite old cookie if exist
					sessionId = Guid.NewGuid().ToString() + "-" + Thread.CurrentThread.ManagedThreadId;

					HttpContext.Current.Items[_sessionCookieName] = ss = new SessionStateItem(sessionId);
				}
			}

			return ss;
		}

		/// <summary>
		/// Get current session state based on cookie
		/// </summary>
		public static SessionStateItem Session
		{
			// ako je context null, nema cookiea pa je i session null
			get {return GetSession();}
		}

	}
}

#endif

