﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Ice.Converter
{
	public class IceBaseConvert
	{
		#region chars table definition

		/// <summary>
		/// RFC base64 standard table
		/// </summary>
		public const string StandardBase64RFC = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/="; // RFC standard table

		/// <summary>
		/// RFC JavaScript table 36 Ice extended to 64 by Ice to work in URLs
		/// </summary>
		public const string JavaScriptRFCTable64 = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_-!"; // acceptable for url:  ! - _ ( )

		/// <summary>
		/// up to base64, Ice specific to support hex
		/// </summary>
		public const string Table64Ice = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_-"; // acceptable for url:  ! - _ ( )  

		/// <summary>
		/// Jubito.mobi base32 special table
		/// </summary>
		public const string Table32Jubito = "IJKL56734ABCDEFOP012VGHQRSTU89MNWXYZabcdefghijklmnopqrstuvwxyz_-"; // acceptable for url:  ! - _ ( )  

		/// <summary>
		/// RFC JavaScript table 36, up to radix 36, extracted from Chrome and Firefox, used with number.toString(radix) or parseInt(num, radix)
		/// </summary>
		public const string Table36JavaScriptRFC = "0123456789abcdefghijklmnopqrstuvwxyz";

		#endregion

		public enum RadixEnum
		{
			/// <summary>
			/// Radix 2, binary
			/// </summary>
			Base2 = 1,
			/// <summary>
			/// Radix 2, binary
			/// </summary>
			Binary = 1,
			/// <summary>
			/// Radix 4
			/// </summary>
			Base4 = 2,
			/// <summary>
			/// Radix 8, octal
			/// </summary>
			Base8 = 3,
			/// <summary>
			/// Radix 8, octal
			/// </summary>
			Octal = 3,
			/// <summary>
			/// Radix 16, hex
			/// </summary>
			Base16 = 4,
			/// <summary>
			/// Radix 16, hex
			/// </summary>
			Hex = 4,
			/// <summary>
			/// Radix 32
			/// </summary>
			Base32 = 5,
			/// <summary>
			/// Radix 64
			/// </summary>
			Base64 = 6,
		}

		#region static helper methods

		private IEnumerable<bool> GetBitsEnumerable(IEnumerable<byte> data, int fromBits, int toBits)
		{
			int bytesCount = 0;

			foreach (byte bb in data)
			{
				bytesCount++;

				for (int n = fromBits - 1; n >= 0; n--)
				{
					yield return (bb & (((byte)1) << n)) > 0;
					//yield return (bb & (((byte)1) << fromBits - n - 1)) > 0; // reversed bits
				}
			}
		}

		private IEnumerable<byte> ConvertBytes(IEnumerable<byte> bytes, int fromBits, int toBits)
		{
			IEnumerable<bool> bitsEnumerable = GetBitsEnumerable(bytes, fromBits, toBits);

			byte bb = 0;
			int bitsCount = 0;
			
			foreach (bool b in bitsEnumerable)
			{
				// take bits, reverse them, create byte from bits
				bitsCount++;
				
				bb = (byte)(bb | ((b ? 1 : 0) << (toBits - bitsCount)));
				//bb = (byte)(bb | ((b ? 1 : 0) << (count-1))); // reversed bits

				if (bitsCount == toBits)
				{
					yield return bb;
					bitsCount = 0;
					bb = 0;
				}
			}

			if (toBits < 8 && bitsCount > 0)
			{
				// ostatak
				// ako bits == 8, ignoriramo ostatak jer je nastao prethodnim paddingom do punog broja, npr kad se reducira sa 8 na 6 bitova, zadnji byte treba popuniti s ostatkom do 6
				// pri povratku sa 6 na 8 bitova, ignoriramo bitove za padding do 6
				yield return bb;
			}
		}

		#endregion

		public string CharsTable { get; set; }

		public IceBaseConvert(string charsTable = IceBaseConvert.JavaScriptRFCTable64)
		{
			this.CharsTable = charsTable;
		}

		private IEnumerable<byte> Convert(byte[] data, int fromBits, int toBits)
		{
			if (fromBits != 8 && (fromBits < 1 || fromBits > 6))
			{
				throw new ArgumentOutOfRangeException(fromBits + " is invalid value. fromBits can be in range [1-6] or 8");
			}
			if (toBits != 8 && (toBits < 1 || toBits > 6))
			{
				throw new ArgumentOutOfRangeException(toBits + " is invalid value. toBits can be in range [1-6] or 8");
			}

			var convertedEnumerable = ConvertBytes(data, fromBits, toBits);
			
			if (this.CharsTable == StandardBase64RFC)
			{
				// standard base64 RFC uses padding at the end

				if (fromBits == 8 && toBits == 6)
				{
					// encoding

					if (data.Length % 3 > 0)
					{
						// ako je Base64 RFC table, use padding
						int pad = 3 - data.Length % 3;
						byte alignmentCharIndex = (byte)(this.CharsTable.Length - 1);
						while (pad > 0)
						{
							pad--;
							convertedEnumerable = convertedEnumerable.Concat(alignmentCharIndex);
						}
					}
				}
				else if(fromBits == 6 && toBits == 8)
				{
					// decoding
					// count alignment chars
					byte alignmentCharIndex = (byte)(this.CharsTable.Length - 1);
					int cnt = 0;
					int len = data.Length;
					while(cnt < len && (data[len - 1 - cnt] == alignmentCharIndex))
					{
						cnt++;
					}

					int take = convertedEnumerable.Count() - cnt;

					convertedEnumerable = convertedEnumerable.Take(take);
				}
			}

			return convertedEnumerable;
		}

		public string ToBaseString(byte[] data, RadixEnum toRadix)
		{
			StringBuilder sb = new StringBuilder();
			foreach (char ch in Convert(data, 8, (int)toRadix).Select(o => this.CharsTable[o]))
			{
				sb.Append(ch);
			}
			return sb.ToString();
		}

		public string ToBaseString(string rawString, RadixEnum toRadix)
		{
			return ToBaseString(Encoding.UTF8.GetBytes(rawString), toRadix);
		}

		public byte[] FromBaseString(string baseString, RadixEnum fromRadix)
		{
			byte[] bytes = baseString.Select(o => (byte)this.CharsTable.IndexOf(o)).ToArray();
			return Convert(bytes, (int)fromRadix, 8).ToArray();
		}

		public string DecodeBaseStringToString(string baseString, RadixEnum fromRadix)
		{
			byte[] bytes = baseString.Select(o => (byte)this.CharsTable.IndexOf(o)).ToArray();
			byte[] data = Convert(bytes, (int)fromRadix, 8).ToArray();
			return Encoding.UTF8.GetString(data, 0, data.Length);
		}

		public string ToHexString(byte[] data)
		{
			return ToBaseString(data, RadixEnum.Hex);
		}

		public byte[] FromHexString(string hexString)
		{
			return FromBaseString(hexString, RadixEnum.Hex);
		}

	}

}
