﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;

namespace Ice.Graphics
{
	public class ImageResize
	{
		int _maxAllowedWidth = 2000;
		int _maxAllowedHeight = 2000;

		public enum CropPositionEnum
		{
			/// <summary>
			/// Default
			/// </summary>
			Center = 0,
			
			/// <summary>
			/// Start from left side and take width
			/// </summary>
			Left = 1,

			/// <summary>
			/// Start from right side and take width
			/// </summary>
			Right = 2
		}

		/// <summary>
		/// Max output width.
		/// </summary>
		public int MaxWidth { get; set; }

		/// <summary>
		/// Max output height.
		/// </summary>
		public int MaxHeight { get; set; }
		
		/// <summary>
		/// Fit smaller dimension, crop larger
		/// </summary>
		public bool Crop { get; set; }

		/// <summary>
		/// Default is Center. Used with Crop == true or ExactRect == true.
		/// </summary>
		public CropPositionEnum CropPosition { get; set; }

		/// <summary>
		/// Stretch to exact dimensions, reduce or enlarge, destroy AR.
		/// </summary>
		public bool Stretch { get; set; }
		
		/// <summary>
		/// Draw rectangle with exact dimensions and background color, than draw image inside. Keep AR.
		/// </summary>
		public bool ExactRect { get; set; }

		/// <summary>
		/// Smaller images will be enlarged.
		/// </summary>
		public bool Enlarge { get; set; }

		/// <summary>
		/// Don't shrink larger images.
		/// </summary>
		public bool NoShrink { get; set; }

		/// <summary>
		/// Force conversion.
		/// </summary>
		public bool AlwaysConvert { get; set; }

		Brush _backColor = Brushes.White;
		public Brush BackColor { get { return _backColor; } set { _backColor = value; } }

		int _quality = 100;
		/// <summary>
		/// Output quality 0 - 100. Default 100.
		/// </summary>
		public int Quality { get { return _quality; } set { _quality = value; } }

		public enum OutputFormatEnum
		{
			Original = 0,
			GIF,
			BMP,
			JPEG,
			PNG,
			TIFF
		}

		public OutputFormatEnum OutputFormat { get; set; }

		private byte[] _imgData;

		/// <summary>
		/// Larger images are always shrinked. Always keep AR.
		/// </summary>
		/// <param name="imgData"></param>
		/// <param name="width"></param>
		/// <param name="height"></param>
		/// <param name="crop"></param>
		public ImageResize(byte[] imgData, int width, int height, bool crop = false)
		{
			if(imgData == null)
			{
				throw new ArgumentNullException("imgData");
			}
			if (this.MaxWidth < 0 || this.MaxWidth > _maxAllowedWidth)
			{
				throw new ArgumentOutOfRangeException("width");
			}
			if (this.MaxHeight < 0 || this.MaxHeight > _maxAllowedHeight)
			{
				throw new ArgumentOutOfRangeException("height");
			}

			this._imgData = imgData;
			this.MaxWidth = width;
			this.MaxHeight = height;
			this.Crop = crop;
		}

		/// <summary>
		/// Resize image using current settings and return resized byte[].
		/// </summary>
		public byte[] Resize()
		{

			System.IO.MemoryStream ms = new System.IO.MemoryStream(_imgData);
			System.Drawing.Image img = System.Drawing.Image.FromStream(ms);
			System.Drawing.Imaging.ImageFormat rawformat = img.RawFormat;

			decimal thumbWidth = img.Width;
			decimal thumbHeight = img.Height;
			decimal ar = thumbWidth / thumbHeight;

			if(this.Stretch)
			{
				thumbWidth = this.MaxWidth;
				thumbHeight = this.MaxHeight;
			}
			else if ((this.Enlarge && (thumbWidth < this.MaxWidth || thumbHeight < this.MaxHeight)) || (!this.NoShrink && (thumbWidth > this.MaxWidth || thumbHeight > this.MaxHeight)))
			{
				thumbWidth = this.MaxWidth;
				thumbHeight = thumbWidth / ar;

				if ((this.Crop && thumbHeight < this.MaxHeight) || (!this.Crop && thumbHeight > this.MaxHeight))
				{
					thumbHeight = this.MaxHeight;
					thumbWidth = thumbHeight * ar;
				}
			}
			else if (this.AlwaysConvert || this.OutputFormat != OutputFormatEnum.Original)
			{
				// convert to current thumbWidth and thumbHeight
			}
			else
			{
				// conversion is not required
				// PAZI: ANIMIRAJUCI GIFOVI IZGUBE ANIMACIJU, ako ne treba resizati sliku, nemoj je ni obradjivati, vrati originalni stream
				ms.Close();
				ms.Dispose();
				img.Dispose();
				return _imgData;
			}


			// NORMAL THUMB
			System.Drawing.Bitmap thumb = new System.Drawing.Bitmap((int) thumbWidth, (int) thumbHeight);
			System.Drawing.Graphics g = System.Drawing.Graphics.FromImage(thumb);
			g.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.HighQualityBicubic;
			g.DrawImage(img, 0, 0, (int) thumbWidth, (int) thumbHeight);

			if(this.Crop || this.ExactRect)
			{
				//Brush brush = new SolidBrush(Color.Green); // Brush is abstract, SolidBrush is Brush type

				System.Drawing.Rectangle rect = new Rectangle();
				switch(this.CropPosition)
				{
					case CropPositionEnum.Center:
						rect = new System.Drawing.Rectangle((int)(thumbWidth - this.MaxWidth) / 2, (int)(thumbHeight - this.MaxHeight) / 2, this.MaxWidth, this.MaxHeight);
						break;
					case CropPositionEnum.Left:
						rect = new System.Drawing.Rectangle(0, 0, this.MaxWidth, this.MaxHeight);
						break;
					case CropPositionEnum.Right:
						rect = new System.Drawing.Rectangle((int)thumbWidth - this.MaxWidth, (int)thumbHeight - this.MaxHeight, this.MaxWidth, this.MaxHeight);
						break;
				}

				// Crop or FitWhite
				System.Drawing.Bitmap cropBmp = new System.Drawing.Bitmap(this.MaxWidth, this.MaxHeight);
				System.Drawing.Graphics g2 = System.Drawing.Graphics.FromImage(cropBmp);
				g2.FillRectangle(this.BackColor, 0, 0, this.MaxWidth, this.MaxHeight);
				g2.DrawImage(thumb,
					new System.Drawing.Rectangle(0, 0, this.MaxWidth, this.MaxHeight),
					rect,
					System.Drawing.GraphicsUnit.Pixel);

				byte[] data = GetImageData(cropBmp, GetImageCodecInfo(rawformat));

				g2.Dispose();
				cropBmp.Dispose();

				return data;
			}
			else
			{
				return GetImageData(thumb, GetImageCodecInfo(rawformat));
			}
		}

		private ImageCodecInfo GetImageCodecInfo(ImageFormat format)
		{
			switch(this.OutputFormat)
			{
				case OutputFormatEnum.Original:
					return ImageCodecInfo.GetImageEncoders().Where(o => o.FormatID == format.Guid).First();
				default:
					return ImageCodecInfo.GetImageEncoders().Where(o => o.FormatDescription == this.OutputFormat.ToString()).First();
			}
		}

		////private static byte[] GetImageData(System.Drawing.Image img, System.Drawing.Imaging.ImageFormat rawformat)
		////{
		////	// ovaj save ponekad baca exception
		////	System.IO.MemoryStream msImg = new System.IO.MemoryStream();
		////	img.Save(msImg, rawformat);
		////	byte[] data = msImg.ToArray();
		////	msImg.Close();
		////	msImg.Dispose();
		////	return data;
		////}

		////private byte[] GetImageData(System.Drawing.Image img, System.Drawing.Imaging.ImageFormat rawformat)
		////{
		////	return GetImageData(img, ImageCodecInfo.GetImageEncoders().Where(o => o.FormatID == rawformat.Guid).First());
		////}

		////private byte[] GetImageData(System.Drawing.Image img, string mime="image/jpeg")
		////{
		////	return GetImageData(img, ImageCodecInfo.GetImageEncoders().Where(o => o.MimeType == mime).First());
		////}

		private byte[] GetImageData(System.Drawing.Image img, ImageCodecInfo ici)
		{
			// .net default quality je oko 70, ali su slike s pastelnim bojama ruzne i razmazane. Ako je quality 100%, slike su ok, ali su cca 3 puta veci fileovi.
			if (img == null)
			{
				return null;
			}

			// KK: buduci da direktni img.Save(...) poziv ponekad baca GDI+ exception ako img nije dobar jpg format, pokusati cu ovako!
			System.Drawing.Bitmap newimg = new System.Drawing.Bitmap(img.Width, img.Height);
			System.Drawing.Graphics g = System.Drawing.Graphics.FromImage(newimg);
			g.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.HighQualityBicubic;
			g.DrawImage(img, 0, 0, img.Width, img.Height);

			EncoderParameters ep = new EncoderParameters();
			ep.Param[0] = new EncoderParameter(System.Drawing.Imaging.Encoder.Quality, (long)this.Quality);

			MemoryStream ms = new MemoryStream();
			newimg.Save(ms, ici, ep);

			return ms.ToArray();

		}

	}
}
