﻿using System;
using System.Net;
using System.Text;
using System.IO;
using System.Threading;

namespace Ice.Net
{
	/// <summary>
	/// HttpWebRequest implementation for Silverlight (async only)
	/// Silverlight has no GZip or Deflate streams support
	/// </summary>
	public class IceWebClient
	{
		// This class stores the State of the request.
		const int BUFFER_SIZE = 8 * 1024;
		string Url { get; set; }
		byte[] _readBuffer = new byte[BUFFER_SIZE];
		public HttpWebRequest Request { get; set; }
		public HttpWebResponse Response { get; set; }
		public Exception LastException { get; set; }
		Stream _responseStream;
		Stream _requestStream;
		public byte[] UploadData { get; set; }
		public int UploadPosition { get; set; }
		MemoryStream _responseMemoryStream = new MemoryStream();
		public byte[] ResponseData { get; set; }
		public int ResponsePosition { get; set; }
		public string ResponseString { get { return Encoding.UTF8.GetString(this.ResponseData, 0, this.ResponseData.Length); } }
		bool _instanceUsed { get; set; }

		public event Action<IceWebClient> ResponseComplete;

		public IceWebClient OnResponseComplete(Action<IceWebClient> action)
		{
			this.ResponseComplete = action;
			return this;
		}

		public event Action<IceWebClient> ResponseProgress;

		public IceWebClient OnResponseProgress(Action<IceWebClient> action)
		{
			this.ResponseProgress = action;
			return this;
		}

		public event Action<IceWebClient> UploadComplete;

		public IceWebClient OnUploadComplete(Action<IceWebClient> action)
		{
			this.UploadComplete = action;
			return this;
		}

		public event Action<IceWebClient> UploadProgress;

		public IceWebClient OnUploadProgress(Action<IceWebClient> action)
		{
			this.UploadProgress = action;
			return this;
		}

		public event Action<IceWebClient> ExceptionThrown;

		public IceWebClient OnExceptionThrown(Action<IceWebClient> action)
		{
			this.ExceptionThrown = action;
			return this;
		}

		private void HandleException(Exception ex)
		{
			this.LastException = ex;

			if (this.ExceptionThrown != null)
			{
				this.ExceptionThrown.Invoke(this);
			}
			else
			{
				throw ex;
			}
		}

		private void StartReadCallBack()
		{
			try
			{
				IAsyncResult asynchronousInputRead = this._responseStream.BeginRead(this._readBuffer, 0, BUFFER_SIZE, new AsyncCallback(delegate(IAsyncResult res)
				{
					try
					{
						int read = this._responseStream.EndRead(res);
						this.ResponsePosition += read;
						if (this.ResponseProgress != null)
						{
							this.ResponseProgress.Invoke(this);
						}

						// Read the HTML page and then do something with it
						if (read > 0)
						{
							this._responseMemoryStream.Write(this._readBuffer, 0, read);

							StartReadCallBack();
						}
						else
						{
							this.ResponseData = this._responseMemoryStream.ToArray();
							this._responseStream.Close();
							////this.IsBussy = false; // don't reset it, instance may be used only once
							if (this.ResponseComplete != null)
							{
								this.ResponseComplete.Invoke(this);
							}
						}
					}
					catch (WebException ex)
					{
						// Need to handle the exception
						HandleException(ex);
					}

				}), null);
			}
			catch (Exception e1)
			{
				HandleException(e1);
			}
		}

		private void ResponseStreamReady(IAsyncResult res)
		{
			try
			{
				// if posting data, stream has to be read on the server, else EndGetResponse throws NotFound exception
				this.Response = (HttpWebResponse)this.Request.EndGetResponse(res);
				
				this._responseStream = this.Response.GetResponseStream();

				StartReadCallBack();
			}
			catch (Exception ex)
			{
				HandleException(ex);
			}
		}

		private void WriteCallback(IAsyncResult res)
		{
			try
			{
				if (this.UploadProgress != null)
				{
					this.UploadProgress.Invoke(this);
				}
				int writeCount = 0;
				if (this.UploadData != null && this.UploadData.Length > 0 && (writeCount = this.UploadData.Length - this.UploadPosition) > 0)
				{
					if (writeCount > BUFFER_SIZE)
					{
						writeCount = BUFFER_SIZE;
					}
					int offset = this.UploadPosition;
					this.UploadPosition += writeCount;
					this._requestStream.BeginWrite(this.UploadData, offset, writeCount, new AsyncCallback(WriteCallback), null);
				}
				else
				{
					// write complete or no data to be sent
					//this._requestStream.Flush();
					//this._requestStream.Close();

					if (this.UploadComplete != null)
					{
						this.UploadComplete.Invoke(this);
					}

					this.Request.BeginGetResponse(new AsyncCallback(ResponseStreamReady), null);
				}
			}
			catch (Exception ex)
			{
				HandleException(ex);
			}
		}

		private void RequestStreamReady(IAsyncResult res)
		{
			try
			{
				using (this._requestStream = this.Request.EndGetRequestStream(res))
				{
					////WriteCallback(res);
					this._requestStream.Write(this.UploadData, 0, this.UploadData.Length);
					this._requestStream.Flush();
					this._requestStream.Close();
				}
				this.Request.BeginGetResponse(new AsyncCallback(ResponseStreamReady), null);
			}
			catch (Exception ex)
			{
				HandleException(ex);
			}
		}

		private IceWebClient StartRequest()
		{
			try
			{
				if (this._instanceUsed)
				{
					throw new Exception("Already downloading!");
				}
				else
				{
					this._instanceUsed = true;
				}

				if (this.Request.Method == "GET")
				{
					this.Request.BeginGetResponse(new AsyncCallback(ResponseStreamReady), null);
				}
				else if (this.Request.Method == "POST")
				{
					this.Request.BeginGetRequestStream(new AsyncCallback(RequestStreamReady), null);
				}
				else
				{
					throw new NotSupportedException();
				}
			}
			catch (Exception ex)
			{
				HandleException(ex);
			}

			return this;
		}

		public IceWebClient(string url)
		{
			this.Url = url;
			this.Request = (HttpWebRequest)HttpWebRequest.Create(this.Url);
		}

		public IceWebClient Download()
		{
			try
			{
				this.Request.Method = "GET";
				this.StartRequest();
			}
			catch (Exception ex)
			{
				HandleException(ex);
			}
			return this;
		}

		public IceWebClient Upload(string sendString)
		{
			try
			{
				this.UploadData = Encoding.UTF8.GetBytes(sendString);
				this.Request.Method = "POST";
				this.Request.ContentType = "application/x-www-form-urlencoded";
				//req.ContentType = "multipart/form-data";
				//req.ContentType = "application/stream";
				this.StartRequest();
			}
			catch (Exception ex)
			{
				HandleException(ex);
			}
			return this;
		}

		public IceWebClient Upload(byte[] sendData)
		{
			try
			{
				this.UploadData = sendData;
				this.Request.Method = "POST";
				//req.ContentType = "application/x-www-form-urlencoded";
				//req.ContentType = "multipart/form-data";
				this.Request.ContentType = "application/stream";
				this.StartRequest();
			}
			catch (Exception ex)
			{
				HandleException(ex);
			}
			return this;
		}
	}

}
