﻿#if !SILVERLIGHT

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Web;
using System.IO;
using System.Threading;
using System.Text.RegularExpressions;
using System.IO.Compression;
using Ice.Threading;


namespace Ice.Net
{
	public class WebDownloader
	{
		/// <summary>
		/// Timeout ms
		/// </summary>
		public int Timeout = 45000;
		public Encoding Encoding = new UTF8Encoding();
		public string Url;
		public Dictionary<string, string> Parameters;
		public bool UsePOST;
		public string ProxyAddress;
		/// <summary>
		/// Default false. Ako je true, ne koristiti proxy.
		/// </summary>
		public bool NoProxy;
		public CookieCollection Cookies;
		/// <summary>
		/// Default 0. Koliko puta pokusati ako ne uspije prvi put. Ako je -1, pokusavati ce beskonacno dok ne uspije.
		/// </summary>
		public int RetryCount = 0;
		/// <summary>
		/// Default false. Ako je true, ne mijenjati proxy ako ne uspije prvi put.
		/// </summary>
		public bool LockToProxy = false;

		public string AcceptEncoding = "gzip,deflate";
		//public string UserAgent = "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0)";
		public string UserAgent = "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)";
		//public string UserAgent = "Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.1; WOW64; Trident/4.0; SLCC2;)";
		public string Accept = "*/*";
		//public string Accept = "application/x-ms-application, image/jpeg, application/xaml+xml, image/gif, image/pjpeg, application/x-shockwave-flash, */*";
		public string AcceptLanguage = "en-US";
		//public string AcceptCharset = "ISO-8859-1,utf-8;q=0.7,*;q=0.3";
		public string AcceptCharset = "utf-8;q=0.7,*;q=0.3";

		private event Action<WebDownloader> TimeoutEvent;
		private event Action<WebDownloader> CompleteEvent;
		private event Action<WebDownloader, Exception> ExceptionEvent;
		private event Action<WebDownloader, string> WriteLogEvent;

		private void WriteLog(string msg)
		{
			if (this.WriteLogEvent != null)
			{
				this.WriteLogEvent(this, msg);
			}
		}

		public HttpWebResponse Response;
		public byte[] ResponseBytes;
		public Exception LastException;
		private string mResponseText;

		public string ResponseText
		{
			get { return mResponseText ?? (mResponseText = this.ResponseBytes != null ? this.Encoding.GetString(this.ResponseBytes) : null); }
		}

		public WebDownloader Set(Action<WebDownloader> action)
		{
			action.Invoke(this);
			return this;
		}

		private void ResetResponse()
		{
			mResponseText = null;
			Response = null;
			ResponseBytes = null;
			LastException = null;
		}

		public WebDownloader(string url)
		{
			this.Url = url;
		}

		public WebDownloader(string url, CookieCollection cookies)
		{
			this.Url = url;
			this.Cookies = cookies;
		}

		public HttpStatusCode StatusCode
		{
			get { return Response == null ? (HttpStatusCode)(-1) : Response.StatusCode; }
		}

		private static byte[] ReadDataFromStream(Stream stream)
		{
			MemoryStream ms = new MemoryStream();
			int read = -1;
			byte[] buffer = new byte[1024];
			do
			{
				read = stream.Read(buffer, 0, buffer.Length);
				ms.Write(buffer, 0, read);
			}
			while (read > 0);

			return ms.ToArray();
		}

		private byte[] DecompressData(string compression, Stream stream)
		{
			switch (compression)
			{
				case "gzip":
					return ReadDataFromStream(new GZipStream(stream, CompressionMode.Decompress));

				case "deflate":
					return ReadDataFromStream(new DeflateStream(stream, CompressionMode.Decompress));

				default: // uncompressed
					return ReadDataFromStream(stream);
			}
		}

		int tryCount;

		private void Download()
		{
			Exception ex;
			bool isHandled = false;

			tryCount = 0;

			do
			{
				isHandled = false;
				ex = null;

				tryCount++; // redni broj pokusaja

				this.ResetResponse();


				if (!this.LockToProxy && tryCount > 1)
				{
					// not first try, remove previous proxy
					this.ProxyAddress = null;
				}

				WriteLog("WebDownloader using proxy " + this.ProxyAddress);

				bool success = new IceThreadPool<bool>(1, this.DownloadWorker)
					.OnThreadException((p) =>
					{
						if (this.ExceptionEvent != null)
						{
							isHandled = true;
							this.ExceptionEvent(this, p.Exception);
						}
						else
						{
							ex = p.Exception;
						}
					})
					.OnPoolTimeout((p) =>
					{
						if (this.TimeoutEvent != null)
						{
							isHandled = true;
							this.TimeoutEvent(this);
						}
						else
						{
							ex = new TimeoutException();
						}
					})
					.StartWait(this.Timeout)
					.Results().All(o => o == true);

				if (success)
				{
					if (this.CompleteEvent != null)
					{
						this.CompleteEvent(this);
					}
					return;
				}
				else
				{
					// will retry (next do-while circle)
				}
			}
			while (this.RetryCount < 0 || (this.RetryCount > 0 && tryCount <= this.RetryCount)); // ako je retryCount == 0, nema ponovnog pokusaja

			if (!isHandled)
			{
				if (ex != null)
				{
					throw ex;
				}
				else
				{
					throw new Exception("Exception occured");
				}
			}
		}

		private bool DownloadWorker()
		{
			string queryString = this.Parameters == null ? null : this.Parameters.Select(o => HttpUtility.UrlEncode(o.Key) + "=" + HttpUtility.UrlEncode(o.Value)).ToStringJoined("&");
			string reqUrl = this.Url.Trim('?');
			if (!this.UsePOST && !string.IsNullOrEmpty(queryString))
			{
				reqUrl += "?" + queryString;
			}

			Uri uri = new Uri(reqUrl);

			HttpWebRequest httpWebRequest = (HttpWebRequest)WebRequest.Create(uri);
			httpWebRequest.CookieContainer = new CookieContainer();

			if (Cookies != null)
			{
				foreach (Cookie cookie in this.Cookies)
				{
					cookie.Domain = uri.Host;
					httpWebRequest.CookieContainer.Add(cookie);
				}
			}

			httpWebRequest.Timeout = this.Timeout;
			httpWebRequest.Accept = this.Accept;
			httpWebRequest.UserAgent = this.UserAgent;
			httpWebRequest.Headers["Accept-Language"] = this.AcceptLanguage;
			httpWebRequest.Headers["Accept-Encoding"] = this.AcceptEncoding;
			httpWebRequest.Headers["Accept-Charset"] = this.AcceptCharset;

			#region Proxy Implementation

			if (!string.IsNullOrEmpty(this.ProxyAddress))
			{
				var webProxy = new WebProxy();
				var proxyUri = new Uri("http://" + this.ProxyAddress);
				webProxy.Address = proxyUri;
				//webProxy.Credentials = new NetworkCredential("username", "password");
				httpWebRequest.Proxy = webProxy;
			}

			#endregion Proxy Implementation

			if (!this.UsePOST)
			{
				httpWebRequest.Method = "GET";
			}
			else // use post
			{
				httpWebRequest.Method = "POST";
				httpWebRequest.ContentType = "application/x-www-form-urlencoded";

				byte[] data = this.Encoding.GetBytes(queryString);
				httpWebRequest.ContentLength = data.Length;

				Stream newStream = httpWebRequest.GetRequestStream();
				newStream.Write(data, 0, data.Length);
				newStream.Close();
			}

			this.Response = (HttpWebResponse)httpWebRequest.GetResponse();

			this.Cookies = httpWebRequest.CookieContainer.GetCookies(uri);

			string compression = this.Response.GetResponseHeader("content-encoding");
			if (!string.IsNullOrEmpty(compression))
			{
				compression = compression.ToLower();
			}

			// treba implementirati: ako nema this.Response.ContentLength postavljen, moram cekati neko vrijeme nakon sto se primi zadnji byte da budem siguran da vise nista ne stize, to treba za Facebook i google, oni ne saljeu ContentLength
			
			this.ResponseBytes = DecompressData(compression, this.Response.GetResponseStream());

			try
			{
				string characterset = this.Response.CharacterSet;
				if (!string.IsNullOrEmpty(characterset))
				{
					this.Encoding = Encoding.GetEncoding(characterset);
				}
			}
			catch
			{
			}

			return true;
		}

		public byte[] DownloadData()
		{
			Download();
			return this.ResponseBytes;
		}

		public string DownloadString()
		{
			Download();
			return this.ResponseText;
		}

		public WebDownloader OnTimeout(Action<WebDownloader> action)
		{
			this.TimeoutEvent += action;
			return this;
		}

		public WebDownloader OnComplete(Action<WebDownloader> action)
		{
			this.CompleteEvent += action;
			return this;
		}

		public WebDownloader OnException(Action<WebDownloader, Exception> action)
		{
			this.ExceptionEvent += action;
			return this;
		}

		public WebDownloader OnWriteLog(Action<WebDownloader, string> action)
		{
			this.WriteLogEvent += action;
			return this;
		}
		

	}
}
#endif
