﻿#if !SILVERLIGHT

using System;
using System.Collections.Generic;
using System.Text;

namespace Ice.Security.Cryptography.RSA
{
	internal class RSACryptoHelper
	{
		#region RSA test

		public static bool RunTest(RSASecretKey skey, string dataToEncrypt)
		{
			byte[] mydata = Encoding.UTF8.GetBytes(dataToEncrypt);
			
			// encrypt and decrypt data
			byte[] enc = Encrypt(mydata, skey);
			byte[] dec = Decrypt(enc, skey);
			string decryptedString = Encoding.UTF8.GetString(dec);

			// compare
			if (dataToEncrypt != decryptedString)
			{
				throw new Exception("Error");
			}
			return true;
		}

		#endregion

		#region encrypt method

		public static byte[] Encrypt(byte[] data, RSASecretKey key)
		{
			int digitSize = key.n.getDigitSize();
			int chunkSize = digitSize -11; // 11 byte for padding if used, else chunkSize = digitSize

			// split into parts:
			List<byte[]> parts = SplitToParts(data, chunkSize);
			//List<byte[]> padedData = GetPadedData(data, chunkSize, digitSize);

			byte[] encryptedBytes = new byte[parts.Count * digitSize];
			// encrypt parts
			for(int n = 0; n < parts.Count; n++)
			{
				BigInteger biData = new BigInteger(parts[n]);
				BigInteger biEncrypted = biData.modPow(key.e, key.n);
				byte[] tmp = biEncrypted.getBytes();
				Array.Copy(tmp, 0, encryptedBytes, n * digitSize, digitSize);
			}
			return encryptedBytes;
		}

		private static List<byte[]> SplitToParts(byte[] data, int chunkSize)
		{
			List<byte[]> parts = new List<byte[]>();

			int partsCount = (data.Length / chunkSize) + (data.Length % chunkSize > 0 ? 1 : 0);
			for (int n = 0; n < partsCount; n++)
			{
				int left = data.Length - (chunkSize * n);
				int partSize = left > chunkSize ? chunkSize : left;

				byte [] tmp = new byte[partSize];
				Array.Copy(data,  (chunkSize * n), tmp, 0, partSize);
				
				parts.Add(tmp);

			}

			return parts;
		}

		private static List<byte[]> GetPadedData(byte[] data, int chunkSize, int digitSize)
		{
			List<byte[]> padedData = new List<byte[]>();

			byte[] a = data;
			int al = data.Length;

			int i;
			Random rand = new Random();

			for (i = 0; i < al; i += chunkSize)
			{
				//block = new BigInteger();
				//j = 0;

				////////////////////////////////// TYF
				// Add PKCS#1 v1.5 padding
				// 0x00 || 0x02 || PseudoRandomNonZeroBytes || 0x00 || Message
				// Variable a before padding must be of at most digitSize-11
				// That is for 3 marker bytes plus at least 8 random non-zero bytes
				int x = 0;
				int msgLength = (i + chunkSize) > al ? al % chunkSize : chunkSize;

				// Variable b with 0x00 || 0x02 at the highest index.
				byte[] b = new byte[digitSize];// new Array();
				for (x = 0; x < msgLength; x++)
				{
					b[x] = a[i + msgLength - 1 - x]; // reversed order from a array
				}
				b[msgLength] = 0; // marker// add 0 to position al + 1
				int paddedSize = Math.Max(8, digitSize - 3 - msgLength); // size of PseudoRandomNonZeroBytes in bytes

				for (x = 0; x < paddedSize; x++) // the rest of b array has to be filled with random data
				{
					b[msgLength + 1 + x] = (byte)rand.Next(1, 255);//.random() * 254) + 1; // [1,255]
				}
				// It can be asserted that msgLength+paddedSize == key.digitSize-3
				b[digitSize - 2] = 2; // marker
				b[digitSize - 1] = 0; // marker

				byte[] finalData = new byte[digitSize];
				Array.Copy(b, finalData, digitSize);
				//Array.Reverse(finalData);
				
				padedData.Add(finalData);

				//block = new BigInteger(b);
				//	for (k = 0; k < digitSize; ++j)
				//	{
				//		block.digits[j] = b[k++];
				//		block.digits[j] += b[k++] << 8;
				//	}
				////////////////////////////////// TYF

				//		var crypt = key.barrett.powMod(block, key.e);
				//		var text = key.radix == 16 ? biToHex(crypt) : biToString(crypt, key.radix);
				//		result += text + " ";
			}

			return padedData;
		}

		#endregion

		#region decrypt methods

		public static byte[] Decrypt(byte[] chiperData, RSASecretKey key)
		{
			int digitSize = key.n.getDigitSize();

			if (chiperData.Length % digitSize != 0)
			{
				return null; // invalid length
			}

			List<byte[]> parts = SplitToParts(chiperData, digitSize); // RSA encrypted block == digitSaze
			List<byte[]> decryptedList = new List<byte[]>();

			int decryptedLenth = 0;

			for (int n = 0; n < parts.Count; n++)
			{
				BigInteger biChiper = new BigInteger(parts[n]);
				BigInteger biData = biChiper.modPow(key.d, key.n);
				byte[] tmp = biData.getBytes();
				decryptedLenth += tmp.Length;
				decryptedList.Add(tmp);
			}

			int destIndex = 0;
			byte[] decryptedData = new byte[decryptedLenth]; 
			foreach(byte[] decData in decryptedList)
			{
				Array.Copy(decData, 0, decryptedData, destIndex, decData.Length);
				destIndex += decData.Length;
			}
			return decryptedData;

		}

		#endregion

		#region sign & verify

		/// <summary>
		/// Sign (encrypt) hash using private key. Returns signature.
		/// </summary>
		/// <param name="hash"></param>
		/// <returns></returns>
		public static byte[] Sign(byte[] hash, RSASecretKey key)
		{
			// encrypt hash using private key
			RSASecretKey skey = new RSASecretKey();
			skey.e = key.d; // use private key to encrypt data
			skey.n = key.n; // modulus
			return Encrypt(hash, skey);
		}

		/// <summary>
		/// Verify hash using public key. Decrypt signature using public key and compare it with hash.
		/// </summary>
		/// <param name="hash"></param>
		/// <param name="signature"></param>
		/// <returns></returns>
		public static bool Verify(byte[] hash, byte[] signature, RSASecretKey key)
		{
			// decrypt signature using public key
			RSASecretKey skey = new RSASecretKey();
			skey.d = key.e; // use public key to decrypt data
			skey.n = key.n;
			byte[] decrypted = Decrypt(signature, skey);
			if (decrypted == null || hash == null)
			{
				return false;
			}
			else
			{
				return Convert.ToBase64String(decrypted) == Convert.ToBase64String(hash);
			}
		}

		#endregion

	}

}
#endif
