﻿namespace Ice.Serialization.Json.Core.Serializer
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Globalization;
    using System.Reflection;
    using System.Runtime.CompilerServices;
	using System.Text;
	using System.Linq;

    public class JavaScriptSerializer
    {
		public enum SerializationFormat
		{
			JSON,
			JavaScript
		}

		private bool _includePropertyType = false; // for type registration in Ice framework

        private int _maxJsonLength = 10000000; // 10 MB
        private int _recursionLimit = 100;
		////private JavaScriptTypeResolver _typeResolver;
		internal static readonly long DatetimeMinTimeTicks = new DateTime(1970, 1, 1, 0, 0, 0, DateTimeKind.Utc).Ticks;
		private int _maxRecursionDepth = 100;
		private List<JavaScriptConverter> _converters;

		internal JavaScriptSerializer(bool includePropertyTypesComments)
        {
			_includePropertyType = includePropertyTypesComments;
        }

		#region converters

		public T ConvertToType<T>(object obj)
		{
			return (T)ObjectConverter.ConvertObjectToType(obj, typeof(T), this);
		}

		internal bool TryConvert(object obj, Type toType, out object result)
		{
			if (this._converters != null && obj != null)
			{
				foreach (JavaScriptConverter cc in this._converters)
				{
					if(cc.TryConvert(obj, toType, this, out result))
					{
						return true;
					}
				}
			}
			result = null;
			return false;
		}

		internal bool TrySerialize(object obj, StringBuilder sb, JavaScriptSerializer serializer, SerializationFormat serializationFormat)
		{
			if (this._converters != null && obj != null)
			{
				foreach (JavaScriptConverter cc in this._converters)
				{
					if (cc.TrySerialize(obj, sb, serializer, serializationFormat))
					{
						return true;
					}
				}
			}
			return false;
		}

        protected void RegisterConverters(IEnumerable<JavaScriptConverter> converters)
        {
			if (this._converters == null)
			{
				this._converters = new List<JavaScriptConverter>();
			}
			this._converters.AddRange(converters);
        }

		#endregion

		#region deserialize methods

		public T Deserialize<T>(string input)
		{
			return (T)Deserialize(this, input, typeof(T), this.RecursionLimit);
		}

		internal static object Deserialize(JavaScriptSerializer serializer, string input, Type type, int depthLimit)
		{
			if (input == null)
			{
				throw new ArgumentNullException("input");
			}
			if (input.Length > serializer.MaxJsonLength)
			{
				throw new ArgumentException(AtlasWeb.JSON_MaxJsonLengthExceeded, "input");
			}
			return ObjectConverter.ConvertObjectToType(JavaScriptObjectDeserializer.BasicDeserialize(input, depthLimit, serializer), type, serializer);
		}

		public object DeserializeObject(string input)
		{
			return Deserialize(this, input, null, this.RecursionLimit);
		}

		#endregion

		#region serialize methods

		public string Serialize(object obj)
        {
            return this.Serialize(obj, SerializationFormat.JSON);
        }

        private void Serialize(object obj, StringBuilder output)
        {
            this.Serialize(obj, output, SerializationFormat.JSON);
        }

        private string Serialize(object obj, SerializationFormat serializationFormat)
        {
            StringBuilder output = new StringBuilder();
            this.Serialize(obj, output, serializationFormat);
            return output.ToString();
        }

        internal void Serialize(object obj, StringBuilder output, SerializationFormat serializationFormat)
        {
            this.SerializeValue(obj, output, 0, null, serializationFormat);
            if ((serializationFormat == SerializationFormat.JSON) && (output.Length > this.MaxJsonLength))
            {
                throw new InvalidOperationException(AtlasWeb.JSON_MaxJsonLengthExceeded);
            }
        }

		#endregion

		#region serializaion by type

		private static void SerializeBoolean(bool o, StringBuilder sb)
        {
            if (o)
            {
                sb.Append("true");
            }
            else
            {
                sb.Append("false");
            }
        }

        private void SerializeCustomObject(object o, StringBuilder sb, int depth, Hashtable objectsInUse, SerializationFormat serializationFormat)
        {
            bool flag = true;
            Type type = o.GetType();
            sb.Append('{');
          
            foreach (FieldInfo info in type.GetFields(BindingFlags.Public | BindingFlags.Instance))
            {
                if (!info.IsDefined(typeof(ScriptIgnoreAttribute), true) && !info.IsDefined(typeof(NonSerializedAttribute), true))
                {
                    if (!flag)
                    {
                        sb.Append(',');
                    }
					if (_includePropertyType)
					{
						sb.AppendLine("\r\n// " + Ice.Utility.IceTools.GetJavaScriptTypeName(info.FieldType));
					}
                    SerializeString(info.Name, sb);
                    sb.Append(':');
                    this.SerializeValue(info.GetValue(o), sb, depth, objectsInUse, serializationFormat);
                    flag = false;
                }
            }
            foreach (PropertyInfo info2 in type.GetProperties(BindingFlags.GetProperty | BindingFlags.Public | BindingFlags.Instance))
            {
				if (!info2.IsDefined(typeof(ScriptIgnoreAttribute), true) && !info2.IsDefined(typeof(NonSerializedAttribute), true))
                {
                    MethodInfo getMethod = info2.GetGetMethod();
                    if ((getMethod != null) && (getMethod.GetParameters().Length <= 0))
                    {
                        if (!flag)
                        {
                            sb.Append(',');
                        }
						if (_includePropertyType)
						{
							sb.AppendLine("\r\n// " + Ice.Utility.IceTools.GetJavaScriptTypeName(info2.PropertyType));
						}
                        SerializeString(info2.Name, sb);
                        sb.Append(':');
						
						this.SerializeValue(getMethod.Invoke(o, null), sb, depth, objectsInUse, serializationFormat);

                        flag = false;
                    }
                }
            }
            sb.Append('}');
        }

		private static void SerializeDateTime(DateTime datetime, StringBuilder sb, SerializationFormat serializationFormat)
		{
			// original MS methods
			if (serializationFormat == SerializationFormat.JSON)
			{
				sb.Append("\"\\/Date(");
				sb.Append((long)((datetime.ToUniversalTime().Ticks - DatetimeMinTimeTicks) / 10000)); // 1 .net tick = 100 ns
				sb.Append(")\\/\"");
			}
			else
			{
				sb.Append("new Date(");
				sb.Append((long)((datetime.ToUniversalTime().Ticks - DatetimeMinTimeTicks) / 10000));
				sb.Append(")");
			}
		}

		private void SerializeDictionary(IDictionary o, StringBuilder sb, int depth, Hashtable objectsInUse, SerializationFormat serializationFormat)
		{
			sb.Append('{');
			bool flag = true;

			foreach (DictionaryEntry entry in o)
			{
				string key = entry.Key as string;
				if (key == null)
				{
					throw new ArgumentException(string.Format(CultureInfo.InvariantCulture, AtlasWeb.JSON_DictionaryTypeNotSupported, new object[] { o.GetType().FullName }));
				}

				if (!flag)
				{
					sb.Append(',');
				}
				this.SerializeDictionaryKeyValue(key, entry.Value, sb, depth, objectsInUse, serializationFormat);
				flag = false;
			}
			sb.Append('}');
		}

        private void SerializeDictionaryKeyValue(string key, object value, StringBuilder sb, int depth, Hashtable objectsInUse, SerializationFormat serializationFormat)
        {
			if (_includePropertyType)
			{
				sb.AppendLine("\r\n// " + Ice.Utility.IceTools.GetJavaScriptTypeName(value));
			}
            SerializeString(key, sb);
            sb.Append(':');
            this.SerializeValue(value, sb, depth, objectsInUse, serializationFormat);
        }

        private void SerializeEnumerable(IEnumerable enumerable, StringBuilder sb, int depth, Hashtable objectsInUse, SerializationFormat serializationFormat)
        {
            sb.Append('[');
            bool flag = true;
            foreach (object obj2 in enumerable)
            {
                if (!flag)
                {
                    sb.Append(',');
                }
                this.SerializeValue(obj2, sb, depth, objectsInUse, serializationFormat);
                flag = false;
            }
            sb.Append(']');
        }

        private static void SerializeGuid(Guid guid, StringBuilder sb)
        {
            sb.Append("\"").Append(guid.ToString()).Append("\"");
        }

        private static void SerializeString(string input, StringBuilder sb)
        {
            sb.Append('"');
            sb.Append(JavaScriptString.QuoteString(input));
            sb.Append('"');
        }

        private static void SerializeUri(Uri uri, StringBuilder sb)
        {
            sb.Append("\"").Append(uri.GetComponents(UriComponents.SerializationInfoString, UriFormat.UriEscaped)).Append("\"");
        }

        private void SerializeValue(object o, StringBuilder sb, int depth, Hashtable objectsInUse, SerializationFormat serializationFormat)
        {
			if (depth > this._maxRecursionDepth) // customization by IceASP
			{
				return;
			}

			++depth;

            if ( depth > this._recursionLimit)
            {
                throw new ArgumentException(AtlasWeb.JSON_DepthLimitExceeded);
            }

			if(!this.TrySerialize(o, sb, this, serializationFormat))
			{
                this.SerializeValueInternal(o, sb, depth, objectsInUse, serializationFormat);
            }
        }

        private void SerializeValueInternal(object o, StringBuilder sb, int depth, Hashtable objectsInUse, SerializationFormat serializationFormat)
        {
            if ((o == null) || DBNull.Value.Equals(o))
            {
                sb.Append("null");
            }
            else
            {
				if (o is IICeSerialization)
				{
					o = ((IICeSerialization)o).ToDictionary();
				}

                string input = o as string;
                if (input != null)
                {
                    SerializeString(input, sb);
                }
                else if (o is char)
                {
                    if (((char) o) == '\0')
                    {
                        sb.Append("null");
                    }
                    else
                    {
                        SerializeString(o.ToString(), sb);
                    }
                }
                else if (o is bool)
                {
                    SerializeBoolean((bool) o, sb);
                }
                else if (o is DateTime)
                {
                    SerializeDateTime((DateTime) o, sb, serializationFormat);
                }
                else if (o is Guid)
                {
                    SerializeGuid((Guid) o, sb);
                }
                else
                {
                    Uri uri = o as Uri;
                    if (uri != null)
                    {
                        SerializeUri(uri, sb);
                    }
                    else if (o is double)
                    {
                        sb.Append(((double) o).ToString("r", CultureInfo.InvariantCulture));
                    }
                    else if (o is float)
                    {
                        sb.Append(((float) o).ToString("r", CultureInfo.InvariantCulture));
                    }
                    else if (o.GetType().IsPrimitive || (o is decimal))
                    {
                        IConvertible convertible = o as IConvertible;
                        if (convertible != null)
                        {
                            sb.Append(convertible.ToString(CultureInfo.InvariantCulture));
                        }
                        else
                        {
                            sb.Append(o.ToString());
                        }
                    }
                    else
                    {
                        Type type = o.GetType();
                        if (type.IsEnum)
                        {
                            sb.Append((int) o);
                        }
                        else
                        {
                            try
                            {
                                if (objectsInUse == null)
                                {
                                    objectsInUse = new Hashtable(new ReferenceComparer());
                                }
                                else if (objectsInUse.ContainsKey(o))
                                {
                                    throw new InvalidOperationException(string.Format(CultureInfo.CurrentCulture, AtlasWeb.JSON_CircularReference, new object[] { type.FullName }));
                                }
                                objectsInUse.Add(o, null);
                                IDictionary dictionary = o as IDictionary;
                                if (dictionary != null)
                                {
                                    this.SerializeDictionary(dictionary, sb, depth, objectsInUse, serializationFormat);
                                }
                                else
                                {
                                    IEnumerable enumerable = o as IEnumerable;
                                    if (enumerable != null)
                                    {
                                        this.SerializeEnumerable(enumerable, sb, depth, objectsInUse, serializationFormat);
                                    }
                                    else
                                    {
                                        this.SerializeCustomObject(o, sb, depth, objectsInUse, serializationFormat);
                                    }
                                }
                            }
                            finally
                            {
                                if (objectsInUse != null)
                                {
                                    objectsInUse.Remove(o);
                                }
                            }
                        }
                    }
                }
            }
        }

		#endregion

		#region helper properties and methods

		public int MaxJsonLength
        {
            get
            {
                return this._maxJsonLength;
            }
            set
            {
                if (value < 1)
                {
                    throw new ArgumentOutOfRangeException(AtlasWeb.JSON_InvalidMaxJsonLength);
                }
                this._maxJsonLength = value;
            }
        }

		/// <summary>
		/// Throw exception is recursion level is reached
		/// </summary>
        protected int RecursionLimit
        {
            get
            {
                return this._recursionLimit;
            }
            set
            {
                if (value < 1)
                {
                    throw new ArgumentOutOfRangeException(AtlasWeb.JSON_InvalidRecursionLimit);
                }
                this._recursionLimit = value;
            }
        }

		/// <summary>
		/// Default value is 100, min value 1
		/// </summary>
		public int MaxRecursionDepth
		{
			get { return this._maxRecursionDepth; }
			set 
			{
				if (value < 1)
				{
					throw new Exception("JSON_InvalidMaxRecursionDepth");
				}
				this._maxRecursionDepth = value; 
			}
		}

		////internal JavaScriptTypeResolver TypeResolver
		////{
		////    get
		////    {
		////        return this._typeResolver;
		////    }
		////}

#if SILVERLIGHT

        private class ReferenceComparer : IEqualityComparer<object>
        {
            bool IEqualityComparer<object>.Equals(object x, object y)
            {
                return (x == y);
            }

            int IEqualityComparer<object>.GetHashCode(object obj)
            {
                return RuntimeHelpers.GetHashCode(obj);
            }
        }
#else
		private class ReferenceComparer : IEqualityComparer
		{
			bool IEqualityComparer.Equals(object x, object y)
			{
				return (x == y);
			}

			int IEqualityComparer.GetHashCode(object obj)
			{
				return RuntimeHelpers.GetHashCode(obj);
			}
		}
#endif

		#endregion
	}
}

