﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace Ice.Threading.Abstract
{
	/*
		AppDomain.CurrentDomain.UnhandledException += delegate(object sender,UnhandledExceptionEventArgs e){
					MessageBox.Show("Error: " + ((Exception)e.ExceptionObject).Message,"Error:",MessageBoxButtons.OK,MessageBoxIcon.Error);
				};
				Application.ThreadException += delegate(object sender,System.Threading.ThreadExceptionEventArgs e){
					MessageBox.Show("Error: " + e.Exception.Message,"Error:",MessageBoxButtons.OK,MessageBoxIcon.Error);
				};

	*/

	public abstract class IceThreadPoolBase<TPool, PItem> 
		where PItem : IceThreadPoolBase<TPool, PItem>.PoolItemBase
		where TPool : IceThreadPoolBase<TPool, PItem>
	{
		public abstract class PoolItemBase
		{
			public TPool Pool { get; set; }

			/// <summary>
			/// Current thread.
			/// </summary>
			public Thread Thread { get; set; }

			/// <summary>
			/// Exception thrown in callback method.
			/// </summary>
			public Exception Exception { get; set; }

			public bool IsFinished { get; set; }

			public bool IsThreadStarted { get; set; }

			public bool IsStartRequested { get; set; }

			public bool IsAborted { get; set; }

			public bool IsSuccessful { get { return this.IsFinished && this.Exception == null; } }

		}

		int mThreadsCount;
		bool mPoolCompleteSent;

		public event Action<PItem> ThreadCompleteEvent;
		public event Action<PItem> ThreadExceptionEvent;
		public event Action<TPool> PoolCompleteEvent;
		public event Action<TPool> PoolStartEvent;
		public event Action<TPool> PoolTimeoutEvent;

		public IceThreadPoolBase(int threadsCount)
		{
			this.mThreadsCount = threadsCount;
		}

		List<PItem> mItems;

		object mWaitLock = new object();
		object mTimeoutLock = new object();
		object mItemsLock = new object();

		protected abstract void InvokeWorkerMethod(PItem ss);

		private void ThreadWorkerMethod(PItem ss)
		{
			lock (mWaitLock)
			{
				if (this.Items().All(o => !o.IsThreadStarted))
				{
					if (this.PoolStartEvent != null)
					{
						try { this.PoolStartEvent.Invoke((TPool)this); }
						catch { }
					}
				}
				ss.IsThreadStarted = true;
			}

			try
			{
				InvokeWorkerMethod(ss);
			}
			catch (Exception ex)
			{
				ss.Exception = ex;

				if (this.ThreadExceptionEvent != null)
				{
					try { this.ThreadExceptionEvent.Invoke(ss); }
					catch { }
				}
				else
				{
					// if there is no ThreadExceptionEvent, exception is not handled
				}
			}
			finally
			{
				ss.IsFinished = true;

				if (this.ThreadCompleteEvent != null)
				{
					try { this.ThreadCompleteEvent.Invoke(ss); }
					catch { }
				}

				lock (mWaitLock)
				{
					if (this.Items().All(o => o.IsFinished))
					{
						if (this.PoolCompleteEvent != null && !this.mPoolCompleteSent)
						{
							this.mPoolCompleteSent = true;
							try { this.PoolCompleteEvent.Invoke((TPool)this); }
							catch { }
						}

						Monitor.PulseAll(mWaitLock);
					}
				}
			}
		}

		/// <summary>
		/// Threads containers.
		/// </summary>
		/// <returns></returns>
		public List<PItem> Items()
		{
			if (mItems == null)
			{
				lock (mItemsLock)
				{
					if (mItems == null)
					{
						List<PItem> items = new List<PItem>(this.mThreadsCount);

						for (int n = 0; n < this.mThreadsCount; n++)
						{
							PItem ss = Activator.CreateInstance<PItem>();
							ss.Pool = (TPool)this;

							ss.Thread = new Thread(delegate()
							{
								ThreadWorkerMethod(ss);
							})
							{
								IsBackground = true, // if true, thread is aborted when calling process aborted. If not background thread, it will stay alive even if main proces aborted.
								Name = "thread[" + n + "]"
							};

							items.Add(ss);
						}

						// assign them at the end
						mItems = items;
					}
				}
			}
			return mItems;
		}

		private IEnumerable<PItem> GetRunningItems()
		{
			return this.Items().Where(o => o.IsStartRequested && !o.IsFinished && !o.IsAborted);
		}

		public bool IsRunning
		{
			get { return GetRunningItems().Any(); }
		}

		/// <summary>
		/// Returns non-null exceptions if any.
		/// </summary>
		/// <returns></returns>
		public IEnumerable<Exception> Exceptions()
		{
			return this.Items().Where(o => o.Exception != null).Select(o => o.Exception);
		}

		private bool _abortRequested;

		public TPool Abort()
		{
			_abortRequested = true;

			lock (mWaitLock)
			{
				this.Items().Where(o => o.IsStartRequested && !o.IsFinished).ForEach(o =>
				{
					o.IsAborted = true;
					try { o.Thread.Abort(); }
					catch { };
				});

				Monitor.PulseAll(mWaitLock); // release wait lock
			}
			return (TPool) this;
		}

		/// <summary>
		/// Start worker threads.
		/// </summary>
		public TPool Start()
		{
			_abortRequested = false;

			lock (mWaitLock)
			{
				if (!_abortRequested)
				{
					if (this.Items().Any(o => o.IsStartRequested))
					{
						throw new Exception("Pool is already running");
					}

					this.mPoolCompleteSent = false;

					this.Items().ForEach(o =>
					{
						if (!_abortRequested)
						{
							o.IsStartRequested = true;
							o.Thread.Start();
						}
					});
				}
			}

			return (TPool) this;
		}

		private void WaitUntillAllFinishedOrTimeouted(int timeoutMs)
		{
			lock (mWaitLock)
			{
				if (this.Items().Where(o => !o.IsFinished).Count() > 0)
				{
					if (timeoutMs >= 0)
					{
						// Returns:
						//     true if the lock was reacquired before the specified time elapsed; false
						//     if the lock was reacquired after the specified time elapsed. The method does
						//     not return until the lock is reacquired.
						if(!Monitor.Wait(mWaitLock, timeoutMs))
						{
							// timeouted
						}
						else
						{
							// not timeouted
						}
					}
					else
					{
						Monitor.Wait(mWaitLock);
					}
				}
			}

			if (this.IsRunning)
			{
				// some threads are timeout
				if (this.PoolTimeoutEvent != null)
				{
					// callback on non-finished threads
					try
					{
						this.PoolTimeoutEvent((TPool)this);
					}
					catch
					{
					}
				}
				// abort unfinished threads
				this.Abort();
			}
		}

		/// <summary>
		/// Blocks calling thread and wait for all worker threads to finish or timeout. Timeouted threads are aborted.
		/// </summary>
		/// <param name="timeoutMs">if 0 or less than 0, wait until all threads are finished, or forever, else wait specified miliseconds</param>
		/// <returns></returns>
		public TPool StartWait(int timeoutMs)
		{
			this.Start();
			WaitUntillAllFinishedOrTimeouted(timeoutMs);
			return (TPool) this;
		}

		/// <summary>
		/// Block calling thread and wait for all worker threads to finish.
		/// </summary>
		/// <returns></returns>
		public TPool StartWait()
		{
			return this.StartWait(-1);
		}

		/// <summary>
		/// Start worker threads. Creates new watchdog thread which kills worker threads if timeouted.
		/// </summary>
		/// <param name="timeoutMs"></param>
		/// <returns></returns>
		public TPool StartTimeout(int timeoutMs)
		{
			new Thread(delegate()
			{
				WaitUntillAllFinishedOrTimeouted(timeoutMs);
			}) { IsBackground = true }.Start();

			this.Start();

			return (TPool) this;
		}

		/// <summary>
		/// Callback on thread comlete, successfully or not, doesn't matter.
		/// </summary>
		/// <param name="action"></param>
		/// <returns></returns>
		public TPool OnThreadComplete(Action<PItem> action)
		{
			this.ThreadCompleteEvent = action;
			return (TPool) this;
		}

		/// <summary>
		/// Callback on thread exception.
		/// </summary>
		/// <param name="action"></param>
		/// <returns></returns>
		public TPool OnThreadException(Action<PItem> action)
		{
			this.ThreadExceptionEvent = action;
			return (TPool) this;
		}

		/// <summary>
		/// Callback on all threads completed, successfully or not, doesn't matter.
		/// </summary>
		/// <param name="action"></param>
		/// <returns></returns>
		public TPool OnPoolComplete(Action<TPool> action)
		{
			this.PoolCompleteEvent = action;
			return (TPool) this;
		}

		/// <summary>
		/// Callback on first thread started.
		/// </summary>
		/// <param name="action"></param>
		/// <returns></returns>
		public TPool OnPoolStart(Action<TPool> action)
		{
			this.PoolStartEvent = action;
			return (TPool) this;
		}

		/// <summary>
		/// Callback on pool timeout.
		/// </summary>
		/// <param name="action"></param>
		/// <returns></returns>
		public TPool OnPoolTimeout(Action<TPool> action)
		{
			this.PoolTimeoutEvent = action;
			return (TPool) this;
		}

	}
}
