﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;
using System.IO;

namespace Ice.Utility
{
	public class EmbeddedHelper
	{
		public class EmbeddedFileInfo
		{
			private Assembly mAssembly;
			public string Name { get; private set; }

			private byte[] mFileBytes;
			public byte[] FileBytes
			{
				get
				{
					if (mFileBytes == null)
					{
						Stream s = this.mAssembly.GetManifestResourceStream(this.Name);
						byte[] data = new byte[s.Length];
						s.Read(data, 0, data.Length);
						s.Dispose();
						mFileBytes = data;
					}
					return mFileBytes;
				}
			}

			private string mFileText;
			public string FileText
			{
				get { return mFileText ?? (mFileText = IceTools.RemoveUtfHeader(Encoding.UTF8.GetString(this.FileBytes, 0, this.FileBytes.Length))); }
			}

			public EmbeddedFileInfo(string name, Assembly assembly)
			{
				this.Name = name;
				this.mAssembly = assembly;
			}
		}

		public static EmbeddedFileInfo[] GetFilesInfo(params Assembly[] assembly)
		{
			return assembly.SelectMany(asse =>
			{
				var names = asse.GetManifestResourceNames();
				return names.Where(k=>!string.IsNullOrEmpty(k)).Select(k => new EmbeddedFileInfo(k, asse));
			}).ToArray();
		}

		private static byte[] ReadFileBytes(string name, Assembly assembly)
		{
			Stream s = assembly.GetManifestResourceStream(name);
			byte[] data = new byte[s.Length];
			s.Read(data, 0, data.Length);
			s.Dispose();

			return data;
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="resourceName">Case non-sensitive search</param>
		/// <param name="assembly"></param>
		/// <returns></returns>
		public static byte[] GetFileBytes(string resourceName, Assembly assembly)
		{
			string[] marray = assembly.GetManifestResourceNames();
			string name = marray.Where(o => o != null && o.EndsWith(resourceName, StringComparison.OrdinalIgnoreCase)).SingleOrDefault(); // must be unique name

			if (!string.IsNullOrEmpty(name))
			{
				return ReadFileBytes(name, assembly);
			}

			throw new Exception("Embedded resource not found: " + resourceName);
		}

		/// <summary>
		/// Get resource from calling assembly
		/// </summary>
		/// <param name="resourceName">Case non-sensitive search.</param>
		/// <returns></returns>
		public static byte[] GetFileBytes(string resourceName)
		{
			return GetFileBytes(resourceName, Assembly.GetCallingAssembly());
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="resourceName">Case non-sensitive search.</param>
		/// <param name="assembly"></param>
		/// <returns></returns>
		public static string GetFileText(string resourceName, Assembly assembly)
		{
			byte[] data = GetFileBytes(resourceName, assembly);
			return IceTools.RemoveUtfHeader(Encoding.UTF8.GetString(data, 0, data.Length));
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="resourceName">Case non-sensitive search.</param>
		/// <returns></returns>
		public static string GetFileText(string resourceName)
		{
			byte[] data = GetFileBytes(resourceName, Assembly.GetCallingAssembly());
			return IceTools.RemoveUtfHeader(Encoding.UTF8.GetString(data, 0, data.Length));
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="resourceName">Case non-sensitive search.</param>
		/// <param name="assembly"></param>
		/// <returns></returns>
		public static bool ContainsFile(string resourceName, Assembly assembly)
		{
			string[] marray = assembly.GetManifestResourceNames();
			return marray.Where(o => o != null && o.EndsWith(resourceName, StringComparison.OrdinalIgnoreCase)).SingleOrDefault() != null; // must be single or null
		}

		/// <summary>
		/// Get resource from calling assembly
		/// </summary>
		/// <param name="resourceName">Case non-sensitive search.</param>
		/// <returns></returns>
		public static bool ContainsFile(string resourceName)
		{
			return ContainsFile(resourceName, Assembly.GetCallingAssembly());
		}

		private static int GetLastSlashIndex(string filename)
		{
			int index = filename.LastIndexOf('/');
			if (!(index >= 0))
			{
				index = filename.LastIndexOf('\\');
			}
			return index;
		}

		private static string GetDirectory(string filename)
		{
			int index = GetLastSlashIndex(filename);
			if(index >= 0)
			{
				return filename.Remove(index + 1);
			}
			return "";
		}

		private static string GetFileName(string filename)
		{
			int index = GetLastSlashIndex(filename);
			if (index >= 0 && filename.Length > index)
			{
				return filename.Substring(index+1);
			}
			return filename;
		}

		public static string GetEmbeddedFilePath(string fileName)
		{
			// - in directories is replaced with _ , but  - in filename is not replaced
			// embedded file
			string dir = GetDirectory(fileName).Replace("-", "_");
			string file = GetFileName(fileName);
			return "." + (dir + file).Trim('~', '/', '.', ' ').Replace("/", ".");
		}

	}
}