﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using System.Text.RegularExpressions;
using System.Diagnostics;
using Ice.Serialization.Json.Core.Serializer;

namespace Ice.Xml.Parser
{
	public enum SourceKindEnum
	{
		Html,
		Xml,
		Json
	}

	public enum SegmentKind
	{
		Unknown,
		FreeText,
		scriptNode,
		styleNode,
		cdataNode,
		phpCodeBlock,
		xmlHeader,
		htmlComment,
		htmlDoctype,
		aspxComment,
		aspxHeader,
		aspxCodeBlock,
		NodeSelfClosed,
		StartTag,
		EndTag,
		
		// inner matches
		attr,
		tag,
		content
	}

	internal class XmlParser
	{
		#region common methods & variables

		XmlNode currentElement = null;
		List<XmlNode> allNodes = new List<XmlNode>();
		StringBuilder freeTextBuffer = new StringBuilder();

		private XmlParser()
		{

		}

		private void FlushFreeTextBuffer()
		{
			if(freeTextBuffer != null)
			{
				AddTextNode(freeTextBuffer.ToString(), XmlNodeType.Text);
				freeTextBuffer = null;
			}
		}

		private void AddFreeTextToBuffer(string txt)
		{
			if (freeTextBuffer == null)
			{
				freeTextBuffer = new StringBuilder();
			}
			freeTextBuffer.Append(txt);
		}

		private void AddChildNode(XmlNode parent, XmlNode child)
		{
			child.Parent = parent;
			child.Level = parent.Level + 1;
			parent.Children.Add(child);
			allNodes.Add(child);
		}

		private void AddTextNode(string nodeValue, XmlNodeType nodeType, object dataItem = null)
		{
			AddChildNode(currentElement, new XmlNode()
			{
				NodeType = nodeType,
				ParsingState = ParsingStateEnum.NodeClosed,
				NodeValue = nodeValue,
				DataItem = dataItem
			});
		}

		#endregion

		#region parse HTML & XML

		#region regex for HTML & XML parsing



		private static string GetRegexPattern(bool isHtml)
		{
			// xml specific part is always included
			// html specific part is included only if isHtml == true
			// warning: do not change patterns order

			return new[]
			{
				new {include = true, regex = "(?<FreeText>((< )|( >)))"}, // space before > or after <, it is not tag, eg. <div>temp >10 and < temp11 10 < ma < ne < Moze </div>
				new {include = isHtml, regex = "(?<scriptNode><script(?<attr>[^>]*)>(?<content>[\\w\\W]*?)</script>)"}, // <script type="text/javascript"> ...content... </script>
				new {include= isHtml, regex = "(?<styleNode><style(?<attr>[^>]*)>(?<content>[\\w\\W]*?)</style>)"}, // <style type="text/css"> ...content... </style>
				new {include = true, regex = "(?<cdataNode><\\!\\[CDATA\\[(?<cdata>[\\w\\W]*?)\\]\\]>)"}, // <![CDATA[   ...some text or script...  ] ]>
				new {include = isHtml, regex = "(?<phpCodeBlock>[<][\\?]php[\\w\\W]*?[\\?][>])"}, // <?php .... ?>
				new {include = true, regex = "(?<xmlHeader>[<][\\?]xml(?<attr>[^>]*?)[\\?][>])"}, // <?xml version="1.0" encoding="utf-8" ?>
				new {include = true, regex = "(?<htmlComment><\\!\\-\\-[\\w\\W]*?\\->)"}, // <!--html comment-->
				new {include = isHtml, regex = "(?<htmlDoctype><\\!DOCTYPE[^>]*>)"}, // <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
				new {include = isHtml, regex = "(?<aspxComment><%\\-\\-[\\w\\W]*?\\-\\-%>)"}, // <%--aspx comment--%>
				new {include = isHtml, regex = "(?<aspxHeader><%@[ ]*(?<tag>[\\w]+)(?<attr>[\\w\\W]*?)%>)"}, // <%@ Page Language="C#" AutoEventWireup="true" CodeBehind="WebForm2.aspx.cs" Inherits="FileUploadWeb.WebForm2" %>
				new {include = isHtml, regex = "(?<aspxCodeBlock><%[\\w\\W]*?%>)"}, // <% aspx code block %>
				new {include = true, regex = "(?<NodeSelfClosed>[<][ ]*(?<tagName>[\\w\\:\\-\\._]+)[ ]*(?<attr>[^<>]*?)/>)"}, // NodeSelfClosed has to be tested before StartTag bacause of /> at the end: < br />, <img .../>
				new {include = true, regex = "(?<StartTag>[<][ ]*(?<tagName>[\\w\\:\\-\\._]+)[ ]*(?<attr>[^<>]*?)>)"}, // <div>, ima nekih htmlova gdje je space iza <, znaci < div ...., ali to je invalid html
				new {include = true, regex = "(?<EndTag>[<]\\/[ ]*(?<tagName>[\\w\\:\\-\\._]+)[ ]*>)"}, // </div>
				new {include = true, regex = "(?<FreeText>(([^><]+)|([><]+?)))"} // free text, take all non <>, or if it is < or >, then take 1 by 1 and buffer it into freeTextBuffer
			}.Where(o => o.include == true).Select(o => o.regex).ToStringJoined("|");
		}

		private static readonly Regex mMainRegexXml = new Regex(GetRegexPattern(false), RegexOptions.IgnoreCase);

		private static readonly Regex mMainRegexHtml = new Regex(GetRegexPattern(true), RegexOptions.IgnoreCase);

		// atribute style: name = "value", name = 'value' name=value, x:Size="234" (xaml)
		// html5 compatible: <audio controls autoplay preload="none" myfiles=mojifileovi drugidio=1234 volume=0.5 raq=1.5>
		private static readonly Regex mAttributesRegex = new Regex("((?<name>[\\w\\:\\-\\._]+)[ ]*=[ ]*((\"(?<val>[^\"]*?)\")|('(?<val>[^']*?)')|((?<val>[^ ='\">]*))))|(?<name>[\\w\\:\\-\\._]+)", RegexOptions.IgnoreCase);

		private static readonly Regex mReStartTagNameOnly = new Regex("([<][\\/]*(?<tagName>[\\w\\:\\-\\._]+))", RegexOptions.IgnoreCase);  // tagname ili attribute name moze [^ >='\"\\/], ali je sigurnije [\\w\\:\\-\\.]

		// tags that can not contain other tags
		private static readonly Dictionary<string, bool> mHtmlSelfClosingTags = new string[]{
										"link",
										 "base",
										 "meta",
										 "isindex",
										 "hr",
										 "col",
										 "img",
										 "param",
										 "embed",
										 "frame",
										 "wbr",
										 "bgsound",
										 "spacer",
										 "keygen",
										 "area",
										 "input",
										 "basefont",
										 "br"
		}.ToDictionary(o => o.ToLower(), o=>true);

		#endregion

		private static bool IsSelfClosingHTMLTag(string tag)
		{
			return mHtmlSelfClosingTags.ContainsKey(tag.ToLower());
		}

		private string ExtractTagName(string html, SourceKindEnum kind)
		{
			Match m = mReStartTagNameOnly.Match(html);
			if (m.Groups["tagName"].Success)
			{
				return m.Groups["tagName"].Value;
			}
			else
			{
				return null;
			}
		}

		private void ExtractAttributes(string str, XmlNode node)
		{
			MatchCollection mc = XmlParser.mAttributesRegex.Matches(str);
			foreach (Match matt in mc)
			{
				// HTML5: attribute value may not be defined: <audio controls autoplay preload="none">....</audio>, set null for value and when building attributes, if value is null, use attribute name only
				node.Attributes.Add(new XmlAttribute(matt.Groups["name"].Value, matt.Groups["val"].Success ? matt.Groups["val"].Value : null));
			}
		}

		private XmlDocument ParseHtmlWorker(string html, SourceKindEnum kind)
		{
			XmlDocument document = new XmlDocument() { ParsingState = ParsingStateEnum.NodeClosed, SourceKind = kind};
			currentElement = document;

			// parse html into segments
			List<Segment> segments = (kind == SourceKindEnum.Html ? mMainRegexHtml : mMainRegexXml).Matches(html ?? "").Cast<Match>().Select(o => new Segment(o)).ToList();

			foreach(Segment seg in segments)
			{
				if(seg.IsSuccessful(SegmentKind.FreeText))
				{
					AddFreeTextToBuffer(seg.Text);
					continue;
				}

				FlushFreeTextBuffer();

				if(seg.IsSuccessful(SegmentKind.cdataNode))
				{
					AddTextNode(seg.Text, XmlNodeType.Cdata);
				}
				else if (seg.IsSuccessful(SegmentKind.phpCodeBlock))
				{
					AddTextNode(seg.Text, XmlNodeType.PhpCodeBlock);
				}
				else if (seg.IsSuccessful(SegmentKind.xmlHeader))
				{
					////AddTextNode(m.Value, XmlNodeType.XmlHeader);
					XmlNode node = new XmlNode();
					node.NodeType = XmlNodeType.XmlHeader;
					node.ParsingState = ParsingStateEnum.NodeSelfClosed;
					ExtractAttributes(seg.GetGroupValue(SegmentKind.attr), node);
					AddChildNode(currentElement, node);
				}
				else if (seg.IsSuccessful(SegmentKind.htmlComment))
				{
					AddTextNode(seg.Text, XmlNodeType.Comment);
				}
				else if (seg.IsSuccessful(SegmentKind.aspxComment))
				{
					AddTextNode(seg.Text, XmlNodeType.AspxComment);
				}
				else if (seg.IsSuccessful(SegmentKind.htmlDoctype))
				{
					AddTextNode(seg.Text, XmlNodeType.Doctype);
				}
				else if (seg.IsSuccessful(SegmentKind.aspxHeader))
				{
					////AddTextNode(m.Value, XmlNodeType.AspxHeader);
					XmlNode node = new XmlNode();
					node.TagName = seg.GetGroupValue(SegmentKind.tag).Trim();
					node.NodeType = XmlNodeType.AspxHeader;
					node.ParsingState = ParsingStateEnum.NodeSelfClosed;
					ExtractAttributes(seg.GetGroupValue(SegmentKind.attr), node);
					AddChildNode(currentElement, node);
				}
				else if (seg.IsSuccessful(SegmentKind.aspxCodeBlock))
				{
					AddTextNode(seg.Text, XmlNodeType.AspxCodeBlock);
				}
				else if (seg.IsSuccessful(SegmentKind.scriptNode) || seg.IsSuccessful(SegmentKind.styleNode))
				{
					XmlNode node = new XmlNode();
					node.TagName = ExtractTagName(seg.Text, kind);
					node.ParsingState = ParsingStateEnum.NodeClosed;
					ExtractAttributes(seg.GetGroupValue(SegmentKind.attr), node);
					AddChildNode(currentElement, node);

					// script node can have text node as child only
					// script may contain html code as string, we don't want to parse it
					string content = seg.GetGroupValue(SegmentKind.content);
					if (!string.IsNullOrEmpty(content))
					{
						XmlNode scriptNode = new XmlNode();
						scriptNode.NodeType = seg.IsSuccessful(SegmentKind.scriptNode) ? XmlNodeType.Script : XmlNodeType.Style;
						scriptNode.NodeValue = content;
						scriptNode.ParsingState = ParsingStateEnum.NodeClosed;
						AddChildNode(node, scriptNode); // add script content as child to node
					}
				}
				else if (seg.IsSuccessful(SegmentKind.NodeSelfClosed))
				{
					XmlNode node = new XmlNode();
					node.NodeType = XmlNodeType.Xml;
					node.ParsingState = ParsingStateEnum.NodeSelfClosed;
					node.TagName = ExtractTagName(seg.Text, kind);
					ExtractAttributes(seg.GetGroupValue(SegmentKind.attr), node);
					AddChildNode(currentElement, node);
				}
				else if (seg.IsSuccessful(SegmentKind.StartTag))
				{
					XmlNode node = new XmlNode();
					node.NodeType = XmlNodeType.Xml;
					node.ParsingState = ParsingStateEnum.StartTag;
					node.TagName = ExtractTagName(seg.Text, kind);
					ExtractAttributes(seg.GetGroupValue(SegmentKind.attr), node);
					AddChildNode(currentElement, node);
					if (kind == SourceKindEnum.Html && IsSelfClosingHTMLTag(node.TagName))
					{
						// this closes nodes like <br> or <img...> which should be self closed
						node.ParsingState = ParsingStateEnum.NodeSelfClosed;
					}
					else
					{
						currentElement = node; // expect children, but only if node is not self closed
					}
				}
				else if (seg.IsSuccessful(SegmentKind.EndTag))
				{
					string tagName = ExtractTagName(seg.Text, kind);

					// for fixing broken html, but not implemented
					XmlNode endNode = new XmlNode() { TagName = tagName, ParsingState = ParsingStateEnum.EndTag, Level = currentElement.Level };
					allNodes.Add(endNode);

					if (kind == SourceKindEnum.Html && IsSelfClosingHTMLTag(tagName))
					{
						// ignore closing tag if node is self-closed already, </img> or </meta>
					}
					else
					{
						if (kind != SourceKindEnum.Html && tagName != currentElement.TagName)
						{
							currentElement.ParsingState = ParsingStateEnum.InvalidEndTag;
						}
						else if (kind == SourceKindEnum.Html && !tagName.Equals(currentElement.TagName, StringComparison.OrdinalIgnoreCase))
						{
							// html tag has to be case nonsensitive
							
							// this is simple and quick fix for broken html, if current node doesn't match, try to match with parent node (required for Freemeteo.com, 7-days forecast)
							// npr. ako se zatvori parent td, automatski se zatvara i div i span: <td><div><span>fdsgsd</td>, closing div and span is missing, but consider div and span closed since parent td is closed, this is standard
							// test all ancestors up to the root
							XmlNode ancestor = currentElement.AncestorNodes().Where(o => tagName.Equals(o.TagName, StringComparison.OrdinalIgnoreCase)).FirstOrDefault();
							if (ancestor != null)
							{
								currentElement = ancestor;
								currentElement.ParsingState = ParsingStateEnum.NodeClosed;
							}
							else
							{
								// mark node as invalid, but don't throw exeption
								currentElement.ParsingState = ParsingStateEnum.InvalidEndTag;
							}
						}
						else
						{
							currentElement.ParsingState = ParsingStateEnum.NodeClosed;
						}

						if (currentElement.Parent != null)
						{
							// if parent is null, document is broken, add other nodes to current node
							currentElement = currentElement.Parent; // no children are expected any more to currentElement, move up level
						}
					}
				}
			}

			FlushFreeTextBuffer();

			return document;
		}

		#endregion

		#region parse JSON

		private void ProcessJsonNode(object obj)
		{
			if (obj is string)
			{
				// string is IEnumerable, that is why we have to test firs if it is string
				AddTextNode(obj + "", XmlNodeType.Text, obj);
			}
			else if (obj is IEnumerable)
			{
				foreach (object cobj in obj as IEnumerable)
				{
					XmlNode node = new XmlNode();
					node.DataItem = cobj;
					AddChildNode(currentElement, node);
					node.NodeType = XmlNodeType.JSObject;
					if (string.IsNullOrEmpty(node.TagName))
					{
						node.TagName = "js:object";
					}
					node.ParsingState = ParsingStateEnum.NodeClosed;
					currentElement = node;
					ProcessJsonNode(cobj);
					currentElement = node.Parent;
				}
			}
			else if (obj is KeyValuePair<string, object>)
			{
				KeyValuePair<string, object> kv = ((KeyValuePair<string, object>)obj);
				currentElement.DataItem = obj;
				currentElement.NodeType = XmlNodeType.JSObject;
				currentElement.TagName = kv.Key;
				ProcessJsonNode(kv.Value);
			}
			else
			{
				// just a value
				AddTextNode(obj + "", XmlNodeType.Text, obj);
			}
		}

		private XmlDocument ParseJsonWorker(string json)
		{
			object obj = Ice.Serialization.Json.IceJson.DeserializeObject(json);
			// tree is dictionary tree output from JavaScriptSerializer.DeserializeObject()
			XmlDocument document = new XmlDocument() { ParsingState = ParsingStateEnum.NodeClosed, SourceKind = SourceKindEnum.Json};
			currentElement = document;
			ProcessJsonNode(obj);
			return document;
		}

		#endregion

		#region parse entry point

		public static XmlDocument Parse(string str, SourceKindEnum kind = SourceKindEnum.Html)
		{
			XmlParser parser = new XmlParser();

			switch (kind)
			{
				case SourceKindEnum.Html:
				case SourceKindEnum.Xml:
					return parser.ParseHtmlWorker(str, kind);
				case SourceKindEnum.Json:
					return parser.ParseJsonWorker(str);
				default:
					throw new NotImplementedException();
			}
		}

		#endregion

#if DEBUG && !SILVERLIGHT
		static XmlParser()
		{
			Ice.Xml.Parser.SelfTextEngine.SelfTest.RunSelfTest();
		}
#endif


	}
}
