﻿#if !SILVERLIGHT

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.Data.Linq.Mapping;
using System.Data.SqlClient;
using System.Text.RegularExpressions;

namespace System.Data.Linq
{
	public enum SaveModeEnum
	{
		Update,
		Insert,
		IdentityInsert
	}

	public static class DataContextExtensions
	{
		public static DataSet ExecuteQueryToDataSet(this System.Data.Linq.DataContext db, string query, params object[] parameters)
		{
			db.CommandTimeout = db.Connection.ConnectionTimeout; // ConnectionTimeout is defined in sql connection string

			SqlConnection conn = (SqlConnection)db.Connection;
			SqlCommand cmd = conn.CreateCommand();
			cmd.Connection = conn;
			if (db.Transaction != null)
			{
				cmd.Transaction = (SqlTransaction)db.Transaction;
			}
			cmd.CommandTimeout = db.CommandTimeout;
			int uniqueNum = 0;
			cmd.CommandText = new Regex("\\{([\\d]+)\\}").Replace(query, delegate(Match m)
			{
				int index = Convert.ToInt32(m.Groups[1].Value);
				if (parameters[index] == null || parameters[index] == DBNull.Value)
				{
					return "NULL";
				}
				else
				{
					string name = "@p" + index + "_" + (uniqueNum++);
					cmd.Parameters.AddWithValue(name, parameters[index]);
					return name;
				}
			});

			SqlDataAdapter da = new SqlDataAdapter(cmd);
			DataSet ds = new DataSet();

			bool shouldCloseConnection = false;

			if (conn.State != ConnectionState.Open)
			{
				shouldCloseConnection = true;
				conn.Open();
			}
	
			da.Fill(ds);

			if (shouldCloseConnection)
			{
				conn.Close();
			}
			return ds;
		}

		public static void SaveToDB(this System.Data.Linq.DataContext db, object obj, SaveModeEnum saveMode)
		{
			var properties = obj.GetType().GetProperties().Select(o => new
			{
				Property = o,
				Mapping = o.GetCustomAttributes(typeof(ColumnAttribute), true).Cast<ColumnAttribute>().FirstOrDefault()
			})
			.Where(o => o.Mapping != null)
			.Select(o=>new {
				o.Property,
				o.Mapping,
				Value = o.Property.GetValue(obj, null)
			})
			.ToList();

			string tableName = obj.GetType().Name;

			int index = 0;
			var func = new Func<object, string>(delegate(object val)
			{
				return val == null ? "NULL" : ("{" + (index++) + "}");
			});

			if (saveMode == SaveModeEnum.Update)
			{
				var reference = properties.Where(o => o.Mapping.IsDbGenerated || o.Mapping.IsPrimaryKey).ToList();
				var values = properties.Where(o => !reference.Contains(o)).ToList();
				db.ExecuteCommand("update [" + tableName + "] set " + (values.Select((o) => "[" + o.Property.Name + "] = " + func(o.Value)).ToStringJoined(", ")) + " where " +
					(reference.Select(o => "[" + o.Property.Name + "] = " + func(o.Value)).ToStringJoined(" and ")),
					values.Concat(reference).Where(o=>o.Value != null).Select(o => o.Value).ToArray());
			}
			else // insert
			{
				var values = (saveMode == SaveModeEnum.IdentityInsert ? properties : properties.Where(o => !o.Mapping.IsDbGenerated).ToList());

				db.ExecuteCommand(
					(saveMode == SaveModeEnum.IdentityInsert ? ("SET IDENTITY_INSERT [" + tableName + "] ON\r\n") : null) +
					"insert into [" + tableName + "] (" + values.Select(o => "[" + o.Property.Name + "]").ToStringJoined(", ") + ")\r\nVALUES (" + values.Select(o => func(o.Value)).ToStringJoined(", ") + ")\r\n" +
					(saveMode == SaveModeEnum.IdentityInsert ? ("SET IDENTITY_INSERT [" + tableName + "] OFF") : null),
					values.Where(o=>o.Value != null).Select(o => o.Value).ToArray());
			}
		}

		/// <summary>
		/// Start transaction. If action returns true, transaction will be commited, else it will rollback. 
		/// </summary>
		/// <param name="db"></param>
		/// <param name="action"></param>
		/// <param name="alwaysRollback"></param>
		public static void UseTransaction(this DataContext db, Func<bool> action, bool alwaysRollback = false)
		{
			bool closeConnection = false;

			try
			{
				if (db.Connection.State != System.Data.ConnectionState.Open)
				{
					closeConnection = true;
					db.Connection.Open();
				}

				db.Transaction = db.Connection.BeginTransaction();

				bool success = action.Invoke();

				if (alwaysRollback || !success)
				{
					db.Transaction.Rollback();
				}
				else
				{
					db.Transaction.Commit();
				}

				db.Transaction = null;

				if (closeConnection)
				{
					db.Connection.Close();
				}

			}
			catch (Exception ex)
			{
				// action.Invoke() failed or Commit() failed
				if (db.Transaction != null)
				{
					try
					{
						db.Transaction.Rollback(); // rollback may fail too
					}
					catch
					{
					}
					db.Transaction = null;
				}

				if (closeConnection)
				{
					db.Connection.Close();
				}

				throw ex;
			}

		}
	}
}

#endif
