﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.Linq;
using System.Collections;

namespace System
{
	public static class StringExtensions
	{
		/// <summary>
		/// Capitalize first char of the string, all other char's case is not changed.
		/// </summary>
		/// <param name="s"></param>
		/// <returns></returns>
		public static string Capitalize(this string s)
		{
			if (string.IsNullOrEmpty(s))
			{
				return s;
			}
			else
			{
				return s[0].ToString().ToUpper() + s.Substring(1);
			}
		}

		public static string ToNullIfEmpty(this string s)
		{
			if(string.IsNullOrEmpty(s))
			{
				return null;
			}
			return s;
		}

		/// <summary>
		/// Take first n chars from string or the whole string if n is larger than length of the string. If string is shorten, add elipsis to the end of the new string.
		/// </summary>
		/// <param name="s"></param>
		/// <param name="n"></param>
		/// <returns></returns>
		public static string TakeFirst(this string s, int n, string elipsis = null)
		{
			if (string.IsNullOrEmpty(s) || n >= s.Length)
			{
				return s;
			}
			else
			{
				int len = (elipsis ?? "").Length;
				return s.Substring(0, n - len) + elipsis;
			}
		}

		/// <summary>
		/// Take last n chars from string or the whole string if n is larger than length of the string. If string is shorten, add elipsis to the start of the new string.
		/// </summary>
		/// <param name="s"></param>
		/// <param name="n"></param>
		/// <returns></returns>
		public static string TakeLast(this string s, int n, string elipsis = null)
		{
			if (string.IsNullOrEmpty(s) || n >= s.Length)
			{
				return s;
			}
			else
			{
				int len = (elipsis ?? "").Length;
				return elipsis + s.Substring(s.Length - n + len, n-len);
			}
		}

		/// <summary>
		/// Remove first n chars from string or the whole string if n is larger than length of the string. If string is shorten, add elipsis to the start of the new string.
		/// </summary>
		/// <param name="s"></param>
		/// <param name="n"></param>
		/// <param name="elipsis"></param>
		/// <returns></returns>
		public static string RemoveFirst(this string s, int n, string elipsis = null)
		{
			if (string.IsNullOrEmpty(s) || n >= s.Length)
			{
				return "";
			}
			else
			{
				int len = (elipsis ?? "").Length;
				return elipsis + s.Remove(n + len);
			}
		}

		/// <summary>
		/// Remove last n chars from string or the whole string if n is larger than length of the string. If string is shorten, add elipsis to the end of the new string.
		/// </summary>
		/// <param name="s"></param>
		/// <param name="n"></param>
		/// <param name="elipsis"></param>
		/// <returns></returns>
		public static string RemoveLast(this string s, int n, string elipsis = null)
		{
			if (string.IsNullOrEmpty(s) || n >= s.Length)
			{
				return "";
			}
			else
			{
				int len = (elipsis ?? "").Length;
				return s.Remove(s.Length - n - len) + elipsis;
			}
		}

		/// <summary>
		/// Replace new line with br
		/// </summary>
		/// <param name="s"></param>
		/// <returns></returns>
		public static string ToHtml(this string s)
		{
			if (string.IsNullOrEmpty(s))
			{
				return s;
			}
			else
			{
				return s.Replace("\r\n", "\n").Replace("\r", "").Replace("\n", "<br/>");
			}
		}

		public static string Repeat(this string str, int count)
		{
			StringBuilder sb = new StringBuilder();
			for (int n = 0; n < count; n++)
			{
				sb.Append(str);
			}
			return sb.ToString();
		}

		public static TResult ToNumber<TResult>(this string str, bool last = false)
		{
			if (string.IsNullOrEmpty(str))
			{
				return default(TResult);
			}

			string s;
			var query = Regex.Matches(str, "[\\- ]*[\\d]+[\\d\\.\\,]*").Cast<Match>().Select(o => o.Value); // including space too
			if (last)
			{
				s = query.LastOrDefault();
			}
			else
			{
				s = query.FirstOrDefault();
			}

			if (!string.IsNullOrEmpty(s))
			{
				// ukloniti space izmedju - i brojke jer Decimal.TryParse ne radi s razmakom
				s = s.Replace(" ", "");
			}

			TResult res;
			Ice.Utility.IceTypeConverter.TryChangeType<TResult>(s, out res);
			return res;
		}

		/// <summary>
		/// if ignoreCase == true: case non-sensitive String.Contains
		/// </summary>
		/// <param name="s"></param>
		/// <param name="str"></param>
		/// <returns></returns>
		public static bool Contains(this string s, string str, bool ignoreCase)
		{
			if(s == str)
			{
				return true;
			}
			else if (string.IsNullOrEmpty(s) || string.IsNullOrEmpty(str))
			{
				return false;
			}
			else if (ignoreCase)
			{
				return s.IndexOf(str, StringComparison.OrdinalIgnoreCase) >= 0;
			}
			else
			{
				return s.Contains(str); // much faster than .IndexOf(..)
			}
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="str">input string with property names or array indexes: My name is {Name}, or My name is {0}</param>
		/// <param name="obj">
		///		<para>object with properties {Name:Tom, Age:22, ...}</para>
		///		<para>or Enuerable or array["Tom", 22"...]</para>
		///		<para>IDictionary with Key-Values</para>
		/// </param>
		/// <param name="placeHolderStart"></param>
		/// <param name="placeHolderEnd"></param>
		/// <returns></returns>
		public static string FormatWith(this string str, object obj, string placeHolderStart = "{", string placeHolderEnd = "}")
		{
			if (string.IsNullOrEmpty(str) || obj == null)
			{
				return str;
			}

			if (placeHolderStart.Length > 1)
			{
				placeHolderStart = placeHolderStart.Select(o => "\\" + o).ToStringJoined();
			}
			else
			{
				placeHolderStart = "\\" + placeHolderStart;
			}

			if (placeHolderEnd.Length > 0)
			{
				placeHolderEnd = placeHolderEnd.Select(o => "\\" + o).ToStringJoined();
			}
			else
			{
				placeHolderEnd = "\\" + placeHolderEnd;
			}

			return new Regex("(" + placeHolderStart + "([\\w_\\-\\.]+)" + placeHolderEnd + ")").Replace(str, delegate(Match m)
			{
				string key = m.Groups[2].Value;

				return Ice.Web.UI.DataBinder.Eval(obj,key) + ""; // works in silverlight
				////return System.Web.UI.DataBinder.Eval(obj, key) + ""; // doesn't work in Silverlight

			});


		}

	}
}
