﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.IO;
using System.Text.RegularExpressions;

namespace Ice.Web
{
	public class IceHtmlMinimizerFilter : MemoryStream
	{
		private MemoryStream mBuffer = null;

		public Stream BaseStream { get; set; } 

		public IceHtmlMinimizerFilter(Stream stream)
		{
			this.BaseStream = stream;
		}

		public override void Close()
		{
			this.Flush();
			base.Close();
		}

		public override void Flush()
		{
			this.FlushBufferedData();
			this.BaseStream.Flush();
		}

		public override void Write(byte[] buffer, int offset, int count)
		{
			// write se poziva vise puta, moram prvo zapisati sve u buffer, a na Flush tek kreirati response
			if (buffer != null)
			{
				if (mBuffer == null)
				{
					mBuffer = new MemoryStream();
				}

				mBuffer.Write(buffer, offset, count);
			}
		}

		static Regex reJsComments = new Regex("(/\\*[\\w\\W]*?\\*/)|(//[^\\r\\n]*[\\r\\n]+)");
		static Regex reCssComments = new Regex("(/\\*[\\w\\W]*?\\*/)");

		static Regex reScripts = new Regex("(<script[\\w\\W]*?</script>)|(<style[\\w\\W]*?</style>)");

		private void FlushBufferedData()
		{
			if (mBuffer != null && mBuffer.Length > 0)
			{
				StreamWriter writer = new StreamWriter(this.BaseStream, System.Text.Encoding.UTF8);

				string str = System.Text.Encoding.UTF8.GetString(mBuffer.ToArray());
				mBuffer = null;

				// remove line comments in script or css
				str = reScripts.Replace(str, delegate(Match m)
				{
					if (m.Groups[1].Success)
					{
						str = reJsComments.Replace(m.Value, " ");
					}
					else
					{
						str = reCssComments.Replace(m.Value, " ");
					}

					// compress xml
					str = Regex.Replace(str, "[\\s]+", " ");

					// remove space near to symbols
					str = Regex.Replace(str, @"[ ]*([\(\)\;\{\}\[\]\:\,\=\-\+\*\/\<\>]+)[ ]*", "$1");

					return str;

				});

				// compress xml
				str = Regex.Replace(str, "[\\s]+", " ");

				str = Regex.Replace(str, ">[ ]+<", "><");
				str = Regex.Replace(str, "[ ]+/>", "/>");
				str = Regex.Replace(str, ">[ ]+", ">");
				str = Regex.Replace(str, "[ ]+<", "<");

				writer.Write(str + "\r\n"); // must terminate line
				writer.Flush();

			}
		}

		const string __noMinimizeFilter = "__noMinimizeFilter";
		const string __minimizeFilterUsed = "__minimizeFilterUsed";

		public static void AddHtmlMinimizerFilter(HttpContext context)
		{
			if (context == null || context.Response == null || context.Response.IsRequestBeingRedirected)
			{
				return;
			}

			if (context.Items.Contains(__noMinimizeFilter) || context.Items.Contains(__minimizeFilterUsed))
			{
				return;
			}

			context.Items[__minimizeFilterUsed] = true;

			string contentType = context.Response.ContentType.ToLower();
			string pathInfo = HttpContext.Current.Request.ServerVariables["PATH_INFO"];
			if (pathInfo.EndsWith(".aspx") && (contentType.Contains("html")))
			{
				context.Response.Filter = new IceHtmlMinimizerFilter(context.Response.Filter);
			}
		}

		public static void RemoveHtmlMinizerFilter(HttpContext context)
		{
			if (context == null || context.Response == null)
			{
				return;
			}

			context.Items[__noMinimizeFilter] = true;

			while (context.Response.Filter is IceHtmlMinimizerFilter)
			{
				context.Response.Filter = (context.Response.Filter as IceHtmlMinimizerFilter).BaseStream;
			}
		}

	}
}

