﻿/// <reference path="Ice.js" />
/// <reference path="Ice.DomQuery.js" />

Ice.Animator = function ()
{
	/// <summary>
	/// Constructor. create new animated task
	/// </summary>

	this._config(); // create new config
};

Ice.Animator.__class = true;

Ice.Animator.prototype = new (function ()
{
	this._typeName = "Ice.Animator";

	this._config = function ()
	{
		this._config =
		{
			startValue: null,
			finalValue: null, // numeric value
			durationMs: 400, // total duration, default predefined
			stepMs: 20, // interval between ticks, default predefined (20 ms = 50fps) // Firefox has problem with > 30 fps
			onComplete: null, // function(sender){...}
			onStart: null, // function (sender){...}
			onProgress: null, // function(sender, percent){...}
			onAbort: null, //function(sender){...}
			easing: Ice.Animator.Easing.linear
		};

		this._instance =
		{
			timeoutId: null,
			elapsedTimeMs: 0
		};

		this._items = [];
	};

	this.Items = function ()
	{
		/// <summary>
		/// return Ice.List of this._items
		/// </summary>

		return new Ice.List(this._items);
	};

	this.ToDomQuery = function ()
	{
		/// <summary>
		/// Returns distinct dom elements.
		/// </summary>
		return new Ice.DomQuery(this._items).Select(function(o){return o.element;}).Where(function(o){return !!o;}).Distinct().ToDomQuery();
	};

	this.AddElement = function (element, styleName, finalValue, onTick)
	{
		/// <summary>
		/// Push DOM element with final value setting. Animation will run from current state to final setting.
		/// </summary>
		/// <param name="element"></param>
		/// <param name="styleName"></param>
		/// <param name="finalValue">can be exact value or eg. "+=10", "-=50", "*=34"</param>
		/// <param name="onTick" optional="true" type="Function">function(animator, currentValue){...}, this is animator instance</param>

		//#if VSINTELLISENSE
		if (Ice.IsVSIntellisene) { onTick.call(this, this, 1); }
		//#endif

		this._items.push({ element: element, name: styleName, finalValue: finalValue, onTick: onTick });
		return this;
	};

	this.AddValue = function (startValue, finalValue, onTick)
	{
		/// <summary>
		/// Animate general value from startValue to finalValue.
		/// </summary>
		/// <param name="startValue"></param>
		/// <param name="finalValue"></param>
		/// <param name="onTick" optional="true" type="Function">function(animator, currentValue){...}, this is animator instance</param>

		//#if VSINTELLISENSE
		if (Ice.IsVSIntellisene) { onTick.call(this, this, 1); }
		//#endif

		this._items.push({ isValue: true, startValue: startValue, finalValue: finalValue, onTick: onTick });
		return this;
	},

	this.Clear = function ()
	{
		/// <summary>
		/// remove all this._items
		/// </summary>
		
		this._items = [];
		return this;
	},

	this.Duration = function (durationMs, stepMs)
	{
		/// <summary>
		/// total duration of animation. Default is 400ms.
		/// </summary>
		/// <param name="durationMs">default is 400ms</param>
		/// <param name="stepMs" optional="true">default is 20ms</param>
		if (!((durationMs || 0) > 0))
		{
			durationMs = 1; // duration must be > 0, or easing functions won't work
		}
		this._config.durationMs = durationMs;
		if (arguments.length == 2)
		{
			this._config.stepMs = stepMs > 0 ? stepMs : 20;
		}
		return this;
	};

	this.Step = function (stepMs)
	{
		/// <summary>
		/// step in ms. Default is 20ms (or 50fps).
		/// </summary>
		/// <param name="stepMs">ms value</param>
		this._config.stepMs = stepMs > 0 ? stepMs : 20;
		return this;
	};
	
	this.FPS = function (framesPerSecond)
	{
		/// <summary>
		/// Calculate step based on frames per second. Eg. 25 fps, step = 40 ms
		/// </summary>
		/// <param name="framesPerSecond"></param>
		
		return this.Step(Math.ceil(1000 / framesPerSecond));
	};

	this.OnProgress = function (callback)
	{
		/// <summary>
		/// send progres in percentage
		/// </summary>
		/// <param name="callback">function(animator, percent){  }, this is animator instance</param>

		//#if VSINTELLISENSE
		if (Ice.IsVSIntellisene){callback.call(this, this, 1);}
		//#endif

		this._config.onProgress = callback;
		return this;
	};

	this.OnStart = function (callback)
	{
		/// <summary>
		/// Send callbak on start.
		/// </summary>
		/// <param name="callback">function(animator){...}, this is animator instance</param>

		//#if VSINTELLISENSE
		if (Ice.IsVSIntellisene) { callback.call(this, this); }
		//#endif

		this._config.onStart = callback;
		return this;
	};

	this.OnComplete = function (callback)
	{
		/// <summary>
		/// Send callback on complete.
		/// </summary>
		/// <param name="callback">function(animator){...}, this is animator instance</param>

		//#if VSINTELLISENSE
		if (Ice.IsVSIntellisene) { callback.call(this, this); }
		//#endif

		this._config.onComplete = callback;
		return this;
	};

	this.OnAbort = function (callback)
	{
		/// <summary>
		/// Send callback on complete.
		/// </summary>
		/// <param name="callback">function(animator){...}, this is animator instance</param>

		//#if VSINTELLISENSE
		if (Ice.IsVSIntellisene) { callback.call(this, this); }
		//#endif

		this._config.onAbort = callback;
		return this;
	};

	this.Easing = function (easingType)
	{
		/// <summary>
		/// set easing method
		/// </summary>
		/// <param name="easingType" type="Ice.Animator.Easing">easing method. If not defined, default linear will be used.</param>
		if (easingType)
		{
			this._config.easing = easingType;
		}
		else
		{
			this._config.easing = Ice.Animator.Easing.linear;
		}
		return this;
	};

	function GetCurrent(startValue, finalValue)
	{
		return this._config.easing.call(this,
			this._instance.elapsedTimeMs / this._config.durationMs, // progress 0-1
			this._instance.elapsedTimeMs,
			startValue,
			finalValue - startValue,
			this._config.durationMs);
	}

	function TimerTick(firstRun)
	{

		if (!this._allowNextStep)
		{
			return; // finish it
		}

		var ticksAtFuncStart = Date.now();

		if (firstRun)
		{
			this._instance.animateStartTicks = ticksAtFuncStart;
		}

		this._instance.elapsedTimeMs = ticksAtFuncStart - this._instance.animateStartTicks;

		var timeLeftMs = this._config.durationMs - this._instance.elapsedTimeMs;
		var isOver = timeLeftMs <= 0;
		

		// first remove items with element property and with no name property (created from domquery with .ToAnimator() with no final values set in .ToAnimator(...final values...)
		this._items = this.Items().Where(function (o) { return o && (o.name || !o.element); }).ToArray();

		for (var nn = 0; nn < this._items.length; nn++)
		{
			var item = this._items[nn];


			if (firstRun)
			{
				if (item.element && !Ice.Type.IsNullOrUndefined(item.finalValue))
				{
					var styleItem;
					if (item.name.Contains("color", true))
					{
						// eg. backgroundColor
						// IE <= v8 returns named colors
						var vvv = Ice.Utility.Dom.GetComputedStyleValue(item.element, item.name);
						item.startValue = Ice.Colorspace.Parse(vvv);
						var dd = document.createElement("div");
						dd.style.position = "absolute";
						dd.style.width = "1px";
						dd.style.height = "1px";
						dd.style.zIndex = -1000;
						document.body.appendChild(dd); // must be in dom to be able to get computed style
						dd.style.backgroundColor = item.finalValue;
						var ccc = Ice.Utility.Dom.GetComputedStyleValue(dd, "backgroundColor");
						dd.parentNode.removeChild(dd);
						item.finalValue = Ice.Colorspace.Parse(ccc);
						////console.log("start: " + item.startValue.ToHEX() + ", finalValue: " + item.finalValue.ToHEX());
						item.isRGB = true; // animate RGB, HSV is not ok
						//item.isHSV = true;
						item.currentValue = item.startValue.Clone();
					}
					else if(item.name == "clip")
					{
						var curr1 = Ice.Utility.Dom.GetClip(item.element);
						item.startValue = curr1; // {top, right, bottom, left}
						item.currentValue = Ice.Extend({}, curr1); // clone to avoid messing with startValue object
						item.finalValue = Ice.Utility.Dom.CalculateClip(item.element, item.finalValue);
						////console.log(item.startValue);
						////console.log(item.finalValue);
						item.isClip = true;
					}
					else if(item.name == "transform")
					{
						// if current transform style is coumpound;
						// translate(100px, -100px) rotate(170deg) scale(0.2)"
						
						var currList = ($$ice(item.element).SelectTransform().FirstOrDefault() || "").split(")").AsEnumerable().Where("o=>o").Select(function (o) { return o.Trim() + ")"; }).Select(function (o)
						{
							return {
								statement: o, // eg. translate(100px, -100px)
								kind: o.match(new RegExp("[\\w]+"))[0], // eg. translate
								args: new Ice.List((o || "").match(new RegExp("([\\-]?[\\.\\d]+)", "ig"))).Select(function (kk) { return parseFloat(kk);}).ToArray() // arguments: 100, -100
							}
						}).ToList();

						var finalList = item.finalValue.split(")").AsEnumerable().Where("o=>o").Select(function (o) { return o.Trim() + ")"; }).Select(function (o)
						{
							return {
								statement: o, // eg. translate(100px, -100px)
								kind: o.match(new RegExp("[\\w]+"))[0], // eg. translate
								args: new Ice.List((o || "").match(new RegExp("([\\-]?[\\.\\d]+)", "ig"))).Select(function (kk) { return parseFloat(kk); }).ToArray()
							}
						}).ToList();

						// select transforms which are to be animated
						// start and finalVal args count has to be the same, if arguments are missing, add them
						var mainList = finalList.Select(function (finalVal)
						{
							var curr = currList.Where(function (cc) { return cc.kind == finalVal.kind; }).FirstOrDefault();
							if (!curr)
							{
								curr = {
									statement: finalVal.statement,
									kind: finalVal.kind,
									args: finalVal.args.AsEnumerable().Select(function (ff) { return finalVal.kind.indexOf("scale") == 0 ? 1 : 0; }).ToArray()
								};
							}

							// ako finalVal ima vise argumenata, uzeti statement as stringFormat iz finalValue, u protivnom, uzeti iz currValue
							var finalHasMoreArgs = false;

							while (curr.args.length < finalVal.args.length)
							{
								finalHasMoreArgs = true;
								curr.args.push(curr.args.length > 0 ? curr.args[curr.args.length - 1] : finalVal.kind.indexOf("scale") == 0 ? 1 : 0);
							}

							while (curr.args.length > finalVal.args.length)
							{
								finalVal.args.push(finalVal.args.length > 0 ? finalVal.args[finalVal.args.length - 1] : finalVal.kind.indexOf("scale") == 0 ? 1 : 0);
							}

							return {
								startValuesArr: curr.args,
								finalValuesArr: finalVal.args,
								stringFormat: (finalHasMoreArgs ? finalVal.statement : curr.statement)
							};

						}).ToList();

						var index = 0;
						// create single item:
						item.startValue = mainList.SelectMany("o=>o.startValuesArr").ToArray();
						item.currentValue = item.startValue.concat(); // clone array with .concat()////.SelectMany("o=>o.startValuesArr").ToArray(); // array has to be cloned to avoid messing with startValue array
						item.finalValue = mainList.SelectMany("o=>o.finalValuesArr").ToArray();
						item.stringFormat = mainList.Select("o=>o.stringFormat").ToStringJoined(" ").replace(new RegExp("([\\-]?[\\.\\d]+)", "ig"), function () { return "{" + (index++) + "}"; });
						item.isTransform = true;
						item.mainList = mainList;
					}
					else
					{
						// width or height always take from clientWidth or clientHeight
						switch (item.name)
						{
							case "height":
								styleItem = Ice.Utility.Dom.GetStyleItem(item.element, "clientHeight");
								break;
							case "width":
								styleItem = Ice.Utility.Dom.GetStyleItem(item.element, "clientWidth");
								break;
							default:
								styleItem = Ice.Utility.Dom.GetStyleItem(item.element, item.name);
								break;
						}

						item.isValue = true;

						item.startValue = styleItem.NumericValue;

						var originalValue;

						if (!item.units)
						{
							//item.units = styleItem.Units;
							// take units from finalValue, eg. : 100%
							if(typeof(item.finalValue) == "string" && item.finalValue.indexOf("%") > 0)
							{
								item.units = "%";
							}
							else
							{
								item.units = styleItem.Units;
							}
						}

						if (item.name.indexOf("margin") == 0)
						{
							// we want to set margin in percent of elmeent dimensions, eg. margin-top: 25%, we'll set top margin: 25% of element.offsetHeight
							// margin is to be animated and finalValue is in percent, use element offset dimensions to calculate final value
							item.units = null;

							switch (Ice.Utility.Dom.ConvertCssNameToCamel(item.name))
							{
								case "marginTop":
								case "marginBottom":
									originalValue = item.element.offsetHeight;
									break;
								case "marginLeft":
								case "marginRight":
									originalValue = item.element.offsetWidth;
									break;
							}
						}

						// finalValue may be number, "+=10", "100px"

						item.finalValue = Ice.Utility.Dom.CalculateNumericValue(styleItem.NumericValue, item.finalValue, originalValue);

					}
				}
				else if(item.isValue)
				{
					// value item (not element)
				}


				// common code: for element animation (excluding color) and for simple value animation
				if (item.isValue)
				{
					item.delta = item.finalValue - item.startValue;
					item.currentValue = item.startValue;
				}

				if ((item.isValue && (typeof (item.startValue) != "number" || typeof (item.finalValue) != "number"))
					|| (item.element && (Ice.Type.IsNullOrUndefined(item.finalValue) || Ice.Type.IsNullOrUndefined(item.startValue))))
				{
					Ice.Diagnostic.Exception("Ice.Animator error, invalid start/stop value. Name: " + item.name + ", startValue: " + item.startValue + ", finalValue: " + item.finalValue);
				}

			}
			else if(item.isHSV)
			{
				Ice.ForEach(["h", "s", "v"], function (o) { item.currentValue[o] = isOver ? item.finalValue[o] : Math.floor(GetCurrent.call(this, item.startValue[o], item.finalValue[o])) }, this);
				// apply to element
				Ice.Utility.Dom.SetStyleValue(item.element, item.name, item.currentValue.ToHEX(), null);
			}
			else if(item.isRGB)
			{
				Ice.ForEach(["r", "g", "b", "op"], function (o) { item.currentValue[o] = isOver ? item.finalValue[o] : Math.floor(GetCurrent.call(this, item.startValue[o], item.finalValue[o])) }, this);
				// apply to element
				Ice.Utility.Dom.SetStyleValue(item.element, item.name, item.currentValue.ToHEX(), null);
			}
			else if (item.isClip)
			{
				Ice.ForEach(["top", "right", "bottom", "left"], function (o) { item.currentValue[o] = isOver ? item.finalValue[o] : Math.floor(GetCurrent.call(this, item.startValue[o], item.finalValue[o])) }, this);
				// css clip: rect(0px 10px 100px 20px );
				// applyt to element
				Ice.Utility.Dom.SetClip(item.element, item.currentValue);
			}
			else if(item.isTransform)
			{
				var val = item.stringFormat;
				for (var nn = 0; nn < item.currentValue.length; nn++)
				{
					item.currentValue[nn] = isOver ? item.finalValue[nn] : GetCurrent.call(this, item.startValue[nn], item.finalValue[nn]);
					val = val.replace("{" + nn + "}", item.currentValue[nn]);
				}
				// apply to element
				$$ice(item.element).ApplyTransform(val);
			}
			else if (item.isValue)
			{
				// value item or element item

				item.currentValue = isOver ? item.finalValue : GetCurrent.call(this, item.startValue, item.finalValue);

				if (item.element)
				{
					if (item.units == "px" || item.units == "%")
					{
						item.currentValue = Math.round(item.currentValue);
					}
					////Ice.Diagnostic.WriteLine(item.currentValue);
					//Ice.Diagnostic.WriteLine(Ice.JSON.Serialize(item, 1));
					// apply to element
					Ice.Utility.Dom.SetStyleValue(item.element, item.name, item.currentValue, item.units);
				}
			}
			else
			{
				Ice.Diagnostic.Exception("Ice.Animator error, unknow or unsuported");
			}



			if (item.onTick)
			{
				item.onTick.call(this, this, item.currentValue);
			}
		}

		if (this._config.onProgress)
		{
			var percent = this._instance.elapsedTimeMs / this._config.durationMs * 100;
			if (percent > 100)
			{
				percent = 100;
			}
			this._config.onProgress.call(this, this, percent); // send progress in percent
		}

		if (!isOver)
		{
			// this code has to be after the tick handler code, so if an exception occures in the code above, next setTimeout will not be executed at all
			// start next timer tick to next tick from here to avoid time between ticks to be extended by the time required to execute the rest of the code in this method
			// sleep time is shorten by time spent in processing tick method
			var nextSleepMs = Math.min(timeLeftMs, this._config.stepMs - (Date.now() - ticksAtFuncStart));
			////console.log("sleep: " + nextSleepMs);
			this._instance.timeoutId = setTimeout(TimerTick.BindContext(this), nextSleepMs > 0 ? nextSleepMs : 0);
		}
		else
		{
			////console.log("over: " + this.Items().Select("o=>o.startValue + ' ' + o.finalValue").ToStringJoined(", "));

			// completed
			this._allowNextStep = false;
			clearTimeout(this._instance.timeoutId);
			this._instance.timeoutId = null;

			// onComplete has to be invoked last because we may have this.Start() in onComplete to restart it all
			if (this._config.onComplete)
			{
				this._config.onComplete.call(this, this);
			}
		}


	}

	this.IsRunning = function ()
	{
		/// <summary>
		/// Is animator running. If it is waiting for start with Start(delay), it is not running yet.
		/// </summary>
		
		return this._instance.timeoutId != null;
	};

	this.IsWaiting = function ()
	{
		/// <summary>
		/// Is waiting for start, with Start(delay). Once the animation is started, IsWaiting returns false.
		/// </summary>

		return this._delayTimeoutId != null;
	};

	this._delayTimeoutId;

	this.Start = function (delayMs)
	{
		/// <summary>
		/// Call onStart() and start animating.
		/// </summary>
		/// <param name="delayMs" optional="true">delay animator start for delayMs miliseconds</param>
		
		if (this.IsRunning())
		{
			Ice.Diagnostic.Exception("animation is already running");
		}

		if (this._delayTimeoutId)
		{
			clearTimeout(this._delayTimeoutId);
		}

		if (delayMs > 0)
		{
			this._delayTimeoutId = setTimeout(function () { this.Start(); }.BindContext(this), delayMs);
			return this;
		}

		this._allowNextStep = true;

		if (this._config.onStart)
		{
			this._config.onStart.call(this, this);
		}

		TimerTick.BindContext(this)(true);

		return this;
	};

	this.Abort = function ()
	{
		/// <summary>
		/// Stop animating at current position. Do not call onComplete(). Call onAbort()
		/// </summary>
		
		if (this._delayTimeoutId)
		{
			clearTimeout(this._delayTimeoutId);
		}

		this._allowNextStep = false;

		if (this._instance.timeoutId)
		{
			clearTimeout(this._instance.timeoutId);
			this._instance.timeoutId = null;
		}

		if (this._config.onAbort)
		{
			this._config.onAbort.call(this, this);
		}

		return this;
	};


})();


Ice.Animator.Easing =
{
	// http://easings.net/

	// arg0: currentValue/finalValue (current position: 0...1)
	// arg1: currentValue
	// arg2: startValue
	// arg3: finalValue - startValue (total value delta)
	// arg4: duration (total duration delta)

	linear: function (p, n, firstNum, diff)
	{
		return firstNum + diff * p;
	},
	swing: function (p, n, firstNum, diff)
	{
		return ((-Math.cos(p * Math.PI) / 2) + 0.5) * diff + firstNum;
	},

	easeInQuad: function (x, t, b, c, d)
	{
		return c * (t /= d) * t + b;
	},
	easeOutQuad: function (x, t, b, c, d)
	{
		return -c * (t /= d) * (t - 2) + b;
	},
	easeInOutQuad: function (x, t, b, c, d)
	{
		if ((t /= d / 2) < 1) return c / 2 * t * t + b;
		return -c / 2 * ((--t) * (t - 2) - 1) + b;
	},
	easeInCubic: function (x, t, b, c, d)
	{
		return c * (t /= d) * t * t + b;
	},
	easeOutCubic: function (x, t, b, c, d)
	{
		return c * ((t = t / d - 1) * t * t + 1) + b;
	},
	easeInOutCubic: function (x, t, b, c, d)
	{
		if ((t /= d / 2) < 1) return c / 2 * t * t * t + b;
		return c / 2 * ((t -= 2) * t * t + 2) + b;
	},
	easeInQuart: function (x, t, b, c, d)
	{
		return c * (t /= d) * t * t * t + b;
	},
	easeOutQuart: function (x, t, b, c, d)
	{
		return -c * ((t = t / d - 1) * t * t * t - 1) + b;
	},
	easeInOutQuart: function (x, t, b, c, d)
	{
		if ((t /= d / 2) < 1) return c / 2 * t * t * t * t + b;
		return -c / 2 * ((t -= 2) * t * t * t - 2) + b;
	},
	easeInQuint: function (x, t, b, c, d)
	{
		return c * (t /= d) * t * t * t * t + b;
	},
	easeOutQuint: function (x, t, b, c, d)
	{
		return c * ((t = t / d - 1) * t * t * t * t + 1) + b;
	},
	easeInOutQuint: function (x, t, b, c, d)
	{
		if ((t /= d / 2) < 1) return c / 2 * t * t * t * t * t + b;
		return c / 2 * ((t -= 2) * t * t * t * t + 2) + b;
	},
	easeInSine: function (x, t, b, c, d)
	{
		return -c * Math.cos(t / d * (Math.PI / 2)) + c + b;
	},
	easeOutSine: function (x, t, b, c, d)
	{
		return c * Math.sin(t / d * (Math.PI / 2)) + b;
	},
	easeInOutSine: function (x, t, b, c, d)
	{
		return -c / 2 * (Math.cos(Math.PI * t / d) - 1) + b;
	},
	easeInExpo: function (x, t, b, c, d)
	{
		return (t == 0) ? b : c * Math.pow(2, 10 * (t / d - 1)) + b;
	},
	easeOutExpo: function (x, t, b, c, d)
	{
		return (t == d) ? b + c : c * (-Math.pow(2, -10 * t / d) + 1) + b;
	},
	easeInOutExpo: function (x, t, b, c, d)
	{
		if (t == 0) return b;
		if (t == d) return b + c;
		if ((t /= d / 2) < 1) return c / 2 * Math.pow(2, 10 * (t - 1)) + b;
		return c / 2 * (-Math.pow(2, -10 * --t) + 2) + b;
	},
	easeInCirc: function (x, t, b, c, d)
	{
		return -c * (Math.sqrt(1 - (t /= d) * t) - 1) + b;
	},
	easeOutCirc: function (x, t, b, c, d)
	{
		return c * Math.sqrt(1 - (t = t / d - 1) * t) + b;
	},
	easeInOutCirc: function (x, t, b, c, d)
	{
		if ((t /= d / 2) < 1) return -c / 2 * (Math.sqrt(1 - t * t) - 1) + b;
		return c / 2 * (Math.sqrt(1 - (t -= 2) * t) + 1) + b;
	},
	easeInElastic: function (x, t, b, c, d)
	{
		var s = 1.70158; var p = 0; var a = c;
		if (t == 0) return b; if ((t /= d) == 1) return b + c; if (!p) p = d * .3;
		if (a < Math.abs(c)) { a = c; var s = p / 4; }
		else var s = p / (2 * Math.PI) * Math.asin(c / a);
		return -(a * Math.pow(2, 10 * (t -= 1)) * Math.sin((t * d - s) * (2 * Math.PI) / p)) + b;
	},
	easeOutElastic: function (x, t, b, c, d)
	{
		var s = 1.70158; var p = 0; var a = c;
		if (t == 0) return b; if ((t /= d) == 1) return b + c; if (!p) p = d * .3;
		if (a < Math.abs(c)) { a = c; var s = p / 4; }
		else var s = p / (2 * Math.PI) * Math.asin(c / a);
		return a * Math.pow(2, -10 * t) * Math.sin((t * d - s) * (2 * Math.PI) / p) + c + b;
	},
	easeInOutElastic: function (x, t, b, c, d)
	{
		var s = 1.70158; var p = 0; var a = c;
		if (t == 0) return b; if ((t /= d / 2) == 2) return b + c; if (!p) p = d * (.3 * 1.5);
		if (a < Math.abs(c)) { a = c; var s = p / 4; }
		else var s = p / (2 * Math.PI) * Math.asin(c / a);
		if (t < 1) return -.5 * (a * Math.pow(2, 10 * (t -= 1)) * Math.sin((t * d - s) * (2 * Math.PI) / p)) + b;
		return a * Math.pow(2, -10 * (t -= 1)) * Math.sin((t * d - s) * (2 * Math.PI) / p) * .5 + c + b;
	},
	easeInBack: function (x, t, b, c, d, s)
	{
		if (s == undefined) s = 1.70158;
		return c * (t /= d) * t * ((s + 1) * t - s) + b;
	},
	easeOutBack: function (x, t, b, c, d, s)
	{
		if (s == undefined) s = 1.70158;
		return c * ((t = t / d - 1) * t * ((s + 1) * t + s) + 1) + b;
	},
	easeInOutBack: function (x, t, b, c, d, s)
	{
		if (s == undefined) s = 1.70158;
		if ((t /= d / 2) < 1) return c / 2 * (t * t * (((s *= (1.525)) + 1) * t - s)) + b;
		return c / 2 * ((t -= 2) * t * (((s *= (1.525)) + 1) * t + s) + 2) + b;
	},
	easeInBounce: function (x, t, b, c, d)
	{
		return c - Ice.Animator.Easing.easeOutBounce(x, d - t, 0, c, d) + b;
	},
	easeOutBounce: function (x, t, b, c, d)
	{
		if ((t /= d) < (1 / 2.75))
		{
			return c * (7.5625 * t * t) + b;
		} else if (t < (2 / 2.75))
		{
			return c * (7.5625 * (t -= (1.5 / 2.75)) * t + .75) + b;
		} else if (t < (2.5 / 2.75))
		{
			return c * (7.5625 * (t -= (2.25 / 2.75)) * t + .9375) + b;
		} else
		{
			return c * (7.5625 * (t -= (2.625 / 2.75)) * t + .984375) + b;
		}
	},
	easeInOutBounce: function (x, t, b, c, d)
	{
		if (t < d / 2) return Ice.Animator.Easing.easeInBounce(x, t * 2, 0, c, d) * .5 + b;
		return Ice.Animator.Easing.easeOutBounce(x, t * 2 - d, 0, c, d) * .5 + c * .5 + b;
	}
};


