﻿/// <reference path="Ice.Event.js" />
/// <reference path="Ice.JSON.js" />

/* example of usage:

	// create instance with UID:
	var ppp = new Ice.PostMessage("win1", onConnected);

	// register method inside of iframe or main window:
	ppp.RegisterMethod("GetWindowTime", function(val1){

	});

	// invoke method from any other iframe or main window, val1 is passed as parameter
	p1.InvokeMethod("destinationUID", "GetWindowTime", val1, function(val3){
		// response handler, val3 is return value from function GetWindowTime(val1);
	});

*/

if (typeof (Ice) == "undefined") { Ice = {}; }

Ice.PostMessage = function (UID, onConnected)
{
	/// <summary>
	/// <para>Enables sending messages to parent window.</para>
	/// <para>Crossdomain, cross window/iframe communications.</para>
	/// <para>Forwards messages between iframes and main window.</para>
	/// <para>Enables to send message between window, iframes and nested iframes from and to each other.</para>
	/// </summary>
	/// <param name="UID" type="String" optional="true">unique window id, define now, if not defined, will create random id</param>
	/// <param name="onConnected" optional="true">
	///		<para>function(data, event, this){...} invoke when connected to it's parent</para>
	///		<para>this inside of function is current PostMessage instance</para>
	/// </param>

	var _this = this;

	function GenerateUID(prefix)
	{
		return (prefix || "") + (Math.random() + "").substr(2) + Date.now();
	}

	var _postMsgUID = window._postMsgUID = UID = UID || GenerateUID("uid");

	var _listeners = window.parent && window.parent != window ? [window.parent] : [];

	var _funcBag = {};

	function EncapsulateFunc(funcRef)
	{
		/// <summary>
		/// If funcRef defined, add funcRef to _funcBag and return funcUID, else return null.
		/// </summary>
		/// <param name="funcRef" type="Function"></param>

		if (funcRef)
		{
			var cnt = 0;
			var funcUid;
			while (cnt++ < 50 && (!(funcUid = GenerateUID("fn")) || _funcBag[funcUid]))
			{
				// if uid already exists, generate new UID, try up to 50 times
			}
			_funcBag[funcUid] = funcRef;
			return funcUid;
		}
		return null;
	}

	// MSIE 9,8,7 allows to send string only

	function EncodeDataObj(obj)
	{
		return Ice.JSON.Serialize(obj);
	}

	function DecodeDataObj(obj)
	{
		try
		{
			return Ice.JSON.Deserialize(obj);
		}
		catch(e)
		{
			return null;
		}
	}

	var _ttNames = ["destinationUID", "funcName", "argument"];

	this.InvokeMethod = function (destinationUID, funcName, arg, callbackFunc)
	{
		/// <summary>
		/// 
		/// </summary>
		/// <param name="destinationUID" type="String">window or UID string</param>
		/// <param name="funcName" type="String">Function name to be executed on destinationUID and pass arg as parameter to the function.</param>
		/// <param name="arg" type="Object" optional="true">object to be passed to invoking method as argument</param>
		/// <param name="callbackFunc" type="Function" optional="true">
		///		<para>function(value, event, this){....}</para>
		///		<para>this inside of function is current PostMessage instance</para>
		///		<para>event is current event {sender:..., origin:..., data:...}</para>
		///		<para>value is sent as argument parameter in method this.InvokeMethod(....)</para>
		/// </param>

		var tt = [destinationUID, funcName, arg];

		for (var n = 0; n < tt.length; n++)
		{
			if(typeof(tt[n]) == "function")
			{
				throw "Ice.PostMessage.InvokeMethod argument " + _ttNames[n] + " can not be function";
			}
		}

		if (callbackFunc && typeof (callbackFunc) != "function")
		{
			throw "Ice.PostMessage.InvokeMethod argument callbackFunc is not function";
		}

		//#if VSINTELLISENSE
		if (Ice.IsVSIntellisene)
		{
			MessageEvent.prototype.data = "{json:true}";
			MessageEvent.prototype.dataObject = {
				// simulated object which will be received in callbackFunc
				senderUID: "",
				destinationUID: "",
				funcName: "",
				argument: {},
				isCallback: true,
				callbackFuncName: ""
			};

			callbackFunc.call(_this, {}, MessageEvent.prototype, _this);
		}
		//#endif

		for (var n = 0; n < _listeners.length; n++)
		{
			var curr;
			if ((curr = _listeners[n]) && curr != window) // do not sent to itself
			{
				curr.postMessage(EncodeDataObj({
					senderUID: _postMsgUID,
					destinationUID: destinationUID._postMsgUID ? destinationUID._postMsgUID : destinationUID, // can not send window ref, so take string _postMsgUID
					funcName: funcName,
					argument: arg,
					callbackFuncName: EncapsulateFunc(callbackFunc) // every callbackFunc may be invoked once, than it has to be removed from _funcBag
				}), "*");
			}
		}

		return _this;
	};

	this.RegisterMethod = function (funcName, funcRef)
	{
		/// <summary>
		/// Register method inside of current window which may be invoked from any other iframe.
		/// </summary>
		/// <param name="funcName" type="String">Function name.</param>
		/// <param name="funcRef" type="Function">
		///		<para>function(value, event, this){....}</para>
		///		<para>this inside of function is current PostMessage instance</para>
		///		<para>event is current event {sender:..., origin:..., data:...}</para>
		///		<para>value is sent as argument parameter in method this.InvokeMethod(....)</para>
		/// </param>

		if (typeof (funcName) != "string")
		{
			throw "Ice.PostMessage.RegisterMethod argument funcName is not string";
		}

		if (typeof (funcRef) != "function")
		{
			throw "Ice.PostMessage.RegisterMethod argument funcRef is not function";
		}

		//#if VSINTELLISENSE
		if (Ice.IsVSIntellisene)
		{
			MessageEvent.prototype.data = "{json:true}";
			MessageEvent.prototype.dataObject = {
				senderUID: "",
				destinationUID: "",
				funcName: "",
				argument: {},
				callbackFuncName:""};

			funcRef.call(_this, {}, MessageEvent.prototype, _this);
		}
		//#endif
		
		_funcBag[funcName] = funcRef;

		return _this;
	};

	function Connect(onConnected)
	{
		/// <summary>
		/// Optional. Connect to it's parent window PostMessage instance. When connected, parent may initiate communication, else, communication may be initiated from child iframe only.
		/// </summary>
		/// <param name="onConnected" optional="true">
		///		<para>function(data, event, this){...} invoke when connected to it's parent</para>
		///		<para>this inside of function is current PostMessage instance</para>
		/// </param>

		//#if VSINTELLISENSE
		if (Ice.IsVSIntellisene)
		{
			MessageEvent.prototype.data = "{json:true}";
			MessageEvent.prototype.dataObject = {
				// simulated object which will be received in onConnected
				senderUID: "",
				destinationUID: "", // callback is always sent to specific UID, never all (*)
				funcName: "",
				isCallback: true,
				argument: true
			};

			onConnected.call(_this, true, MessageEvent.prototype, _this);
		}
		//#endif

		// since we can not send reference to any object, send funcUid
		////console.log("Connect request from " + _postMsgUID);

		if (window.parent && window.parent != window)
		{
			window.parent.postMessage(EncodeDataObj({
				senderUID: _postMsgUID,
				isRegisterListener: true,
				callbackFuncName: EncapsulateFunc(onConnected) // to be executed
			}), "*");
		}

		return _this;
	};

	Ice.Event.Attach(window, "message", function (ev)
	{
		// ev.source
		// ev.origin
		// ev.data

		var dat1;

		// if any 3rd party code uses postMessage, make sure this message is from Ice.PostMessage engine:
		// Connect doesn't have destinationUID, only isRegisterListener

		if (ev && ev.data && ev.type == "message" && typeof(ev.data) == "string" && (dat1 = DecodeDataObj(ev.data)) && dat1.senderUID && (dat1.destinationUID || dat1.isRegisterListener))
		{
			////console.log("message received in " + _postMsgUID + ": " + Ice.JSON.Serialize(ev.data));

			////console.log(ev);

			//ev.data is read only, can not be set here, so use dataObject
			ev.dataObject = dat1; // replace json string with deserialized object

			var msgHandled = false;

			if (dat1.isRegisterListener)
			{
				msgHandled = true;

				_listeners.push(ev.source);

				if (dat1.callbackFuncName)
				{
					////console.log("RegisterListener response send from " + _postMsgUID + ": " + Ice.JSON.Serialize(true));

					// send response to sender only
					ev.source.postMessage(EncodeDataObj({
						senderUID: _postMsgUID,
						destinationUID: dat1.senderUID,
						funcName: dat1.callbackFuncName, // callback is always sent to specific UID, never all (*)
						isCallback: true,
						argument: true
					}), "*");
				}
			}
			else if (dat1.destinationUID == "*" || dat1.destinationUID == _postMsgUID)
			{
				var result;

				////console.log("call to func: " + dat1.funcName);

				if (dat1.funcName && _funcBag[dat1.funcName])
				{
					msgHandled = true;

					// execute funcName, and get result:
					result = _funcBag[dat1.funcName].call(_this, dat1.argument, ev, _this); //pass original event as 2nd parameter

					if (dat1.isCallback)
					{
						////console.log("delete func: " + dat1.funcName);
						// callback executed, remove it
						delete (_funcBag[dat1.funcName]);
					}
					else if (dat1.callbackFuncName)
					{
						////console.log("callback response from " + _postMsgUID + ": " + Ice.JSON.Serialize(result));

						// send response to sender only
						ev.source.postMessage(EncodeDataObj({
							senderUID: _postMsgUID,
							destinationUID: dat1.senderUID, // callback is always sent to specific UID, never all (*)
							funcName: dat1.callbackFuncName,
							isCallback: true,
							argument: result
						}), "*");
					}
				}
				else//// if (!dat1.isCallback)
				{
					// throw ex only if it isCallback, when msg is forwarded between iframes, only destination frame must have funcName defined.
					throw "Unknown funcName: " + dat1.funcName + " in "  + window._postMsgUID;
				}
			}

			// forward to other listeners:
			if(dat1.destinationUID == "*" || !msgHandled)
			{
				var curr;
				for (var n = 0; n < _listeners.length; n++)
				{
					// forward to all, but not to sender and not to itself
					if ((curr = _listeners[n]) && curr != ev.source && curr != window)
					{
						////console.log("forwarding from " + _postMsgUID + " to " + curr._postMsgUID + ": " + Ice.JSON.Serialize(dat1));
						curr.postMessage(ev.data, "*"); // forward original data object or json string
					}
				}
			}


		}
	});

	// finally, connect to parent:
	Connect(onConnected);
};

Ice.PostMessage.FindIframe = function (ev)
{
	/// <summary>
	/// Find iframe element from ifrContentWindow
	/// </summary>
	/// <param name="ifrContentWin"></param>

	var iframes = document.getElementsByTagName("iframe");
	var ifr;
	for (var n = 0; n < iframes.length; n++)
	{
		if((ifr = iframes[n]) && ifr.contentWindow == ev.source)
		{
			return ifr;
		}
	}
};

Ice.PostMessage.GetEventData = function (evnt)
{
	/// <summary>
	/// Encapsulate evnt.data object
	/// </summary>
	/// <param name="evnt">original postMessage event</param>
	
	var ev = evnt && evnt.data ? evnt.data : { data: {} };

	return {
		senderUID: ev.data.senderUID,
		destinationUID: ev.data.destinationUID,
		funcName: ev.data.funcName,
		callbackFuncName: ev.data.callbackFuncName,
		argument: ev.data.argument,
		isRegisterListener: ev.data.isRegisterListener,
		isCallback: ev.data.isCallback
	};
};
