﻿/// <reference path="Ice.js" />
/// <reference path="Ice.Diagnostic.js" />

if (typeof (Ice) == "undefined") { Ice = {}; }

Ice.Utility = new (function ()
{
	var __utility = {};

	__utility.Caret = new (function ()
	{
		function setInputSelectionRange(object, start, end)
		{
			/// <summary>
			/// Select text from start to end index. If start and end are not set, will unselect all.
			/// </summary>
			/// <param name="object"></param>
			/// <param name="start" optional="true">if not set, select from start</param>
			/// <param name="end" optional="true">if not set, select to the end of string</param>

			if (object == null || !object.type)
			{
				return;
			}
			try
			{
				if (!start || start < 0)
				{
					start = 0;
				}
				if (arguments.length < 3 || end > object.value.length)
				{
					end = object.value.length;
				}
				if (object.setSelectionRange)
				{
					object.focus();
					object.setSelectionRange(start, end);
				}
				else if (object.createTextRange)
				{
					object.focus();
					var range;
					if (object.caret)
					{
						range = object.caret;
						range.moveStart("textedit", -1);
						range.moveEnd("textedit", -1);
					} else
					{
						range = object.createTextRange();
					}
					range.moveEnd('character', end);
					range.moveStart('character', start);
					range.select();
				}
			} catch (e)
			{
				// Ignore
			}
		}

		function setCaret(object, pos)
		{
			/// <summary>
			/// Set caret position in text.
			/// </summary>
			/// <param name="object"></param>
			
			setInputSelectionRange(object, pos, pos);
		}

		function getCaret(object)
		{
			/// <summary>
			/// Returns caret position in text.
			/// </summary>
			/// <param name="object"></param>
			
			if (object == null || !object.type)
			{
				return null;
			}
			try
			{
				if (object.createTextRange && object.caret)
				{
					var range = object.caret.duplicate();
					range.moveStart('textedit', -1);
					return range.text.length;
				} else if (object.selectionStart || object.selectionStart == 0)
				{
					return object.selectionStart;
				}
			} catch (e)
			{
			}
			return null;
		}

		function getInputSelectionRange(object)
		{
			/// <summary>
			/// Returns array with start and end index of current text selection.
			/// </summary>
			/// <param name="object"></param>
			
			if (object == null || !object.type)
			{
				return null;
			}
			try
			{
				if (object.selectionEnd)
				{
					return [object.selectionStart, object.selectionEnd];
				} else if (object.createTextRange && object.caret)
				{
					var end = getCaret(object);
					return [end - object.caret.text.length, end];
				}
			} catch (e)
			{
				// Ignore
			}
			return null;
		}

		var __caret = {};
		__caret.GetCaret = getCaret;
		__caret.SetCaret = setCaret;
		__caret.GetInputSelectionRange = getInputSelectionRange;
		__caret.SetInputSelectionRange = setInputSelectionRange;
		return __caret;
	})();

	function RemoveHtmlTags(html, replacement)
	{
		replacement = Ice.Type.Coalesce(replacement, "");

		var re = new RegExp("\\<[^\\>]+\\>");
		while (html.match(re) != null)
		{
			html = html.replace(re, replacement);
		}
		return html;
	};

	__utility.Dom = new (function ()
	{
		var __dom = {};

		//#region Ice.Utility.XmlNodeType enumeration

		__dom.XmlNodeType =
		{
			ELEMENT_NODE: 1,
			ATTRIBUTE_NODE: 2,
			TEXT_NODE: 3,
			CDATA_SECTION_NODE: 4,
			ENTITY_REFERENCE_NODE: 5,
			ENTITY_NODE: 6,
			PROCESSING_INSTRUCTION_NODE: 7,
			COMMENT_NODE: 8,
			DOCUMENT_NODE: 9,
			DOCUMENT_TYPE_NODE: 10,
			DOCUMENT_FRAGMENT_NODE: 11,
			NOTATION_NODE: 12
		};

		//#endregion

		__dom.IsElementInDom = function (el)
		{
			var tmp = el;
			while (tmp && tmp != document && ((tmp.offsetParent && (tmp = tmp.offsetParent)) || (tmp.parentNode && (tmp = tmp.parentNode))))
			{
				// use offsetParent to find offset ncestor, it is less iterations thant with parentNode which finds first parent
			}

			return tmp == document; // el is in dom if tmp == document
		};

		__dom.GetInnerText = function (element)
		{
			/// <summary>
			/// Get innerText or innerHtml with removed html tags or nodeValue
			/// </summary>
			/// <param name="element"></param>
			
			return "innerText" in element ? element.innerText :
			"innerHTML" in element ? RemoveHtmlTags(element.innerHTML, "") :
			"nodeValue" in element ? element.nodeValue :
			element.value || ""; // force to return "" if all values are empty or null
		};

		__dom.GetElementScreenPosition = function (element)
		{
			/// <summary>
			/// Get element's absolute position and offsetWidth and offsetHeight, relative to page. {top:n, left:n, width:n, height:d}. 
			/// If there is content inside div with overflow:auto or scroll, always return position relative to page.
			/// </summary>
			/// <param name="element"></param>

			var tmpEl = element;

			var result = {
				left: 0,
				top: 0,
				width: element.offsetWidth,
				height: element.offsetHeight
			};

			do
			{
				result.left += (tmpEl.offsetLeft || 0) - (tmpEl.scrollLeft || 0);
				result.top += (tmpEl.offsetTop || 0) - (tmpEl.scrollTop || 0);

				if (tmpEl.style && (tmpEl.style.position == "fixed" || GetComputedStyleObject(tmpEl).position == "fixed"))
				{
					// ako je fixed element, treba uzeti offsetLeft i offsetTop i scrollLeft i scrollTop i prekinuti daljnje penjanje po DOM-u
					break;
				}
				else if (tmpEl == document.body)
				{
					// IE if FF page scroll drzi u document.documentElement, ostali browseri drze u document.body
					// zato ako je tmpEl document.body, obavezno uzeti i scroll iz document.documentElement-a
					tmpEl = tmpEl.parentNode;
				}
				else
				{
					tmpEl = tmpEl.offsetParent; // if tmpEl.style.position == "fixed", offsetParent is null, but not always, zato treba provjeriti GetComputedStyleObject(tmpEl).position == "fixed"
				}
			}
			while (tmpEl);

			return result;
		};

		__dom.GetElementPagePosition = function (element)
		{
			/// <summary>
			/// Get element's absolute position on page. {top:n, left:n, width:n, height:d}. 
			/// </summary>
			/// <param name="element"></param>

			var pos = __dom.GetElementScreenPosition(element);

			// dodati page scrolls
			var page = __dom.GetPageDimensions();

			return {
				left: pos.left + page.scrollLeft,
				top: pos.top + page.scrollTop,
				width: pos.width,
				height: pos.height
			};

		};

		var _fixedDiv;

		function GetViewportDiv()
		{
			if (!_fixedDiv)
			{
				// more secure way is to create fixed div and mesure it's dimmension: <div style='position:fixed;top:0;left:0;bottom:0;right:0;z-index:-100000;'></div>
				var tmp = document.createElement("div");
				tmp.innerHTML = "<div id='iceViewportSizeDiv' style='position:fixed;top:0;left:0;bottom:0;right:0;z-index:-10000000;'></div>";
				_fixedDiv = tmp.childNodes[0];
				document.documentElement.appendChild(_fixedDiv);
			}
			return _fixedDiv;
		}

		__dom.GetPageDimensions = function ()
		{
			/// <summary>
			/// Get current html page scroll dimensions as object: {scrollLeft:.., scrollTop:..., scrollWidth:..., scrollHeight:..., clientWidth:..., clientHeight:...}
			/// <para>scrollHeight: full height including scroll</para>
			/// <para>scrollWidth: full width including scroll</para>
			/// </summary>
			
			var dim =
			{
				// seems that scroll dimensions should be taken from body. Safari: if content height < window.innerHeight, body.scrollHeight and body.clientHeight is window.innerHeight
				// scrollHeight is offsetHeight (clientHeight + border + padding)
				// the best way to mesure window inner dimensions is to create div position:absolute;top:0;left:0;bottom:0;right:0 and then mesure div's offsetWidth and offsetHeight

				scrollLeft: Math.max(document.body.scrollLeft, document.documentElement.scrollLeft, window.pageXOffset || 0), // scroll position from left, pageXOffset is NaN on IE, max(1,500,40, NaN) is NaN
				scrollTop: Math.max(document.body.scrollTop, document.documentElement.scrollTop, window.pageYOffset || 0), // scroll position from top

				scrollWidth: Math.max(document.body.scrollWidth, document.documentElement.scrollWidth), // full width including scroll
				scrollHeight: Math.max(document.body.scrollHeight, document.documentElement.scrollHeight), // full height including scroll
				
				clientWidth: GetViewportDiv().offsetWidth, // visible size, window.innerHeigh includes scrollbar height, so better use ViewportDiv instead which excludes scrollbar width or height
				clientHeight: GetViewportDiv().offsetHeight // visible size
			};

			return dim;
		};

		__dom.GetViewportDimensions = function ()
		{
			/// <summary>
			/// Get vieport size and position on page.
			/// </summary>
			
			var dim = __dom.GetPageDimensions();
			return {
				page: dim,
				left: dim.scrollLeft,
				top: dim.scrollTop,
				width: dim.clientWidth,
				height: dim.clientHeight,
				left1: dim.scrollLeft,
				left2: dim.scrollLeft + dim.clientWidth,
				top1: dim.scrollTop,
				top2: dim.scrollTop + dim.clientHeight,
				bottom1: dim.scrollHeight - dim.clientHeight - dim.scrollTop,
				bottom2: dim.scrollHeight - dim.scrollTop,
				right1: dim.scrollWidth - dim.clientWidth - dim.scrollLeft,
				right2: dim.scrollWidth - dim.scrollLeft
			};
		};

		__dom.IsElementInViewport = function (el, onlyPart)
		{
			/// <summary>
			/// Test if element is in visible viewport.
			/// </summary>
			/// <param name="el"></param>
			/// <param name="onlyPart">if true, at least 1 dimension must be included in viewport, if false, all dimensions must be included</param>

			var pos = __dom.GetElementScreenPosition(el);
			var viewport = __dom.GetElementScreenPosition(GetViewportDiv());

			var top = pos.top >= viewport.top && pos.top <= (viewport.top + viewport.height);
			var bottom = (pos.top + pos.height) >= viewport.top && (pos.top + pos.height) <= (viewport.top + viewport.height);
			var left = pos.left >= viewport.left && pos.left <= (viewport.left + viewport.width);
			var right = (pos.left + pos.width) >= viewport.left && (pos.left + pos.width) <= (viewport.left + viewport.width);

			if (onlyPart)
			{
				var w = pos.width <= viewport.width ? (left || right) : ((left && !right) || (!left && right) || (!left && !right && pos.left < viewport.left && (pos.left + pos.width) > viewport.left));
				var h = pos.height <= viewport.height ? (top || bottom) : ((top && !bottom) || (!top && bottom) || (!top && !bottom && pos.top < viewport.top && (pos.top + pos.height) > viewport.top));
				return w && h;
			}
			else
			{
				return top && bottom && left && right;
			}
		};

		__dom.ParseClip = function(el, clipStyleString)
		{
			//margins is ice specific, defines margin from top, right, bottom, left
			// rect is css standard, defined as left, right: from left, top, bottom: from top
			var m = clipStyleString.match(new RegExp("(rect|margins)[\\(][\\s]*([\\d\\+\\-px\\%]+)[\\s]+([\\d\\+\\-px\\%]+)[\\s]+([\\d\\+\\-px\\%]+)[\\s]+([\\d\\+\\-px\\%]+)[\\s]*[\\)]"));
			m = m || [];

			return {
				match: m,
				useMargins: m && m[1] == "margins",
				top: parseFloat(m[2]) || 0,
				right: m[3] ? parseFloat(m[3]) : el.offsetWidth,
				bottom: m[4] ? parseFloat(m[4]) : el.offsetHeight,
				left:parseFloat(m[5]) || 0
			};
		};

		__dom.GetClip = function (el)
		{
			/// <summary>
			/// Get element's current clip value.
			/// </summary>
			/// <param name="el"></param>
			
			// clip: rect(0px 10px 100px 20px );
			// clip is always in px, all 4 values has to be set: top right bottom left
			
			return __dom.ParseClip(el, __dom.GetComputedStyleValue(el, "clip"));
		};

		__dom.CalculateClip = function (el, objVal)
		{
			/// <summary>
			/// Calculate clip.
			/// </summary>
			/// <param name="el"></param>
			/// <param name="objVal" type="object|string">rect|margins(top right bottom left)</param>
			/// <param name="fromRight"></param>
			/// <param name="fromBottom"></param>
			/// <param name="fromLeft"></param>

			////// ice clip value is different from css clip: has to be provided as string with 4 values "from top, from right, from bottom, from left", default unit is px, but % may be used
			////var arr = item.finalValue.split(new RegExp("[, ]")).AsEnumerable().Where(function (o) { return !!o; }).ToArray();
			////var final = { top: fromTop, right: fromRight, bottom: fromBottom, left: fromLeft };

			// css clip: rect(0px 10px 100px 20px );
			// clip is always in px, all 4 values has to be set: top right bottom left
			var curr = Ice.Utility.Dom.GetClip(el);

			if(typeof(objVal) == "string")
			{
				var mm = objVal.match(new RegExp("[\\d\\+\\-px\\%]+", "g"));
				mm = mm || [];
				// must use string values to include % if exists
				objVal = { top: mm[0], right: mm[1], bottom: mm[2], left: mm[3], useMargins: objVal.indexOf("margins") >= 0 };
			}

			////console.log(objVal);

			if (objVal.useMargins)
			{
				return {
					top: Ice.Utility.Dom.CalculateNumericValue(curr.top, objVal.top, el.offsetHeight),
					right: el.offsetWidth - Ice.Utility.Dom.CalculateNumericValue(curr.right, objVal.right, el.offsetWidth),
					bottom: el.offsetHeight - Ice.Utility.Dom.CalculateNumericValue(curr.bottom, objVal.bottom, el.offsetHeight),
					left: Ice.Utility.Dom.CalculateNumericValue(curr.left, objVal.left, el.offsetWidth)
				};
			}
			else
			{
				return {
					top: Ice.Utility.Dom.CalculateNumericValue(curr.top, objVal.top, el.offsetHeight),
					right: Ice.Utility.Dom.CalculateNumericValue(curr.left, objVal.right, el.offsetWidth),
					bottom: Ice.Utility.Dom.CalculateNumericValue(curr.top, objVal.bottom, el.offsetHeight),
					left: Ice.Utility.Dom.CalculateNumericValue(curr.left, objVal.left, el.offsetWidth)
				};
			}
		};

		__dom.SetClip = function (el, clipValue)
		{
			/// <summary>
			/// 
			/// </summary>
			/// <param name="el"></param>
			/// <param name="clipValue" type="object|string">already parsed into object or "margins(100% 0 0 0)" or "rect:(10px 100px 50px 53px)" </param>

			// this will add position:absolute to an element because clip requires absolute position
			// ice clip: "margins(fromTop  fromRight  fromBottom  fromLeft)"; // units none, px, %, +=...%, eg. .ToAnimator({ clip: "margins(100% 0 0 0)", marginTop: "-100%" })
			// css clip: "rect(fromTop fromLeftToRight fromTopToBottom fromLeft)// eg. ss clip: rect(0px 10px 100px 20px ); // always in px
			// ice clip value is different from css clip: has to be provided as string with 4 values "from top, from right, from bottom, from left", default unit is px, but % may be used


			if (typeof (clipValue) == "string")
			{
				clipValue = __dom.CalculateClip(el, clipValue);
			}
			////console.log(clipValue);

			// css clip: rect(0px 10px 100px 20px );
			var cc = clipValue;
			el.style.clip = "rect(" + cc.top + "px " + cc.right + "px " + cc.bottom + "px " + cc.left + "px" + ")";
			el.style.position = "absolute";
			////console.log(el.style.clip);
		};

		function TrimSpaces(str)
		{
			if (!str) { return str; }

			return (str || "").replace(new RegExp("(^[\\s]+)|([\\s]+$)", "ig"), "");
		}

		__dom.HasTagName = function (obj, tagName)
		{
			if (obj && obj.tagName && obj.tagName.toLowerCase() == tagName.toLowerCase())
			{
				return true;
			}
			return false;
		};

		__dom.HasClass = function (obj, cls)
		{
			return obj && obj.className && cls && (' ' + obj.className + ' ').indexOf(' ' + (cls) + ' ') >= 0;
		};

		__dom.AddClass = function (obj, cls)
		{
			if (cls && obj)
			{
				__dom.RemoveClass(obj, cls); // to prevent duplicates

				obj.className = TrimSpaces(obj.className + " " + cls);
			}
		};

		__dom.RemoveClass = function (obj, cls)
		{
			if (cls && obj)
			{
				// make sure className is not substring of some other className
				// replace with space " " if class is inside other classes: cls1 cls2 cls3, replace " " + cls2 + " " with " "
				obj.className = TrimSpaces((" " + obj.className + " ").replace(new RegExp(" " + cls + " ", "g"), " ")); // replace all matches of cls
			}
		};

		__dom.SetClass = function (obj, cls)
		{
			/// <summary>
			/// Set exact cls value, removing all previous values.
			/// </summary>
			/// <param name="obj"></param>
			/// <param name="cls"></param>/
			if (obj)
			{
				obj.className = TrimSpaces(cls) || "";
			}
		};

		__dom.ToggleClass = function (obj, cls)
		{
			if (obj && obj.className && cls)
			{
				if (__dom.HasClass(obj, cls))
				{
					__dom.RemoveClass(obj, cls);
				}
				else
				{
					__dom.AddClass(obj, cls);
				}
			}
		};

		__dom.ConvertCssNameToCamel = function (name)
		{
			/// <summary>
			/// takes border-color and returns borderColor
			/// </summary>
			/// <param name="name"></param>

			var re = new RegExp("(([\\w]*)-([a-z]))|([^\\-]+)", "g");
			var exarr = null;
			var ss = "";
			while ((exarr = re.exec(name)) != null)
			{
				ss += (exarr[2] || "") + (exarr[3] || "").toUpperCase() + (exarr[4] || "");
			}
			return ss;
		};

		__dom.ConvertCssNameFromCamel = function (name)
		{
			/// <summary>
			/// takes borderColor and returns border-color
			/// </summary>
			/// <param name="name"></param>
			var re = new RegExp("(([a-z]*)([A-Z]))|([\\w\\-]+)", "g");
			var exarr = null;
			var ss = "";
			while ((exarr = re.exec(name)) != null) // IE has "" if group not found, FF has undefined as default value in exarr
			{
				ss += (exarr[2] || "") + ((exarr[3] || "") != "" ? "-" + (exarr[3] || "").toLowerCase() : "") + (exarr[4] || "");
			}
			return ss;
		};

		////	__dom.GetParentByTagName = function (element, tagName)
		////	{
		////		var parent = element.parentNode ? element.parentNode : element.parentElement;
		////		var upperTagName = tagName.toUpperCase();
		////		while (parent && (parent.tagName.toUpperCase() != upperTagName))
		////		{
		////			parent = parent.parentNode ? parent.parentNode : parent.parentElement;
		////		}
		////		return parent;
		////	};

		function SetOpacity(element, value)
		{
			/// <summary>
			/// set opacity. If value is 100, remove opacity setting - defalut value.
			/// </summary>
			/// <param name="element"></param>
			/// <param name="value">0...100</param>
			if (element && element.style)
			{
				if (typeof (value) != "number")
				{
					value = 100;
				}

				if (value >= 100)
				{
					value = 100;
				}

				if (value < 0)
				{
					value = 0;
				}

				// chrome requires for opacity to be set, ie7 and ie8 requires filter to be set :(
				// must test filter first, ie7 has opacity, but setting opacity value doesn't work
				// must set both properties for compatibility

				if ("opacity" in element.style)
				{
					// opacity: 0...1
					element.style.opacity = value >= 0 && value < 100 ? (value / 100) : ""; // if value is 100%, set "" value
				}

				if ("filter" in element.style)
				{
					// filter opacity: 0...100
					element.style.filter = value >= 0 && value < 100 ? ("alpha(opacity=" + value + ")") : ""; // if value is 100%, set "" value
				}

			}
		};

		function GetOpacity(element)
		{
			/// <summary>
			/// Returns opacity value 0...100 or 100 if not set.
			/// </summary>
			/// <param name="element"></param>
			/// <param name="useDefault" type="Boolean">When opacity is not defined. If true, return 100, else return "".</param>

			var op = null;
			if (element && element.style)
			{
				if ("opacity" in element.style) // 0...1
				{
					op = parseFloat(element.style.opacity) * 100;
				}
				else if("filter" in element.style) // 0...100
				{
					var match = element.style.filter.match(new RegExp("opacity[\\s\\=]+([\\d]+)"));
					if (match)
					{
						op = parseFloat(match[1]); // may be 0
					}
				}
			}
			
			return typeof(op) == "number" && op >= 0 && op < 100 ? op : 100;
		};

		var rePxNumericProperties = new RegExp("top|left|bottom|right|width|height|size|radius|padding|margin", "i"); // ignore case, eg. borderTop

		function ExtractUnits(value, name)
		{
			if (name == "opacity")
			{
				return null;
			}

			var match;

			if (value && typeof (value) == "string" && (match = value.match(new RegExp("(px|%)$", "ig"))) && match && match[1])
			{
				return match[1];
			}

			if (typeof(name) == "string" && rePxNumericProperties.test(name))
			{
				return "px";
			}

			return null;
		}

		function ExtractNumericValue(value, defaultValue)
		{
			var cval = value;

			if (cval && typeof (cval) == "string")
			{
				// values: +=10, -10, +=10px, +=10%,
				// color should not be parsed: rgb(0, 3, 5), 
				var match = cval.match(new RegExp("^[\\*\\-\\+\\=]*?([\\-]{0,1}[\\d\\.]+)(px|%)?$")); // extract negative decimal number
				if (match)
				{
					cval = parseFloat(match[1]);
				}
			}

			return isNaN(parseFloat(cval)) ? defaultValue : cval;
		}

		function IsIncrementValue(value)
		{
			// eg. += 10 or -= 10 or *=20
			return value && typeof (value) == "string" && value.indexOf("=") == 1;
		}

		function IsPercent(value)
		{
			return value && typeof (value) == "string" && value.indexOf("%") > 0;
		}

		function CreateStyleValue(value, units)
		{
			if (typeof (value) == "string")
			{
				// if value is "10", add px
				// if value is "10%", replace % with units
				// "auto", "inherit" should not have units
				return value.replace(new RegExp("^([\\-]?[\\.\\d]+)(px|%)?$", "i"), "$1" + units);
			}
			else if (units)
			{
				return Math.round(value) + "" + units;
			}
			else
			{
				return Math.round(value);
			}
		}

		__dom.CalculateNumericValue = function (currValue, newValue, originalValue)
		{
			/// <summary>
			/// Calculate style value from currValue value corrected by newValue. Returns number. If there is originalValue and newValue is in %, return percentage value of originalValue
			/// </summary>
			/// <param name="currValue">numeric or string, with px, %, or no units</param>
			/// <param name="newValue">new value or eg. +=10, auto, -=20, *=30, 10px, with px, % or no units</param>
			/// <param name="originalValue">required to calculate if newValue is %</param>

			// newValue may be number, "+=10", "auto", "10px"

			if (IsIncrementValue(newValue))
			{
				var numVal = ExtractNumericValue(newValue.substr(2));
				currValue = parseFloat(currValue); // make sure it is number
				var isCurrPercent = IsPercent(currValue);

				if (IsPercent(newValue))
				{
					if (isCurrPercent)
					{
						// nothing
					}
					else
					{
						// numVal is percentage, get value from percent
						numVal = currValue * numVal / 100;
					}
				}

				var oper = newValue.match(new RegExp("^[\\+\\-\\*]\\="))[0];
				switch (oper)
				{
					case "+=":
						currValue += numVal;
						break;
					case "-=":
						currValue -= numVal;
						break;
					case "*=":
						currValue *= numVal;
						break;
				}
				return currValue;
			}
			else if(IsPercent(newValue) && typeof(originalValue) != "undefined")
			{
				if(IsPercent(originalValue))
				{
					// originalValue is in %, so just return newValue which is in % too
					return newValue;
				}
				else
				{
					// calculate value
					return parseFloat(originalValue) * parseFloat(newValue) / 100;
				}
			}
				// if there is no originalValue, don't recalculate eg. 25%, leave as is, because we don't have original value
				////else if (IsPercent(newValue))
				////{
				////	// numValue is percentage, get value from percent
				////	var numVal = ExtractNumericValue(newValue);
				////	return currValue * numVal / 100;
				////}
			else
			{
				return ExtractNumericValue(newValue);
			}
		};

		function GetComputedStyleObject(element)
		{
			// computed style on Chrowe won't work if html is not attached to dom
			if (window.getComputedStyle)
			{
				// ie 9, chrome
				return window.getComputedStyle(element);
			}
			else if (element.ownerDocument && element.ownerDocument.defaultView && element.ownerDocument.defaultView.getComputedStyle)
			{
				// firefox and other
				return element.ownerDocument.defaultView.getComputedStyle(element, null);
			}
			else if (element.currentStyle)
			{
				// ie 6, ie7, ie8
				return element.currentStyle;
			}
			return element.style;
		}

		__dom.GetComputedStyleValue = function (element, name)
		{
			/// <summary>
			/// Returns raw stayle value as string.
			/// </summary>
			/// <param name="element"></param>
			/// <param name="name"></param>
			
			var styleName = __dom.ConvertCssNameToCamel(name);
			return GetComputedStyleObject(element)[name];
		};

		__dom.GetStyleItem = function (element, name)
		{
			/// <summary>
			/// Get element style value if available, else get computed value.
			/// </summary>
			/// <param name="element"></param>
			/// <param name="name"></param>

			var styleName = __dom.ConvertCssNameToCamel(name);
			var style;
			var styleValue;
			var computedStyleValue;
			var numValue = null;
			var cunits;
			var isNumber = false;

			if (name == "opacity")
			{
				styleValue = numValue = GetOpacity(element);
			}
			else if (name.indexOf("client") == 0 || name.indexOf("offset") == 0 || name.indexOf("scroll") == 0)
			{
				// names: client|offset|scroll + Width|Height|Left|Top
				styleValue = numValue = element[styleName];
			}
			else
			{
				// get value from .style object first, if value is not set, then get from computed style object
				// eg. if name is "border-width", it may be retreived from element.style only, computed style value is always ""

				styleValue = element.style[styleName];

				if (!styleValue || name.indexOf("color") >= 0)
				{
					// styleValue is always string, so "" is false
					styleValue = GetComputedStyleObject(element)[styleName];
				}

				var num = ExtractNumericValue(styleValue, "invalid_number");
				if (num != "invalid_number")
				{
					isNumber = true;
					numValue = Math.round(num); // 0 is default for height, width, etc...
					cunits = ExtractUnits(styleValue, name);
				}
			}

			if (numValue == null)
			{
				////numValue = styleValue; // can't be converted to number
			}
			else if(numValue == "auto")
			{
				// fix for IE7 and older which returns "auto" if value is not set
				switch (styleName)
				{
					case "height":
						numValue = element.clientHeight;
						break;
					case "width":
						numValue = element.clientWidth;
						break;
				}
			}

			// return final object:
			// "name" may be: background-color, "styleName" will be backgroundColor
			// styleValue may be: 10px, then value is 10, units is px
			return { Element: element, Name: name, NumericValue: numValue, IsNumber:isNumber, Units: cunits, StyleValue: styleValue, StyleName: styleName };
		};

		__dom.GetParsedStyleValue = function (element, name)
		{
			/// <summary>
			/// Return style NumericValue or StyleValue, which one is available first.
			/// </summary>
			/// <param name="element"></param>
			/// <param name="name"></param>
			/// <param name="defaultValue"></param>
			if (element && element.style)
			{
				var item = __dom.GetStyleItem(element, name);
				return Ice.Convert.ToNumberOrDefault(item.NumericValue, item.StyleValue);
			}
			// return undefined
		};

		__dom.SetStyleValue = function (element, name, value, units)
		{
			/// <summary>
			/// 
			/// </summary>
			/// <param name="element"></param>
			/// <param name="name"></param>
			/// <param name="value">can be string '+=10' which wil increment current value or '-=10" to decrement current value</param>

			if (element && element.style)
			{
				var styleName = __dom.ConvertCssNameToCamel(name);
				var styleValue;
				var numVal;

				if (IsIncrementValue(value))
				{
					var curr = __dom.GetStyleItem(element, name); // get current first

					numVal = __dom.CalculateNumericValue(curr.NumericValue, value);
					styleValue = CreateStyleValue(numVal, units || curr.Units);
				}
				else if (typeof (value) == "string")
				{
					// eg. value = "10", use px units if units are required
					var units = ExtractUnits(value, name);
					styleValue = CreateStyleValue(value, units); // "10px", "auto", "10", ""...
					////alert(units + ", " + value + ", " + name + ", " + styleValue);
				}
				else
				{
					// numeric value
					var units = units || ExtractUnits(value, name);
					numVal = value;
					styleValue = CreateStyleValue(value, units)
				}

				if (name == "opacity")
				{
					SetOpacity(element, numVal);
				}
				else if (name.indexOf("client") == 0 || name.indexOf("offset") == 0 || name.indexOf("scroll") == 0)
				{
					// set attribute value, this is required for animator
					element[name] = numVal;
				}
				else
				{
					try
					{
						// ie 7 may throw exception if setting negative width or height
						element.style[styleName] = styleValue;

						////Ice.Diagnostic.WriteLine(styleName + ":" + styleValue);
					}
					catch (e)
					{
					}
				}
			}
		};

		__dom.GetAttributeNode = function (element, name)
		{
			/// <summary>
			/// 
			/// </summary>
			/// <param name="element" type="HTMLElement"></param>
			/// <param name="name" type="String">attribute name</param>

			if(element && element.getAttributeNode)
			{
				return element.getAttributeNode(name);
			}
		};

		__dom.GetAttributeValue = function (element, name)
		{
			/// <summary>
			/// returns attribute value
			/// </summary>
			/// <param name="element"></param>
			/// <param name="name"></param>

			if (element)
			{
				return element.getAttribute ? element.getAttribute(name) : "";
			}
		};

		__dom.SetAttributeValue = function (element, name, value)
		{
			if (element)
			{
				if (element.setAttribute)
				{
					element.setAttribute(name, value);
				}
				else
				{
					element[name] = value;
				}
			}
		};

		__dom.RemoveAttribute = function (element, name)
		{
			if (element)
			{
				if (element.removeAttribute)
				{
					element.removeAttribute(name);
				}
				else
				{
					element[name] = null;
					delete (element[name]);
				}
			}
		};

		__dom.HasAttribute = function (element, name)
		{
			if (element)
			{
				// IE8< nema .hasAttribute, zato koristiti .getAttribute
				if (element.hasAttribute)
				{
					return element.hasAttribute(name);
				}
				else if(element.getAttribute)
				{
					return !Ice.Type.IsNullOrUndefined(element.getAttribute(name));
				}
				else
				{
					return !Ice.Type.IsNullOrUndefined(element[name]);
				}
			}
			return false;
		};

		__dom.SetEnabled = function (element, enabled)
		{
			if (element)
			{
				if (enabled)
				{
					// restore onclick event
					if (element.__onclickbak)
					{
						element.onclick = element.__onclickbak;
					}

					this.RemoveAttribute(element, "disabled");

					// restore opacity
					SetOpacity(element, 100);
				}
				else
				{
					// disable

					// if a tag is disabled, firefox and chrome ignores disabled attribute and acts like it is not disabled. We have to disable click event on element.

					if (!element.__onclickbak && element.onclick)
					{
						element.__onclickbak = element.onclick;
						element.onclick = function () { return false; };
					}

					this.SetAttributeValue(element, "disabled", "disabled");

					SetOpacity(element, 40);
				}
			}
		};

		__dom.CloneNodeIfHasParent = function (node)
		{
			if (node && node.parentNode && node.cloneNode)
			{
				// if node has parent, clone it, full depth
				return node.cloneNode(true);
			}
			else
			{
				return node;
			}
		};

		__dom.RemoveNodeFromDom = function (node)
		{
			if (node && node.parentNode)
			{
				node.parentNode.removeChild(node);
			}
		};

		__dom.InsertBefore = function (refNode, elArray)
		{
			/// <summary>
			/// Insert elements before refNode without cloning - remove elements from previous position.
			/// </summary>
			/// <param name="refNode"></param>
			/// <param name="elArray">array of elements to insert, element are not cloned, so they will be removed from previous position and will be inserted into new position</param>
			if (refNode && refNode.parentNode)
			{
				if (elArray && elArray.length > 0)
				{
					// insert backwards because we have system .insertBefore() only, there is no .insertAfter()
					for (var n = elArray.length - 1; n >= 0; n--)
					{
						var n1 = elArray[n];
						if (n1)
						{
							////n1 = __dom.CloneNodeIfHasParent(n1);
							refNode.parentNode.insertBefore(n1, refNode);
							refNode = n1;
						}
					}
				}
			}
		};

		__dom.InsertAfter = function (refNode, elArray)
		{
			/// <summary>
			/// Insert elements after refNode without cloning - remove elements from previous position.
			/// </summary>
			/// <param name="refNode"></param>
			/// <param name="elArray">array of elements to insert, element are not cloned, so they will be removed from previous position and will be inserted into new position</param>
			if (refNode && refNode.parentNode)
			{
				if (elArray && elArray.length > 0)
				{
					// insert element before, than swap nodes
					// insert backwards because we have system .insertBefore() only, there is no .insertAfter()
					for (var n = 0; n < elArray.length; n++)
					{
						var n1 = elArray[n];
						if (n1)
						{
							////n1 = __dom.CloneNodeIfHasParent(n1);
							refNode.parentNode.insertBefore(n1, refNode);
							// swap nodes
							refNode.parentNode.insertBefore(refNode, n1);
							// change anchor node
							refNode = n1;
						}
					}
				}
			}
		};

		__dom.ReplaceChildrenWith = function (parent, childrenArray)
		{
			/// <summary>
			/// Replace parent's children with childrenArray.
			/// </summary>
			/// <param name="parent"></param>
			/// <param name="childrenArray">If null, remove children only. Array of elements to insert, element are not cloned, so they will be removed from previous position and will be inserted into new position</param>

			if (parent && parent.childNodes && parent.nodeType == 1)
			{
				// remove children first
				// parent.innerHTML = ""; // remove children first this way, or use removeChild()
				for (var n = parent.childNodes.length - 1; n >= 0; n--)
				{
					var nn = parent.childNodes[n];
					nn.parentNode.removeChild(nn);
				}
				if (childrenArray && childrenArray.length > 0)
				{
					// add children
					for (var n = 0; n < childrenArray.length; n++)
					{
						////parent.appendChild(__dom.CloneNodeIfHasParent(childrenArray[n]));
						parent.appendChild((childrenArray[n]));
					}
				}
			}
		};

		__dom.AddChildren = function (parent, childrenArray)
		{
			/// <summary>
			/// Add childrenArray to parent's children without clonign.
			/// </summary>
			/// <param name="parent">parent parent</param>
			/// <param name="childrenArray" type="Array">Array of elements to insert, element are not cloned, so they will be removed from previous position and will be inserted into new position</param>

			if (parent && parent.childNodes && parent.nodeType == 1)
			{
				if (childrenArray && childrenArray.length > 0)
				{
					for (var n = 0; n < childrenArray.length; n++)
					{
						var n1 = childrenArray[n];
						if (n1)
						{
							////n1 = __dom.CloneNodeIfHasParent(n1);
							parent.appendChild(n1);
						}
					}
				}
			}
		};

		__dom.InsertChildren = function (parent, childrenArray)
		{
			/// <summary>
			/// Insert childrenArray as first child to parent node without cloning.
			/// </summary>
			/// <param name="parent">parent node</param>
			/// <param name="childrenArray">Array of elements to insert, element are not cloned, so they will be removed from previous position and will be inserted into new position.</param>

			if (parent && parent.childNodes)
			{
				if (parent.childNodes.length > 0)
				{
					__dom.InsertBefore(parent.childNodes[0], childrenArray);
				}
				else
				{
					__dom.AddChildren(parent, childrenArray);
				}
			}
		};

		__dom.CenterElementIntoParent = function (el, crop, enlarge)
		{
			/// <summary>
			/// Center element into parent. Larger element is shrinked to fit into container.
			/// </summary>
			/// <param name="el" type="HTMLImageElement"></param>
			/// <param name="crop" type="Boolean" optional="true">false: fit larger dimension; true: overscan to fit smaller dimension</param>
			/// <param name="enlarge" type="Boolean" optional="true">true: smaller element is enlarged to fit into parent, false: smaller element is not enlarged</param>

			if (!el || !el.parentNode || !(el.parentNode.clientHeight > 0 && el.parentElement.clientWidth > 0))
			{
				//on IE 11 in full screen ajax loader image has offsetHeight == 0 && offsetWidth == 0
				// if el is not visible or not in dom
				return;
			}

			/// <param name="shrink" type="Boolean" optional="true" default="true">true: larger element is shrinked to fit into parent, smaller element is not resized, false: larger element is not shrinked</param>
			var shrink = true; // always shrink larger element

			el.parentNode.style.overflow = "hidden";

			if (enlarge || shrink)
			{
				if ((el.tagName || "").toLowerCase() == "img")
				{
					el.removeAttribute("width");
					el.removeAttribute("height");
					el.style.width = "";
					el.style.height = "";
				}
				el.style.margin = "";
			}

			// width and height are needed to calcucate AR only
			var elWidth = el.__initialWidth || el.offsetWidth || el.clientWidth;
			var elHeight = el.__initialHeight || el.offsetHeight || el.clientHeight;

			var ar = elWidth / elHeight || 1; // if elHeight is 0

			var w = elWidth;
			var h = elHeight;

			// if fit always shring or enlarge to fit into parent
			// if not fit: shrink if larger; if smaller, don't change size

			if ((enlarge && (w < el.parentElement.clientWidth || h < el.parentElement.clientHeight)) || (shrink && (w > el.parentElement.clientWidth || h > el.parentElement.clientHeight)))
			{
				w = el.parentElement.clientWidth;
				h = w / ar;

				if ((crop && h < el.parentElement.clientHeight) || (!crop && h > el.parentElement.clientHeight))
				{
					h = el.parentElement.clientHeight;
					w = h * ar;
				}
			}

			// WARNING: el must be positioned as relative or nothing, set position:relative or nothing
			// position has to be set using margin-top and margin-left
			// don't use position:absolute, it won't work in Firefox and Android mobile

			el.style.position = "relative"; // user relative or none, works better on Firefox and Mobile
			el.style.top = "0";
			el.style.left = "0";
			el.style.margin = "0";

			w = Math.ceil(w); // ceil: round to upper level to avoid empty space around the picture
			h = Math.ceil(h);

			// floor: round to lower level to avoid empty space around the picture
			var mleft = Math.floor((el.parentNode.clientWidth - w) / 2) - el.offsetLeft; // npr. ako parent ima align='center' i el.position='relative', moramo relativno pozicionirati
			var mtop = Math.floor((el.parentNode.clientHeight - h) / 2) - el.offsetTop;
			// positioning with margin-top and margin-right doesn't require parent to have position:absolute, while positioning with top and left requires position:absolute on parent
			el.style.marginTop = mtop + "px";
			el.style.marginLeft = mleft + "px";
			el.style.width = w + "px";
			el.style.height = h + "px";
		};

		return __dom;
	})();

	__utility.Expressions = new (function ()
	{
		var __expr = {};

		__expr.CompileLambdaExpression = function (expression, context)
		{
			/// <summary>
			/// create lambda from expression and bind method with context
			/// </summary>
			/// <param name="expression" type="String|Function">example: o=>o == 10, or: o=> new {Name: o.name, Age: o.age}, or {this.DoTask(o); return 10;}</param>
			/// <param name="context" type="Object|Null">default context that expression executes in</param>

			var func = null;
			var type = typeof (expression);

			if (expression == null)
			{
				func = function (x) { return x; };
			}
			else if (type == "function")
			{
				func = expression;
			}
			else if (type == "string")
			{
				if (expression == "")
				{
					func = function (x) { return x; };
				}
					////				else if (expression.indexOf("=>") == -1)
					////				{
					////					func = new Function("$,$$,$$$,$$$$", "return " + expression);
					////				}
				else
				{
					//var mm = expression.match(new RegExp("^[\\(\\s]*([^\\(\\)]+?)[\\)\\s]*\\=\\>[\\s]*(new)*[\\s]*(\\{)*([\\w\\W]+?)(\\})*[\\s]*$"));
					var mm = expression.match(new RegExp("^[\\(\\s]*([^\\(\\)]*)[\\)\\s]*\\=\\>[\\s]*(new)*[\\s]*(\\{)*([\\w\\W]+?)(\\})*[\\s]*$"));

					/*	
					// accepts ()=>.... or (a,b)=>fadsfsd
					{...}
					[0]: "o, n => {var a = 10; var b = 20; return a + b;}"
					[1]: "o, n"
					[2]: ""
					[3]: "{"
					[4]: "var a = 10; var b = 20; return a + b;"
					[5]: "}"
					index: 0
					input: "o, n => {var a = 10; var b = 20; return a + b;}"
					lastIndex: 47

					{...}
					[0]: "o, n => new {Name: o.Name, Age: o.Age}"
					[1]: "o, n"
					[2]: "new"
					[3]: "{"
					[4]: "Name: o.Name, Age: o.Age"
					[5]: "}"
					index: 0
					input: "o, n => new {Name: o.Name, Age: o.Age}"
					lastIndex: 51

					{...}
					[0]: "o, n => o.Age > 10 && o.Legs > 1"
					[1]: "o, n"
					[2]: ""
					[3]: ""
					[4]: "o.Age > 10 && o.Legs > 1"
					[5]: ""
					index: 0
					input: "o, n => o.Age > 10 && o.Legs > 1"
					lastIndex: 32
					*/
					if (mm[3] == "{" && mm[5] == "}")
					{
						if (mm[2] == "new") // "o, n => new {Name: o.Name, Age: o.Age}"
						{
							func = new Function(mm[1], "return {" + mm[4] + "}");
						}
						else // "o, n => {var a = 10; var b = 20; return a + b;}"
						{
							func = new Function(mm[1], mm[4]);
						}
					}
					else // "o, n => o.Age > 10 && o.Legs > 1"
					{
						try
						{
							func = new Function(mm[1], "return " + mm[4] + ";");
						}
						catch(e)
						{
							var dd = 10;
						}
					}
				}
			}

			if (func == null)
			{
				Ice.Diagnostic.Exception("Unsupported expression in Ice.Utility.Expressions.CompileLambdaExpression(" + expression + ")");
			}

			if (context != null)
			{
				return function () { return func.apply(context, arguments); }; // bind method with context
			}
			return func;
		};

		return __expr;
	})();

	__utility.LZW = new (function ()
	{
		var __lzw = {};
		__lzw.CompressString = function (s)
		{
			///<summary>LZW-compress a string</summary>
			var dict = {};
			var data = (s + "").split("");
			var out = [];
			var currChar;
			var phrase = data[0];
			var code = 256;
			for (var i = 1; i < data.length; i++)
			{
				currChar = data[i];
				if (dict[phrase + currChar] != null)
				{
					phrase += currChar;
				}
				else
				{
					out.push(phrase.length > 1 ? dict[phrase] : phrase.charCodeAt(0));
					dict[phrase + currChar] = code;
					code++;
					phrase = currChar;
				}
			}
			out.push(phrase.length > 1 ? dict[phrase] : phrase.charCodeAt(0));
			for (var i = 0; i < out.length; i++)
			{
				out[i] = String.fromCharCode(out[i]);
			}
			return out.join("");
		};

		__lzw.DecompressString = function (s)
		{
			///<summary>Decompress an LZW-encoded string</summary>
			var dict = {};
			var data = (s + "").split("");
			var currChar = data[0];
			var oldPhrase = currChar;
			var out = [currChar];
			var code = 256;
			var phrase;
			for (var i = 1; i < data.length; i++)
			{
				var currCode = data[i].charCodeAt(0);
				if (currCode < 256)
				{
					phrase = data[i];
				}
				else
				{
					phrase = dict[currCode] ? dict[currCode] : (oldPhrase + currChar);
				}
				out.push(phrase);
				currChar = phrase.charAt(0);
				dict[code] = oldPhrase + currChar;
				code++;
				oldPhrase = phrase;
			}
			return out.join("");
		};
		return __lzw;
	})();

	// RFC JavaScript table 36 Ice extended to 64
	var base64CharsTable = {
		JavaScriptRFCTable64: "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_-!", // acceptable for url:  ! - _ ( )
		StandardBase64RFC: "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/="
	};

	__utility.IceBaseConvert = function (useStandardBase64RFCTable)
	{
		/// <summary>
		/// ctor.
		/// </summary>
		/// <param name="useStandardBase64RFCTable" optional="true" type="Boolean">default is false, if false, will use JavaScriptRFCTable64, if true, will use StandardBase64RFC</param>

		this.CharsTable = useStandardBase64RFCTable ? base64CharsTable.StandardBase64RFC : base64CharsTable.JavaScriptRFCTable64;
	};

	__utility.IceBaseConvert.prototype = new (function ()
	{
		this.CharsTable = base64CharsTable.JavaScriptRFCTable64;

		//#region private methods

		function BitsEnumerator(array, fromBits)
		{
			/// <summary>
			/// 
			/// </summary>
			/// <param name="array">byte enumerable</param>
			/// <param name="fromBits">1-6, 8</param>
			/// <returns type="">bit enumerable</returns>
			var bytesEnumerator = null;
			var bitsEnumerator = null;

			return Ice.Collections.CreateEnumerator(function ()
			{
				bytesEnumerator = Ice.Collections.CreateEnumerator(array);
			},
			function (yldobj)
			{
				while (true)
				{
					if (bitsEnumerator != null && bitsEnumerator.YieldNext(yldobj))
					{
						return true;
					}

					if (!bytesEnumerator.YieldNext(yldobj))
					{
						return false; // no more data, exit all
					}

					var bitsArr = [];
					for (var n = fromBits - 1; n >= 0; n--)
					{
						bitsArr.push((yldobj.YieldValue & ((1) << n)) > 0);
					}

					bitsEnumerator = Ice.Collections.CreateEnumerator(bitsArr);
				}
			});
		}

		function ConvertBytes(array, fromBits, toBits)
		{
			/// <summary>
			/// 
			/// </summary>
			/// <param name="array">byte array</param>
			/// <param name="fromBits">1-6, 8</param>
			/// <param name="toBits">1-6, 8</param>
			/// <returns type="">byte[]</returns>

			var bitsEnumerator = new BitsEnumerator(array, fromBits);
			var bytes = [];
			var bb = 0;
			var count = 0;
			var yldobj = new Ice.Collections.Yield();
			while (bitsEnumerator.YieldNext(yldobj))
			{
				count++;
				bb = (bb | ((yldobj.YieldValue ? 1 : 0) << (toBits - count)));

				if (count == toBits)
				{
					bytes.push(bb);
					count = 0;
					bb = 0;
				}
			}

			if (toBits < 8 && count > 0)
			{
				// ostatak
				// ako bits == 8, ignoriramo ostatak jer je nastao prethodnim paddingom do punog broja, npr kad se reducira sa 8 na 6 bitova, zadnji byte treba popuniti s ostatkom do 6
				// pri povratku sa 6 na 8 bitova, ignoriramo bitove za padding do 6
				bytes.push(bb);
			}

			return bytes;
		};

		function Convert(data, fromBits, toBits)
		{
			if (!fromBits || (fromBits != 8 && (fromBits < 1 || fromBits > 6)))
			{
				throw (fromBits + " is invalid value. fromBits can be in range [1-6] or 8");
			}
			if (!toBits || (toBits != 8 && (toBits < 1 || toBits > 6)))
			{
				throw (toBits + " is invalid value. toBits can be in range [1-6] or 8");
			}

			var convertedEnumerable = ConvertBytes(data, fromBits, toBits);

			if (this.CharsTable == base64CharsTable.StandardBase64RFC)
			{
				// standard base64 RFC uses padding at the end

				if (fromBits == 8 && toBits == 6)
				{
					// encoding
					if (data.length % 3 > 0)
					{
						// ako je Base64 RFC table, use padding
						var pad = 3 - data.length % 3;
						var alignmentCharIndex = (this.CharsTable.length - 1);
						while (pad > 0)
						{
							pad--;
							convertedEnumerable.push(alignmentCharIndex);
						}
					}
				}
				else if (fromBits == 6 && toBits == 8)
				{
					// decoding
					// count alignment chars
					var alignmentCharIndex = (this.CharsTable.length - 1);
					var cnt = 0;
					var len = data.length;
					while (cnt < len && (data[len - 1 - cnt] == alignmentCharIndex))
					{
						cnt++;
					}

					var take = convertedEnumerable.length - cnt;

					convertedEnumerable = convertedEnumerable.slice(0, take);
				}
			}

			return convertedEnumerable;
		}

		//#endregion

		this.ToBaseString = function (arrayOrString, toBits)
		{
			/// <summary>
			/// <para>Convert string or array of chars or 8-bits based items to base(toBits) string using.</para>
			/// <para>Use Ice.Utility.UTF8.GetBytes() and Ice.Utility.UTF8.GetString() to converto from or to arrayOrString</para>
			/// </summary>
			/// <param name="arrayOrString">byte[] or string</param>
			/// <param name="toBits">1-6</param>
			/// <returns type="">baseString</returns>

			if (arrayOrString && arrayOrString.charAt)
			{
				// convert to array
				arrayOrString = Ice.Utility.UTF8.GetBytes(arrayOrString);
			}

			var arr = Convert.call(this, arrayOrString, 8, toBits);
			var sb = [];
			for (var n = 0; n < arr.length; n++)
			{
				sb.push(this.CharsTable.charAt(arr[n]));
			}
			return sb.join("");
		};

		this.FromBaseString = function (baseString, fromBits)
		{
			/// <summary>
			/// <para>Convert base(fromBits) encoded string to aray of 8-bits based items using.</para>
			/// <para>Use Ice.Utility.UTF8.GetBytes() and Ice.Utility.UTF8.GetString() to converto from or to array</para>
			/// </summary>
			/// <param name="baseString">string</param>
			/// <param name="fromBits">1-6</param>
			/// <returns type="">byte[]</returns>

			var arr = [];
			for (var n = 0; n < baseString.length; n++)
			{
				arr.push(this.CharsTable.indexOf(baseString.charAt(n)));
			}
			return Convert.call(this, arr, fromBits, 8);
		};

		this.DecodeBaseStringToString = function (baseString, fromBits)
		{
			return Ice.Utility.UTF8.GetString(this.FromBaseString(baseString, fromBits));
		};

		this.ToHexString = function (array)
		{
			/// <summary>
			/// Convert array or 8-bits based items to hex string.
			/// </summary>
			/// <param name="array"></param>

			return this.ToBaseString(array, 4);
		};

		this.FromHexString = function (hexString)
		{
			/// <summary>
			/// Convert base(fromBits) encoded string to aray of 8-bits based items. Returns array of bytes.
			/// </summary>
			/// <param name="hexString"></param>

			return this.FromBaseString(hexString, 4);
		};

	})();

	function EncodeUTF8(string)
	{
		string = string.replace(new RegExp("\r\n", "g"), "\n");
		var utftext = "";

		for (var n = 0; n < string.length; n++)
		{

			var c = string.charCodeAt(n);

			if (c < 128)
			{
				utftext += String.fromCharCode(c);
			}
			else if ((c > 127) && (c < 2048))
			{
				utftext += String.fromCharCode((c >> 6) | 192);
				utftext += String.fromCharCode((c & 63) | 128);
			}
			else
			{
				utftext += String.fromCharCode((c >> 12) | 224);
				utftext += String.fromCharCode(((c >> 6) & 63) | 128);
				utftext += String.fromCharCode((c & 63) | 128);
			}

		}

		return utftext;
	}

	function DecodeUTF8(utftext)
	{
		var string = "";
		var i = 0, c = 0, c1 = 0, c2 = 0, c3 = 0;

		while (i < utftext.length)
		{
			c = utftext.charCodeAt(i);

			if (c < 128)
			{
				string += String.fromCharCode(c);
				i++;
			}
			else if ((c > 191) && (c < 224))
			{
				c2 = utftext.charCodeAt(i + 1);
				string += String.fromCharCode(((c & 31) << 6) | (c2 & 63));
				i += 2;
			}
			else
			{
				c2 = utftext.charCodeAt(i + 1);
				c3 = utftext.charCodeAt(i + 2);
				string += String.fromCharCode(((c & 15) << 12) | ((c2 & 63) << 6) | (c3 & 63));
				i += 3;
			}

		}

		return string;
	}

	__utility.UTF8 = new (function ()
	{
		var __utf8 = {};
		__utf8.GetBytes = function (str)
		{
			var input = EncodeUTF8(str);
			var arr = [];
			for (var n = 0; n < input.length; n++)
			{
				arr.push(input.charCodeAt(n));
			}
			return arr;
		};

		__utf8.GetString = function (arr)
		{
			var chars = [];
			for (var n = 0; n < arr.length; n++)
			{
				chars.push(String.fromCharCode(arr[n]));
			}
			return DecodeUTF8(chars.join(""));
		};
		return __utf8;
	})();

	function Base64Engine(table)
	{
		var __base64 = {};

		var _keyStr = table || base64CharsTable.StandardBase64RFC;

		__base64.ToBase64String = function (str)
		{
			/// <summary>
			/// Encode str to base64 using RFC base64 table. Returns base64 encoded string.
			/// </summary>
			/// <param name="str">raw string</param>

			var output = "";
			var chr1, chr2, chr3, enc1, enc2, enc3, enc4;
			var i = 0;

			var input = __utility.UTF8.GetBytes(str);
			var inplen = input.length;

			while (i < inplen)
			{
				chr1 = input[i++];
				chr2 = input[i++];
				chr3 = input[i++];

				enc1 = chr1 >> 2;
				enc2 = ((chr1 & 3) << 4) | (chr2 >> 4);
				enc3 = ((chr2 & 15) << 2) | (chr3 >> 6);
				enc4 = chr3 & 63;

				if (isNaN(chr2))
				{
					enc3 = enc4 = 64;
				}
				else if (isNaN(chr3))
				{
					enc4 = 64;
				}

				output = output +
			_keyStr.charAt(enc1) + _keyStr.charAt(enc2) +
			_keyStr.charAt(enc3) + _keyStr.charAt(enc4);

			}

			return output;
		};

		__base64.FromBase64String = function (str)
		{
			/// <summary>
			/// Decode str from base64 using RFC base64 table. Returns raw string.
			/// </summary>
			/// <param name="str">RFC base64 encoded string</param>

			var output = [];
			var chr1, chr2, chr3;
			var enc1, enc2, enc3, enc4;
			var i = 0;

			var input = str.replace(/[^A-Za-z0-9\+\/\=]/g, "");

			while (i < input.length)
			{
				enc1 = _keyStr.indexOf(input.charAt(i++));
				enc2 = _keyStr.indexOf(input.charAt(i++));
				enc3 = _keyStr.indexOf(input.charAt(i++));
				enc4 = _keyStr.indexOf(input.charAt(i++));

				chr1 = (enc1 << 2) | (enc2 >> 4);
				chr2 = ((enc2 & 15) << 4) | (enc3 >> 2);
				chr3 = ((enc3 & 3) << 6) | enc4;

				output.push(chr1);

				if (enc3 != 64)
				{
					output.push(chr2);
				}
				if (enc4 != 64)
				{
					output.push(chr3);
				}

			}

			return __utility.UTF8.GetString(output);

		};

		return __base64;
	}

	// Standard base64 encoder
	__utility.RfcBase64 = new Base64Engine(base64CharsTable.StandardBase64RFC);

	// Ice base64 encoder for urls
	__utility.IceBase64 = new Base64Engine(base64CharsTable.JavaScriptRFCTable64);


	__utility.Html = new (function ()
	{
		var __html = {};

		__html.HtmlEncodeSimple = function (html)
		{
			return html.replace(new RegExp("<", "g"), "&lt;").replace(new RegExp(">", "g"), "&gt;");
		};

		__html.HtmlDecodeSimple = function (html)
		{
			return html.replace(new RegExp("&lt;", "g"), "<").replace(new RegExp("&gt;", "g"), ">");
		};

		__html.UrlEncode = function (string)
		{
			return escape(EncodeUTF8(string)); // will not encode right slash in http://something
			//return encodeURIComponent(EncodeUTF8(string)); // will encode right slash in http://something
		};

		__html.UrlDecode = function (string)
		{
			return DecodeUTF8(unescape(string));
		};

		__html.RemoveHtmlTags = RemoveHtmlTags;

		return __html;
	})();

	__utility.HEX = new (function ()
	{
		var hex_tab = "0123456789abcdef";
		var __hex = {};

		this.GetHexFromBytes = function (binarray)
		{
			var str = "";
			for (var i = 0; i < binarray.length * 4; i++)
			{
				str += hex_tab.charAt((binarray[i >> 2] >> ((3 - i % 4) * 8 + 4)) & 0xF) +
			hex_tab.charAt((binarray[i >> 2] >> ((3 - i % 4) * 8)) & 0xF);
			}
			return str;
		};
		return __hex;
	})();


	return __utility;
})();
