﻿using Ice.Utility;
using Ice.Web.UI.Direct;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Web;
using System.Web.UI;
using System.Web.UI.WebControls;

namespace Ice.Web.Script.Services
{
	public class AjaxUpdateManager
	{
		internal class Item
		{
			public string selector { get; set; }
			public string html { get; set; }
			public string pageTitle { get; set; }
		}

		internal List<Item> Items = new List<Item>();

		public AjaxUpdateManager Append(string selector, string html, string pageTitle = null)
		{
			this.Items.Add(new Item() { selector = selector, html = html, pageTitle = pageTitle });
			return this;
		}
	}

	public class IceResponse
	{
		public static IceResponse Current
		{
			get
			{
				IceResponse curr = HttpContext.Current.Items["_iceResponseCurrent"] as IceResponse;
				if(curr == null)
				{
					HttpContext.Current.Items["_iceResponseCurrent"] = curr = new IceResponse();
				}
				return curr;
			}
		}

		private List<string> _javaScripts = new List<string>();
		private List<object> _responseObj = new List<object>();
		private List<string> _exceptionMsg = new List<string>();

		internal bool ResponseEnded { get; private set; }

		/// <summary>
		/// Clear both locked and current response data.
		/// </summary>
		/// <returns></returns>
		public IceResponse ClearResponse()
		{
			_javaScripts.Clear();
			_responseObj.Clear();
			_exceptionMsg.Clear();
			return this;
		}

		/// <summary>
		/// Invoke Response.End() which ends current execution and send current data to the client.
		/// </summary>
		public void EndResponse()
		{
			this.ResponseEnded = true;
			HttpContext.Current.Response.End();
		}

		/// <summary>
		/// Set response object, remove previous object if exists.
		/// </summary>
		/// <param name="obj"></param>
		/// <returns></returns>
		public IceResponse SetResponseObject(object obj)
		{
			////_dic.Clear(); // don't clear, it may already have JavaScript
			_responseObj.Clear();
			_responseObj.Add(obj);
			return this;
		}

		/// <summary>
		/// Clear all existing data and set exception message.
		/// </summary>
		/// <param name="exceptionMessage"></param>
		/// <returns></returns>
		public IceResponse SetExceptionMessage(string exceptionMessage)
		{
			ClearResponse();
			_exceptionMsg.Add(exceptionMessage);
			return this;
		}

		/// <summary>
		/// Append javascript to be executed on client.
		/// </summary>
		/// <param name="js"></param>
		/// <returns></returns>
		public IceResponse ExecuteJavaScript(string js, bool insertBefore = false)
		{
			if(!string.IsNullOrEmpty(js))
			{
				if (insertBefore)
				{
					_javaScripts.Insert(0, js);
				}
				else
				{
					_javaScripts.Add(js);
				}
			}
			return this;
		}

		/// <summary>
		/// Append alert(message) to javascritp which will be executed on client.
		/// </summary>
		/// <param name="message"></param>
		/// <returns></returns>
		public IceResponse DisplayAlert(string message)
		{
			return this.ExecuteJavaScript("alert(\"" + Ice.Web.Extensions.ControlExtensions.JSEncodeString(message) + "\")");
		}

		/// <summary>
		/// Append window.location=url to execute on client. If waitMs &gt; 0, create setTimeout(function(){window.location=url;}, waitMs).
		/// </summary>
		/// <param name="url"></param>
		/// <param name="waitMs"></param>
		/// <returns></returns>
		public IceResponse SetWindowLocation(string url, int waitMs = 0)
		{
			if (waitMs == 0)
			{
				return this.ExecuteJavaScript("window.location='" + url + "'");
			}
			else
			{
				return this.ExecuteJavaScript("setTimeout(function(){window.location='" + url + "';}, " + waitMs + ")");
			}
		}

		public IceResponse ReloadWindowLocation(int waitMs = 0)
		{
			string js;
			if (waitMs == 0)
			{
				js = "window.location.reload()";
			}
			else
			{
				js = "setTimeout(function(){window.location.reload();}, " + waitMs + ");";
			}
			return this.ExecuteJavaScript(js);
		}

		/// <summary>
		/// Use iceSelector to select control(s). Uses IceAjax to send response.
		/// </summary>
		/// <param name="iceSelector"></param>
		/// <param name="action"></param>
		/// <returns></returns>
		public IceResponse UseIceDirect(string iceSelector, Action<IceDirect> action)
		{
			IceDirect direct = IceDirect.FromSelector(iceSelector);
			action(direct);
			// there is no Page, so it can't be registered with MS ScriptManager
			// register with IceWeb
			return this.ExecuteJavaScript(direct.CurrentJs);
		}

		public IceResponse UseAjaxUpdate(bool saveAjaxHistory, Action<AjaxUpdateManager> action)
		{
			AjaxUpdateManager manager = new AjaxUpdateManager();
			action(manager);

			if (manager.Items.Any())
			{
				List<string> list = new List<string>();

				list.Add("Ice.AjaxHistory.BeginUpdate()");

				manager.Items.ForEach(o =>
					list.Add(".Update(" +
						Ice.Web.Extensions.ControlExtensions.JSEncodeString(o.selector, "\"") + ", " +
						Ice.Web.Extensions.ControlExtensions.JSEncodeString(o.html, "\"")
						+ ")")
				);

				list.Add(".Commit(" + (saveAjaxHistory ? "true" : "false") + ");");

				// be carefull, this js has to be inserted before any other js (eg. IceToolTip, when rendered, creates js Setup func which has to be executed AFTER the html is inserted into dom)
				this.ExecuteJavaScript(list.ToStringJoined(), true);
			}

			return this;
		}

		internal object GetObjectForSerialization()
		{
			Dictionary<string, object> dic = new Dictionary<string, object>();
			
			dic.Add("Type", "IceWebServiceResponse");
			
			if(_exceptionMsg.Any())
			{
				dic.Add("ExceptionMessage", _exceptionMsg.First());
			}

			if(_responseObj.Any())
			{
				dic.Add("ResponseObject", _responseObj.First());
			}

			if(_javaScripts.Any())
			{
				dic.Add("JavaScript", _javaScripts.ToStringJoined(";") + ";");
			}

			return dic;
		}
		
	}


}
