﻿namespace Ice.Web.Script.Services
{
	using System;
	using System.Collections.Generic;
	using System.Globalization;
	using System.Reflection;
	using System.Web.Services;
	using Resources;
	using Ice.Serialization.Json.Core.Serializer;
	using System.Linq;

    class WebServiceMethodData
    {
        private Dictionary<string, WebServiceParameterData> _parameterData;
		private List<KeyValuePair<string, WebServiceParameterData>> _parameterDataList;
		private IceWebMethodAttribute _methodAttribute;

		internal WebServiceMethodData(WebServiceData owner, System.Reflection.MethodInfo methodInfo)
		{
			this.Owner = owner;
			this.MethodInfo = methodInfo;
			_methodAttribute = methodInfo.GetCustomAttributes(typeof(IceWebMethodAttribute), true).OfType<IceWebMethodAttribute>().FirstOrDefault() ?? new IceWebMethodAttribute();
		}

        private object CallMethod(object target, IDictionary<string, object> parameters)
        {
            this.EnsureParameters();
            object[] objArray = new object[this._parameterData.Count];
            foreach (WebServiceParameterData data in this._parameterData.Values)
            {
                object obj2;
                if (!parameters.TryGetValue(data.ParameterInfo.Name, out obj2))
                {
                    throw new InvalidOperationException(string.Format(CultureInfo.InvariantCulture, AtlasWeb.WebService_MissingArg, new object[] { data.ParameterInfo.Name }));
                }
                objArray[data.Index] = obj2;
            }
            return this.MethodInfo.Invoke(target, objArray);
        }

        internal object CallMethodFromRawParams(object target, IDictionary<string, object> parameters, IceAjaxVersionEnum ajaxVersion)
        {
			parameters = this.StrongTypeParameters(parameters, ajaxVersion);
            return this.CallMethod(target, parameters);
        }

        private void EnsureParameters()
        {
            if (this._parameterData == null)
            {
                lock (this)
                {
                    Dictionary<string, WebServiceParameterData> dictionary = new Dictionary<string, WebServiceParameterData>();
                    int index = 0;
                    foreach (ParameterInfo info in this.MethodInfo.GetParameters())
                    {
                        dictionary[info.Name] = new WebServiceParameterData(info, index);
                        index++;
                    }
                    this._parameterData = dictionary;
                }
            }
        }

		private IDictionary<string, object> StrongTypeParameters(IDictionary<string, object> rawParams, IceAjaxVersionEnum ajaxVersion)
        {
            IDictionary<string, object> dictionary2 = new Dictionary<string, object>(rawParams.Count);

			if(ajaxVersion == IceAjaxVersionEnum.Array) // method arguments as List<object> - each argument is an object
			{
				List<KeyValuePair<string, WebServiceParameterData>> parameterDataList = this.ParameterDataList;

				int index = 0;

				foreach (KeyValuePair<string, object> pair in rawParams)
				{
					string key = pair.Key;
					
					Type parameterType = parameterDataList[index].Value.ParameterInfo.ParameterType;
					dictionary2[parameterDataList[index].Key] = ObjectConverter.ConvertObjectToType(pair.Value, parameterType, this.Owner.Serializer);

					index++;
				}
				return dictionary2;

			}
			else // method arguments as Dictionary<string, object> - string is method parameter name, object is an argument to method
			{
				IDictionary<string, WebServiceParameterData> parameterDataDictionary = this.ParameterDataDictionary;
				foreach (KeyValuePair<string, object> pair in rawParams)
				{
					string key = pair.Key;
					if (parameterDataDictionary.ContainsKey(key))
					{
						Type parameterType = parameterDataDictionary[key].ParameterInfo.ParameterType;
						dictionary2[key] = ObjectConverter.ConvertObjectToType(pair.Value, parameterType, this.Owner.Serializer);
					}
				}
				return dictionary2;
			}
        }

		/// <summary>
		/// Current method ajax request timeout in ms. Default is null meaning will use global timeout value.
		/// </summary>
		internal int? Timeout { get { return _methodAttribute.Timeout > 0 ? _methodAttribute.Timeout : (int?) null; } }

		internal bool IsSynchronous { get { return _methodAttribute.Synchronous; } }

		internal bool SessionState { get { return _methodAttribute.SessionState; } }

        internal bool IsStatic{get{return this.MethodInfo.IsStatic;}}

		internal System.Reflection.MethodInfo MethodInfo { get; set; }

        internal string MethodName{get{return this.MethodInfo.Name;}}

        internal WebServiceData Owner{get;set;}

        internal IDictionary<string, WebServiceParameterData> ParameterDataDictionary
        {
            get
            {
                this.EnsureParameters();
                return this._parameterData;
            }
        }

		internal List<KeyValuePair<string, WebServiceParameterData>> ParameterDataList
		{
			get
			{
				if (this._parameterDataList == null)
				{
					this._parameterDataList = new List<KeyValuePair<string, WebServiceParameterData>>();
					foreach (KeyValuePair<string, WebServiceParameterData> kv in this.ParameterDataDictionary)
					{
						this._parameterDataList.Add(kv);
					}
				}
				return this._parameterDataList;
			}
		}

        internal ICollection<WebServiceParameterData> ParameterDatas
        {
            get
            {
                return this.ParameterDataDictionary.Values;
            }
        }

        internal Type ReturnType{get{return this.MethodInfo.ReturnType;}}

    }
}

